//
// SydB☮x: seccomp and landlock based application sandbox with support for namespaces
// src/1/main.rs: Simple init system to run under SydB☮x
//
// Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
// Based in part upon sinit which is:
//     Copyright (c) 2014-2015 Dimitris Papastamos <sin@2f30.org>
//     Released under the terms of: MIT/X Consortium License
//
// SPDX-License-Identifier: GPL-3.0-or-later

//! syd-init: Simple init system to run under SydB☮x

// We like clean and simple code with documentation.
#![deny(missing_docs)]
#![deny(clippy::allow_attributes_without_reason)]
#![deny(clippy::arithmetic_side_effects)]
#![deny(clippy::as_ptr_cast_mut)]
#![deny(clippy::as_underscore)]
#![deny(clippy::assertions_on_result_states)]
#![deny(clippy::borrow_as_ptr)]
#![deny(clippy::branches_sharing_code)]
#![deny(clippy::case_sensitive_file_extension_comparisons)]
#![deny(clippy::cast_lossless)]
#![deny(clippy::cast_possible_truncation)]
#![deny(clippy::cast_possible_wrap)]
#![deny(clippy::cast_precision_loss)]
#![deny(clippy::cast_ptr_alignment)]
#![deny(clippy::cast_sign_loss)]
#![deny(clippy::checked_conversions)]
#![deny(clippy::clear_with_drain)]
#![deny(clippy::clone_on_ref_ptr)]
#![deny(clippy::cloned_instead_of_copied)]
#![deny(clippy::cognitive_complexity)]
#![deny(clippy::collection_is_never_read)]
#![deny(clippy::copy_iterator)]
#![deny(clippy::create_dir)]
#![deny(clippy::dbg_macro)]
#![deny(clippy::debug_assert_with_mut_call)]
#![deny(clippy::decimal_literal_representation)]
#![deny(clippy::default_trait_access)]
#![deny(clippy::default_union_representation)]
#![deny(clippy::derive_partial_eq_without_eq)]
#![deny(clippy::doc_link_with_quotes)]
#![deny(clippy::doc_markdown)]
#![deny(clippy::explicit_into_iter_loop)]
#![deny(clippy::explicit_iter_loop)]
#![deny(clippy::fallible_impl_from)]
#![deny(clippy::missing_safety_doc)]
#![deny(clippy::undocumented_unsafe_blocks)]

use std::process::ExitCode;

use errno::{errno, Errno};
use libc::*;

// CONFIGURATION
const RC_INIT: [&[u8]; 1] = [b"rc.init\0"];
const RC_SHUT: &[u8] = b"rc.shutdown\0";
const RC_REBOOT: [&[u8]; 2] = [RC_SHUT, b"reboot\0"];
const RC_PWROFF: [&[u8]; 2] = [RC_SHUT, b"poweroff\0"];
////////////////

const TIMEO: c_uint = 30;
const PR_SET_PDEATHSIG: c_int = 1;

fn sigreap() {
    let mut status: c_int = 0;
    // SAFETY: Status is a valid pointer and other flags are safe.
    while unsafe { waitpid(0, &mut status, WNOHANG) } == 0 {}

    // SAFETY: The alarm function is used to set a
    // timer that will send the SIGALRM signal to
    // the current process after TIMEO seconds.
    // This is safe as long as the program has an
    // appropriate signal handler set up for SIGALRM
    // or it's being correctly managed.
    unsafe {
        alarm(TIMEO);
    }
}

fn spawn(argv: &[&[u8]]) -> Result<(), Errno> {
    // SAFETY: In libc we trust.
    match unsafe { fork() } {
        -1 => Err(errno()),
        0 => {
            // SAFETY: Create an uninitialized signal set
            let mut sigset: sigset_t = unsafe {
                let mut uninit = core::mem::MaybeUninit::<sigset_t>::uninit();
                sigemptyset(uninit.as_mut_ptr());
                uninit.assume_init()
            };
            // SAFETY: Unblock signals
            let _ = unsafe { sigprocmask(libc::SIG_UNBLOCK, core::ptr::null(), &mut sigset) };
            // SAFETY: Create a new session
            let _ = unsafe { setsid() };
            // SAFETY: In libc we trust.
            let _ = unsafe { prctl(PR_SET_PDEATHSIG, SIGKILL, 0, 0, 0) };
            let mut c_args: [*const c_char; 3] = [core::ptr::null::<c_char>(); 3]; // +1 for the null terminator
            for (i, arg) in argv.iter().enumerate() {
                c_args[i] = arg.as_ptr() as *const c_char;
            }
            // SAFETY:
            // The arguments are null-terminated, and the array is null-terminated
            unsafe { execvp(c_args[0], c_args.as_ptr()) };
            Err(errno())
        }
        _ => Ok(()),
    }
}

fn sigpoweroff() {
    if spawn(&RC_PWROFF).is_err() {
        let prefix = b"syd-init: poweroff\0";
        // SAFETY: The argument is null-terminated.
        unsafe { perror(prefix.as_ptr() as *const c_char) };
    }
}

fn sigreboot() {
    if spawn(&RC_REBOOT).is_err() {
        let prefix = b"syd-init: reboot\0";
        // SAFETY: The argument is null-terminated.
        unsafe { perror(prefix.as_ptr() as *const c_char) };
    }
}

fn main() -> ExitCode {
    // SAFETY: In libc, we trust.
    let pid = unsafe { getpid() };
    if pid != 1 {
        eprintln!("syd-init: not running as init (PID {pid} != 1)!");
        return ExitCode::from(1);
    }

    let path = b"/\0";
    // SAFETY: path is a valid pointer to a null-terminated string.
    if unsafe { libc::chdir(path.as_ptr() as *const libc::c_char) } != 0 {
        eprintln!("syd-init: Changing directory to / failed: {}", errno());
        return ExitCode::from(2);
    }

    // SAFETY: Initialize the signal set to include all signals
    let sigset: sigset_t = unsafe {
        let mut uninit = core::mem::MaybeUninit::<sigset_t>::uninit();
        sigfillset(uninit.as_mut_ptr());
        uninit.assume_init()
    };

    // SAFETY: Block all signals
    if unsafe { sigprocmask(SIG_BLOCK, &sigset, core::ptr::null_mut()) } != 0 {
        eprintln!("syd-init: Failed to block signals: {}", errno());
        return ExitCode::from(3);
    }

    if let Err(error) = spawn(&RC_INIT) {
        eprintln!("syd-init: Executing init command failed: {error}");
        return ExitCode::from(4);
    }
    loop {
        let mut sig: c_int = 0;
        // SAFETY: In libc we trust.
        unsafe { sigwait(std::ptr::addr_of!(sigset) as *mut _, &mut sig) };
        match sig {
            SIGCHLD | SIGALRM => sigreap(),
            SIGINT => sigreboot(),
            SIGUSR1 => sigpoweroff(),
            SIGUSR2 => return ExitCode::from(0),
            _ => {}
        }
        // SAFETY: See the comment under sigreap function.
        unsafe {
            alarm(TIMEO);
        }
    }
}
