//
// SydB☮x: seccomp and landlock based application sandbox with support for namespaces
// src/fs.rs: Filesystem utilities
//
// Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
// Based in part upon uutils coreutils package's src/lib/features/fs.rs which is:
//   (c) Joseph Crail <jbcrail@gmail.com>
//   (c) Jian Zeng <anonymousknight96 AT gmail.com>
// Tests base based in part upon gnulib packages' tests/test-canonicalize.c which is:
//   (c) Free Software Foundation, Inc.
//
// SPDX-License-Identifier: GPL-3.0-or-later

//! Set of functions to manage files and symlinks

use std::{
    collections::{HashMap, VecDeque},
    ffi::{CStr, OsStr, OsString},
    fs,
    hash::Hash,
    os::unix::ffi::OsStrExt,
    path::{Component, Path, PathBuf},
};

use nix::{
    errno::Errno,
    fcntl::AtFlags,
    sys::stat::{fstatat, FileStat, Mode, SFlag},
    unistd::{faccessat, AccessFlags},
    NixPath,
};
use once_cell::sync::Lazy;
use regex::bytes::{Regex, RegexBuilder};
use serde::{ser::SerializeMap, Serialize};

static RE_PROC_FD: Lazy<Regex> = Lazy::new(|| {
    #[allow(clippy::disallowed_methods)]
    RegexBuilder::new(
        r"
        \A
        /proc/[1-9][0-9]*/
            (task/[1-9][0-9]*/)?
        fd/[0-9]+
        \z",
    )
    .ignore_whitespace(true)
    .build()
    .expect("Invalid proc fd regex, please file a bug!")
});

static RE_PROC_FD_PIPE: Lazy<Regex> = Lazy::new(|| {
    #[allow(clippy::disallowed_methods)]
    RegexBuilder::new(
        r"
        \A
        (
            ((pipe|socket):\[[0-9][0-9]*\]) |
            (anon_inode:[^:]+)
        )
        \z",
    )
    .ignore_whitespace(true)
    .build()
    .expect("Invalid proc fd pipe regex, please file a bug!")
});

// Wrapper around FileStat to implement Serialize
pub(crate) struct FileStatWrapper(pub(crate) FileStat);

impl Serialize for FileStatWrapper {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        let ftype = match SFlag::from_bits_truncate(self.0.st_mode) {
            SFlag::S_IFREG => Some("reg"),
            SFlag::S_IFDIR => Some("dir"),
            SFlag::S_IFLNK => Some("lnk"),
            SFlag::S_IFCHR => Some("chr"),
            SFlag::S_IFBLK => Some("blk"),
            SFlag::S_IFIFO => Some("fifo"),
            SFlag::S_IFSOCK => Some("sock"),
            _ => None,
        };
        let perm = format!("{:o}", Mode::from_bits_truncate(self.0.st_mode).bits());

        let mut map = serializer.serialize_map(Some(7))?;
        map.serialize_entry("dev", &self.0.st_dev)?;
        map.serialize_entry("ino", &self.0.st_ino)?;
        map.serialize_entry("type", &ftype)?;
        map.serialize_entry("perm", &perm)?;
        map.serialize_entry("uid", &self.0.st_uid)?;
        map.serialize_entry("gid", &self.0.st_gid)?;
        map.serialize_entry("size", &self.0.st_size)?;
        map.serialize_entry("mtime", &self.0.st_mtime)?;
        map.end()
    }
}

/// Information to uniquely identify a file
struct FileInformation(nix::sys::stat::FileStat);

impl FileInformation {
    /// Get information for a given path.
    ///
    /// If `path` points to a symlink, information about the link will be returned.
    pub fn from_path<P: AsRef<Path>>(path: P) -> Result<Self, Errno> {
        fstatat(
            nix::libc::AT_FDCWD,
            path.as_ref(),
            AtFlags::AT_SYMLINK_NOFOLLOW,
        )
        .map(Self)
    }
}

impl PartialEq for FileInformation {
    fn eq(&self, other: &Self) -> bool {
        self.0.st_dev == other.0.st_dev && self.0.st_ino == other.0.st_ino
    }
}

impl Eq for FileInformation {}

impl Hash for FileInformation {
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        self.0.st_dev.hash(state);
        self.0.st_ino.hash(state);
    }
}

/// Controls how symbolic links should be handled when canonicalizing a path.
#[derive(Clone, Copy, Debug, Eq, PartialEq)]
pub enum MissingHandling {
    /// Return an error if any part of the path is missing.
    Normal,

    /// Resolve symbolic links, ignoring errors on the final component.
    Existing,

    /// Resolve symbolic links, ignoring errors on the non-final components.
    Missing,
}

/// Checks if a given path ends with a dot component.
///
/// This function iterates through the bytes of the path from end to start,
/// and determines whether the last component before any slashes is a dot.
///
/// # Arguments
///
/// * `path`: A reference to a `Box<CStr>` representing the path.
///
/// # Returns
///
/// * `bool`: Returns `true` if the path ends with a dot component, otherwise `false`.
///
/// # Examples
///
/// ```ignore
/// use std::ffi::CString;
/// assert_eq!(path_ends_with_dot(&Box::new(CString::new("some/path/.").unwrap())), true);
/// assert_eq!(path_ends_with_dot(&Box::new(CString::new("some/path/foo").unwrap())), false);
/// assert_eq!(path_ends_with_dot(&Box::new(CString::new("some/path/./").unwrap())), true);
/// assert_eq!(path_ends_with_dot(&Box::new(CString::new("some/path/././././///").unwrap())), true);
/// ```
#[allow(clippy::arithmetic_side_effects)]
#[allow(clippy::if_same_then_else)]
pub(crate) fn path_ends_with_dot(path: &CStr) -> bool {
    let bytes = path.to_bytes();

    // Start from the end of the string and move backwards
    let mut index = bytes.len();
    if index == 0 {
        return false;
    }

    // Skip trailing slashes
    while index > 0 && bytes[index - 1] == b'/' {
        index -= 1;
    }

    // If the path is empty after removing trailing slashes, it does not end with a dot
    if index == 0 {
        return false;
    }

    // Check for '.' or '..'
    if bytes[index - 1] == b'.' {
        if index == 1 || bytes[index - 2] == b'/' {
            return true; // Matches '.' or '*/.'
        } else if index > 1 && bytes[index - 2] == b'.' && (index == 2 || bytes[index - 3] == b'/')
        {
            return true; // Matches '..' or '*/..'
        }
    }

    false
}

/// Normalize a path by removing relative information
/// For example, convert 'bar/../foo/bar.txt' => 'foo/bar.txt'
/// copied from `<https://github.com/rust-lang/cargo/blob/2e4cfc2b7d43328b207879228a2ca7d427d188bb/src/cargo/util/paths.rs#L65-L90>`
/// both projects are MIT `<https://github.com/rust-lang/cargo/blob/master/LICENSE-MIT>`
/// for std impl progress see rfc `<https://github.com/rust-lang/rfcs/issues/2208>`
/// Note, this has been adapted to keep double dots at root position,
/// such that ../././config.status returns ../config.status
pub(crate) fn normalize_path<P: AsRef<Path>>(path: P) -> PathBuf {
    let components = path.as_ref().components().peekable();
    let mut ret = PathBuf::new();
    let ends_with_slash = path.as_ref().as_os_str().as_bytes().last() == Some(&b'/');
    let mut seen_curdir = false;

    for component in components {
        match component {
            Component::Prefix(..) => unreachable!(),
            Component::RootDir => {
                ret.push(component.as_os_str());
            }
            Component::CurDir => {
                // ignore all single dots except one
                seen_curdir = true;
                continue;
            }
            Component::ParentDir => {
                match ret.components().next_back() {
                    Some(Component::Normal(_)) => {
                        // Pop only if there's a normal component to remove.
                        ret.pop();
                    }
                    Some(Component::RootDir) => {
                        // If we're at the root, keep the `..` intact.
                        ret.push("..");
                    }
                    Some(Component::ParentDir) | None => {
                        // If there's nothing to pop, or we already have ParentDir,
                        // just push another ParentDir.
                        ret.push("..");
                    }
                    _ => unreachable!(),
                }
            }
            Component::Normal(c) => {
                ret.push(c);
            }
        }
    }

    if ret.is_empty() && seen_curdir {
        ret = PathBuf::from(".");
    } else if ends_with_slash && ret.as_os_str().as_bytes().last() != Some(&b'/') {
        // If the original path ended with a separator, append one to the normalized path.
        ret.push("");
    }

    ret
}

fn resolve_symlink<P: AsRef<Path>>(path: P) -> Result<PathBuf, Errno> {
    fs::read_link(&path).map_err(|e| Errno::from_i32(e.raw_os_error().unwrap_or(nix::libc::EINVAL)))
}

enum OwningComponent {
    Prefix(OsString),
    RootDir,
    CurDir,
    ParentDir,
    Normal(OsString),
}

impl OwningComponent {
    fn as_os_str(&self) -> &OsStr {
        match self {
            Self::Prefix(s) => s.as_os_str(),
            Self::RootDir => Component::RootDir.as_os_str(),
            Self::CurDir => Component::CurDir.as_os_str(),
            Self::ParentDir => Component::ParentDir.as_os_str(),
            Self::Normal(s) => s.as_os_str(),
        }
    }
}

impl<'a> From<Component<'a>> for OwningComponent {
    fn from(comp: Component<'a>) -> Self {
        match comp {
            Component::Prefix(_) => Self::Prefix(comp.as_os_str().to_os_string()),
            Component::RootDir => Self::RootDir,
            Component::CurDir => Self::CurDir,
            Component::ParentDir => Self::ParentDir,
            Component::Normal(s) => Self::Normal(s.to_os_string()),
        }
    }
}

/// Return the canonical, absolute form of a path
///
/// This function is a generalization of [`std::fs::canonicalize`] that
/// allows controlling how symbolic links are resolved and how to deal
/// with missing components. It returns the canonical, absolute form of
/// a path.
///
/// The `resolve_symlinks` is a boolean parameter which controls how symbolic links are resolved.
///
/// The `miss_mode` parameter controls how missing path elements are handled
///
/// * [`MissingHandling::Normal`] makes this function behave like
///   [`std::fs::canonicalize`], resolving symbolic links and returning
///   an error if the path does not exist.
/// * [`MissingHandling::Missing`] makes this function ignore non-final
///   components of the path that could not be resolved.
/// * [`MissingHandling::Existing`] makes this function return an error
///   if the final component of the path does not exist.
#[allow(clippy::cognitive_complexity)]
pub fn canonicalize<P: AsRef<Path>>(
    original: P,
    mut resolve_symlinks: bool,
    miss_mode: MissingHandling,
) -> Result<PathBuf, Errno> {
    const MAX_SYMLINKS: u8 = 32;
    let original = original.as_ref();
    let original = if original.is_empty() {
        return Err(Errno::ENOENT);
    } else if original.is_absolute() {
        original.to_path_buf()
    } else {
        // relative path passed to canonicalize is not supported.
        return Err(Errno::EINVAL);
    };
    let path = normalize_path(original);
    let has_to_be_directory = path.as_os_str().as_bytes().last() == Some(&(b'/'));
    resolve_symlinks = resolve_symlinks || has_to_be_directory;
    let mut parts: VecDeque<OwningComponent> = path.components().map(|part| part.into()).collect();
    let mut result = PathBuf::new();
    let mut visited_files = HashMap::new();
    while let Some(part) = parts.pop_front() {
        match part {
            OwningComponent::Prefix(s) => {
                result.push(s);
                continue;
            }
            OwningComponent::RootDir | OwningComponent::Normal(..) => {
                result.push(part.as_os_str());
            }
            OwningComponent::CurDir => {}
            OwningComponent::ParentDir => {
                result.pop();
            }
        }
        let should_resolve = if parts.is_empty() {
            resolve_symlinks
        } else {
            true
        };
        if !should_resolve {
            continue;
        }

        match FileInformation::from_path(&result) {
            Ok(file_info) => {
                let is_symlink = SFlag::from_bits_truncate(file_info.0.st_mode) == SFlag::S_IFLNK;
                #[allow(clippy::arithmetic_side_effects)]
                if !is_symlink {
                    // Not a symbolic link, no need to resolve.
                    continue;
                } else {
                    let count = visited_files.entry(file_info).or_insert(0);
                    *count += 1; // SAFETY: MAX_SYMLINKS < u8::MAX
                    if *count >= MAX_SYMLINKS {
                        if has_to_be_directory
                            || miss_mode != MissingHandling::Missing
                            || !parts.is_empty()
                        {
                            return Err(Errno::ELOOP);
                        }

                        // From this point on, stop
                        // resolving symlinks and
                        // just accumulate the path.
                        resolve_symlinks = false;
                        continue;
                    }
                }
            }
            Err(error) => {
                if miss_mode == MissingHandling::Existing
                    || (miss_mode == MissingHandling::Normal && !parts.is_empty())
                {
                    return Err(error);
                }
                continue;
            }
        };

        match resolve_symlink(&result) {
            Ok(link_path) => {
                if RE_PROC_FD.is_match(result.as_os_str().as_bytes())
                    && RE_PROC_FD_PIPE.is_match(link_path.as_os_str().as_bytes())
                {
                    continue;
                }
                for link_part in link_path.components().rev() {
                    parts.push_front(link_part.into());
                }
                result.pop();
            }
            Err(error) => {
                if miss_mode == MissingHandling::Existing
                    || (miss_mode == MissingHandling::Normal && !parts.is_empty())
                {
                    return Err(error);
                }
            }
        }
    }

    // raise Not a directory if required
    match miss_mode {
        MissingHandling::Existing => {
            if has_to_be_directory && !result.is_dir() {
                return Err(Errno::ENOTDIR);
            }
        }
        MissingHandling::Normal => {
            if result.exists() {
                if has_to_be_directory && !result.is_dir() {
                    return Err(Errno::ENOTDIR);
                }
            } else if let Some(parent) = result.parent() {
                if !parent.exists() {
                    return Err(Errno::ENOENT);
                } else if !parent.is_dir() {
                    return Err(Errno::ENOTDIR);
                }
            }
        }
        MissingHandling::Missing => {
            // SAFETY: we use faccessat rather than `result.exists()`
            // because the latter may call stat and update access time.
            // Note: `result.is_dir()` updates access time too.
            if faccessat(
                None,
                &result,
                AccessFlags::F_OK,
                AtFlags::AT_SYMLINK_NOFOLLOW,
            )
            .is_ok()
                && has_to_be_directory
                && !result.is_dir()
            {
                return Err(Errno::ENOTDIR);
            }
        }
    }

    // Preserve trailing slash as necessary
    if has_to_be_directory && result.as_os_str().as_bytes().last() != Some(&(b'/')) {
        result.push("");
    }

    Ok(result)
}

#[cfg(test)]
mod tests {
    use std::{
        ffi::CString,
        fs::OpenOptions,
        os::unix::fs::{symlink, OpenOptionsExt},
        process::Command,
        thread::sleep,
        time::{Duration, SystemTime},
    };

    use nix::unistd::getpid;

    use super::{MissingHandling::*, *};

    type TestResult = Result<(), Box<dyn std::error::Error>>;

    // A helper function to get the current atime of a file
    fn get_atime<P: AsRef<Path>>(path: P) -> SystemTime {
        let metadata = fs::metadata(path).expect("Failed to get metadata");
        metadata.accessed().expect("Failed to get accessed time")
    }

    // Helper function to assert that the atime of a file or directory has not changed
    fn assert_atime_unchanged<P: AsRef<Path>, F>(path: P, func: F)
    where
        F: FnOnce() -> Result<PathBuf, Errno>,
    {
        let original_atime_f = get_atime(&path);
        let original_atime_p = get_atime(path.as_ref().parent().unwrap());
        sleep(Duration::from_secs(7));
        assert!(
            func().is_ok(),
            "canonicalize {} failed",
            path.as_ref().display()
        );
        // We allow a 1-second tolerance since some filesystems do not have nanosecond precision.
        let new_atime_f = get_atime(&path);
        let new_atime_p = get_atime(path.as_ref().parent().unwrap());
        assert!(
            new_atime_f <= original_atime_f + Duration::new(1, 0),
            "The atime of the file should not have significantly changed."
        );
        assert!(
            new_atime_p <= original_atime_p + Duration::new(1, 0),
            "The atime of the parent dir should not have significantly changed."
        );
    }

    // std::fs::remove_dir_all stops on the first error.
    // we need something more forceful.
    fn remove_dir_all<P: AsRef<Path>>(path: P) -> std::io::Result<()> {
        let status = Command::new("rm")
            .arg("-rf")
            .arg(path.as_ref().to_string_lossy().to_string())
            .status()?;
        if status.success() {
            Ok(())
        } else {
            Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                format!("Failed to remove directory: {}", path.as_ref().display()),
            ))
        }
    }

    // Helper function to create a symlink chain that eventually loops back to an earlier link
    fn setup_symlink_loop(tmp_dir: &tempfile::TempDir, links: &[(&str, &str)]) {
        for &(src, dst) in links {
            let src_path = tmp_dir.path().join(src);

            // Check and create parent directory for the source if necessary
            if let Some(parent) = src_path.parent() {
                if !parent.exists() {
                    fs::create_dir_all(parent).unwrap();
                }
            }

            // The destination is given relative to the source
            let dst_path = Path::new(dst);

            // Avoid creating a symlink if the source already exists
            if src_path.exists() {
                fs::remove_file(&src_path).unwrap();
            }

            // If the destination is an absolute path or starts with '/', we strip the '/' and prefix with tmp_dir
            let full_dst_path = if dst_path.is_absolute() {
                tmp_dir.path().join(dst_path.strip_prefix("/").unwrap())
            } else {
                src_path.parent().unwrap().join(dst_path)
            };

            // Create parent directories for the destination if they don't exist
            if let Some(parent) = full_dst_path.parent() {
                if !parent.exists() {
                    fs::create_dir_all(parent).unwrap();
                }
            }

            // Create the symlink
            symlink(&full_dst_path, &src_path).expect(&format!(
                "Unable to symlink {} -> {}",
                src_path.display(),
                full_dst_path.display()
            ));
        }
    }

    struct NormalizePathTestCase<'a> {
        path: &'a str,
        test: &'a str,
    }

    const NORMALIZE_PATH_TESTS: [NormalizePathTestCase; 38] = [
        NormalizePathTestCase {
            path: "usr",
            test: "usr",
        },
        NormalizePathTestCase {
            path: "usr/",
            test: "usr/",
        },
        NormalizePathTestCase {
            path: "usr//",
            test: "usr/",
        },
        NormalizePathTestCase {
            path: "./././././././usr",
            test: "usr",
        },
        NormalizePathTestCase {
            path: "./././.././././usr",
            test: "../usr",
        },
        NormalizePathTestCase {
            path: "./foo/bar.txt",
            test: "foo/bar.txt",
        },
        NormalizePathTestCase {
            path: "bar/../foo/bar.txt",
            test: "foo/bar.txt",
        },
        NormalizePathTestCase {
            path: "foo///bar.txt",
            test: "foo/bar.txt",
        },
        NormalizePathTestCase {
            path: "foo///bar",
            test: "foo/bar",
        },
        NormalizePathTestCase {
            path: "foo//./bar",
            test: "foo/bar",
        },
        NormalizePathTestCase {
            path: "/foo//./bar",
            test: "/foo/bar",
        },
        NormalizePathTestCase {
            path: "../config.status",
            test: "../config.status",
        },
        NormalizePathTestCase {
            path: ".././../config.status",
            test: "../../config.status",
        },
        NormalizePathTestCase {
            path: ".././.././././../config.status",
            test: "../../../config.status",
        },
        // Relative Path with Parent Directory References
        NormalizePathTestCase {
            path: "a/../b/c",
            test: "b/c",
        },
        // Path with Current Directory References
        NormalizePathTestCase {
            path: "a/./b/./c",
            test: "a/b/c",
        },
        // Path with Multiple Sequential Slashes
        NormalizePathTestCase {
            path: "a///b/c",
            test: "a/b/c",
        },
        // Path with Trailing Slashes
        NormalizePathTestCase {
            path: "a/b/c/",
            test: "a/b/c/",
        },
        // Path with Leading Slashes
        NormalizePathTestCase {
            path: "/a/b/c",
            test: "/a/b/c",
        },
        // Path with "../" at Root
        NormalizePathTestCase {
            path: "/../a/b/c",
            test: "/../a/b/c",
        },
        // Path with "../" at Root (no normalization)
        NormalizePathTestCase {
            path: "/../../a/b/c",
            test: "/../../a/b/c",
        },
        // Path with Leading and Trailing "../" (at Root)
        NormalizePathTestCase {
            path: "/../a/../b/../c/",
            test: "/../c/",
        },
        // Path with Leading and Trailing "../" (not at Root)
        NormalizePathTestCase {
            path: "x/y/z/../../a/b/../../c",
            test: "x/c",
        },
        // Path with "../" after "../" (at Root)
        NormalizePathTestCase {
            path: "/a/b/c/../../../x/y/z",
            test: "/x/y/z",
        },
        // Path with "../" after "../" (not at Root)
        NormalizePathTestCase {
            path: "a/b/c/../../../x/y/z",
            test: "x/y/z",
        },
        // Path with "./" at Root
        NormalizePathTestCase {
            path: "/./a/b/c",
            test: "/a/b/c",
        },
        // Path with "./" (no normalization)
        NormalizePathTestCase {
            path: "/././././a/b/c",
            test: "/a/b/c",
        },
        // Path with Leading and Trailing "./"
        NormalizePathTestCase {
            path: "./x/./y/./z/./",
            test: "x/y/z/",
        },
        // Path with Leading "./" (at Root)
        NormalizePathTestCase {
            path: "/./x/y/z",
            test: "/x/y/z",
        },
        // Path with Leading "./" (not at Root)
        NormalizePathTestCase {
            path: "a/./b/./c",
            test: "a/b/c",
        },
        // A single .. is preserved as is.
        NormalizePathTestCase {
            path: "..",
            test: "..",
        },
        // A single .. is preserved as is no matter the slashes.
        NormalizePathTestCase {
            path: "../././////.///.////",
            test: "../",
        },
        // A single . is preserved as is.
        NormalizePathTestCase {
            path: ".",
            test: ".",
        },
        // A single . is preserved as is no matter the slashes.
        NormalizePathTestCase {
            path: "./././////.///.////",
            test: ".",
        },
        // Multiple .. at start are preserved as is
        NormalizePathTestCase {
            path: "../../foo",
            test: "../../foo",
        },
        // Multiple .. are preserved as needed
        NormalizePathTestCase {
            path: "a/../../foo",
            test: "../foo",
        },
        // Multiple .. are preserved as needed (take 2)
        NormalizePathTestCase {
            path: "a/./././/../../foo",
            test: "../foo",
        },
        // Multiple .. are preserved as needed (take 3)
        NormalizePathTestCase {
            path: "a/./././b/./././///////..////././/..////./foo",
            test: "foo",
        },
    ];

    struct EndsWithDotTestCase<'a> {
        path: &'a str,
        test: bool,
    }

    const ENDS_WITH_DOT_TESTS: [EndsWithDotTestCase; 10] = [
        EndsWithDotTestCase {
            path: ".",
            test: true,
        },
        EndsWithDotTestCase {
            path: "..",
            test: true,
        },
        EndsWithDotTestCase {
            path: "...",
            test: false,
        },
        EndsWithDotTestCase {
            path: "/.",
            test: true,
        },
        EndsWithDotTestCase {
            path: "/..",
            test: true,
        },
        EndsWithDotTestCase {
            path: "/...",
            test: false,
        },
        EndsWithDotTestCase {
            path: "foo.",
            test: false,
        },
        EndsWithDotTestCase {
            path: "foo./.",
            test: true,
        },
        EndsWithDotTestCase {
            path: "foo/./././/./",
            test: true,
        },
        EndsWithDotTestCase {
            path: "conftest.dir/././././////",
            test: true,
        },
    ];

    fn tempdir() -> Result<PathBuf, Box<dyn std::error::Error>> {
        let path = Path::new(".syd-test");
        std::fs::create_dir_all(path)?;
        let _ = OpenOptions::new()
            .write(true)
            .create(true)
            .mode(0o600)
            .open(path.join("test"))?;
        Ok(path.to_path_buf())
    }

    #[test]
    fn test_normalize_path() {
        for (idx, test) in NORMALIZE_PATH_TESTS.iter().enumerate() {
            let path = Path::new(test.path);
            let normalized = normalize_path(path);
            let normalized = normalized.to_str().expect("Path is not valid utf-8!");
            assert_eq!(
                test.test, normalized,
                "NormalizePathTestCase {} -> \"{}\": {} != {}",
                idx, test.path, test.test, normalized
            );
        }
    }

    #[test]
    fn test_ends_with_dot() {
        for (idx, test) in ENDS_WITH_DOT_TESTS.iter().enumerate() {
            let path = CString::new(test.path).unwrap();
            let ends = path_ends_with_dot(&path);
            assert_eq!(
                test.test, ends,
                "EndsWithDotTestCase {} -> \"{}\": {} != {}",
                idx, test.path, test.test, ends
            );
        }
    }

    #[test]
    fn test_canonicalize_empty_path() -> TestResult {
        assert_eq!(canonicalize("", true, Normal), Err(Errno::ENOENT));
        assert_eq!(canonicalize("", true, Existing), Err(Errno::ENOENT));
        assert_eq!(canonicalize("", true, Missing), Err(Errno::ENOENT));
        assert_eq!(canonicalize("", false, Normal), Err(Errno::ENOENT));
        assert_eq!(canonicalize("", false, Existing), Err(Errno::ENOENT));
        assert_eq!(canonicalize("", false, Missing), Err(Errno::ENOENT));

        Ok(())
    }

    #[test]
    fn test_canonicalize_repetitive_slashes() -> TestResult {
        let result_test = canonicalize("/etc/passwd", true, Normal)?;
        let paths = vec![
            "/etc/passwd",
            "/etc//passwd",
            "/etc///passwd",
            "//etc/passwd",
            "//etc//passwd",
            "//etc///passwd",
            "///etc/passwd",
            "///etc//passwd",
            "///etc///passwd",
        ];
        for path in &paths {
            let result = canonicalize(path, true, Missing)?;
            assert_eq!(result, result_test);
        }

        Ok(())
    }

    #[test]
    fn test_canonicalize_dots_slashes() -> TestResult {
        let base = tempdir()?;

        let cwd = std::env::current_dir()?.display().to_string();
        let path = base.display().to_string();

        let result1 = canonicalize(format!("{cwd}/{path}//./..//{path}/test"), true, Normal)?
            .display()
            .to_string();
        let result2 = canonicalize(format!("{cwd}/{path}//./..//{path}/test"), true, Existing)?
            .display()
            .to_string();

        assert!(!result1.is_empty(), "result:{result1}");
        assert!(!result2.is_empty(), "result:{result2}");
        assert_eq!(result1, result2);

        Ok(())
    }

    #[test]
    fn test_canonicalize_non_directory_with_slash() -> TestResult {
        let cwd = std::env::current_dir()?.display().to_string();
        let path = tempdir()?.display().to_string();
        let test = format!("{cwd}/{path}/test/");

        assert_eq!(canonicalize(&test, true, Normal), Err(Errno::ENOTDIR));
        assert_eq!(canonicalize(&test, true, Existing), Err(Errno::ENOTDIR));
        assert_eq!(canonicalize(&test, true, Missing), Err(Errno::ENOTDIR));
        assert_eq!(canonicalize(&test, false, Normal), Err(Errno::ENOTDIR));
        assert_eq!(canonicalize(&test, false, Existing), Err(Errno::ENOTDIR));
        assert_eq!(canonicalize(&test, false, Missing), Err(Errno::ENOTDIR));

        Ok(())
    }

    /// FIXME: The asserts return success rather than failure.
    /// Bug or feature?
    #[test]
    #[ignore]
    fn test_canonicalize_missing_directory_returns_enoent() -> TestResult {
        assert_eq!(canonicalize("/zzz/..", true, Normal), Err(Errno::ENOENT));
        assert_eq!(canonicalize("/zzz/..", true, Existing), Err(Errno::ENOENT));
        assert_eq!(canonicalize("/zzz/..", false, Normal), Err(Errno::ENOENT));
        assert_eq!(canonicalize("/zzz/..", false, Existing), Err(Errno::ENOENT));

        Ok(())
    }

    #[test]
    fn test_relative_symlink_resolution() -> TestResult {
        // Setup
        let root_test_dir = Path::new("test_root_relative_symlink_resolution");
        let deep_dir = root_test_dir.join("a/b/c");
        let _ = remove_dir_all(&root_test_dir);
        fs::create_dir_all(&root_test_dir.join("d"))?;
        fs::create_dir_all(&deep_dir)?;

        // Create a symlink in "b" that points upwards to "a"
        let rel_link = root_test_dir.join("a/b/rel_link");
        symlink("../..", &rel_link)?;

        // Append /proc/self/cwd to get an absolute path to our symlinked path
        let abs_link_path = Path::new("/proc/self/cwd").join(root_test_dir.join("a/b/rel_link/d"));

        // Call canonicalize
        let result = canonicalize(&abs_link_path, true, MissingHandling::Existing);
        assert!(
            matches!(result, Ok(_)),
            "canonicalize:{} result:{:?}",
            abs_link_path.display(),
            result
        );
        let resolved_path = result.unwrap();

        // We expect the path to be resolved to "test_root/a/d", but we need to canonicalize it
        let expected_path =
            fs::canonicalize(Path::new("/proc/self/cwd").join(root_test_dir.join("d")))?;

        // Cleanup
        let _ = remove_dir_all(&root_test_dir);

        assert_eq!(resolved_path, expected_path);

        Ok(())
    }

    #[test]
    fn test_complex_interplay_symlinks_dots() -> TestResult {
        // Setup
        let cwd = Path::new("/proc/self/cwd").canonicalize()?;
        let root_test_dir = cwd.join("test_root_complex_interplay_symlinks_dots");
        let _ = remove_dir_all(&root_test_dir);
        fs::create_dir_all(root_test_dir.join("a/b/c")).unwrap();
        fs::create_dir(root_test_dir.join("d")).unwrap();
        fs::create_dir(root_test_dir.join("e")).unwrap();
        fs::create_dir(root_test_dir.join("x")).unwrap();

        // Create several symlinks
        symlink("./a", root_test_dir.join("link_to_a")).unwrap();
        symlink("e", root_test_dir.join("link_to_e")).unwrap();
        symlink("a/b", root_test_dir.join("link_to_b")).unwrap();
        symlink("../../x", root_test_dir.join("a/b/rel_link")).unwrap();

        let path = root_test_dir.join("link_to_a/../link_to_b/rel_link/../..");
        let resolved_path = canonicalize(&path, true, MissingHandling::Existing).unwrap();

        // Cleanup
        let _ = remove_dir_all(&root_test_dir);

        // Assertion
        assert_eq!(resolved_path, root_test_dir);

        Ok(())
    }

    #[test]
    fn test_trailing_slash_handling() -> TestResult {
        let path = Path::new("/usr/");
        let pabs = canonicalize(&path, true, MissingHandling::Missing).unwrap();
        assert_eq!(path, pabs);

        let path = Path::new("/proc/self/");
        let pexp = format!("/proc/{}/", getpid());
        let pexp = Path::new(&pexp);
        let pabs = canonicalize(&path, true, MissingHandling::Missing).unwrap();
        assert_eq!(pabs, pexp);
        let pabs = canonicalize(&path, false, MissingHandling::Missing).unwrap();
        assert_eq!(pabs, pexp);

        Ok(())
    }

    #[ignore]
    #[test]
    fn test_canonicalize_no_atime_change_normal() -> TestResult {
        let cdir = std::env::current_dir()?;
        let base = cdir.join(tempdir()?);
        let path = base.join("file");
        fs::File::create(&path)?;

        assert_atime_unchanged(&path, || canonicalize(&path, true, MissingHandling::Normal));

        let _ = remove_dir_all(&base);
        Ok(())
    }

    #[ignore]
    #[test]
    fn test_canonicalize_no_atime_change_existing() -> TestResult {
        let cdir = std::env::current_dir()?;
        let base = cdir.join(tempdir()?);
        let path = base.join("file");
        fs::File::create(&path)?;

        assert_atime_unchanged(&path, || {
            canonicalize(&path, true, MissingHandling::Existing)
        });

        let _ = remove_dir_all(&base);
        Ok(())
    }

    #[test]
    fn test_canonicalize_symlink_loop() {
        let tmp_dir = tempfile::tempdir().expect("Failed to create temp dir");
        let dir_path = tmp_dir.path();

        // Create a symlink loop: link_a -> link_b -> link_a
        let mut link_a = dir_path.join("link_a");
        let mut link_b = dir_path.join("link_b");
        symlink(&link_b, &link_a).expect("Failed to create symlink a");
        symlink(&link_a, &link_b).expect("Failed to create symlink b");

        // Now check that canonicalize detects the loop correctly
        let result = canonicalize(&link_a, false, MissingHandling::Normal);
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(&link_a, true, MissingHandling::Normal);
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(&link_a, false, MissingHandling::Existing);
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(&link_a, true, MissingHandling::Existing);
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(&link_a, false, MissingHandling::Missing);
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(&link_a, true, MissingHandling::Missing);
        assert!(result.is_ok(), "{result:?}");

        // Add a trailing slash and retest.
        link_a.push("");

        let result = canonicalize(&link_a, false, MissingHandling::Missing);
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(&link_a, true, MissingHandling::Missing);
        assert_eq!(result, Err(Errno::ELOOP));

        // Add a trailing slash and retest.
        link_b.push("");

        let result = canonicalize(&link_b, false, MissingHandling::Missing);
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(&link_b, true, MissingHandling::Missing);
        assert_eq!(result, Err(Errno::ELOOP));
    }

    #[test]
    fn test_canonicalize_nonexistent_final_component() {
        let tmp_dir = tempfile::tempdir().expect("Failed to create temp dir");
        let dir_path = tmp_dir.path();

        // Create a valid symlink to a non-existent final component
        let mut valid_link = dir_path.join("valid_link");
        let nonexistent_target = dir_path.join("nonexistent");
        symlink(&nonexistent_target, &valid_link)
            .expect("Failed to create symlink to non-existent target");

        // Now check that canonicalize handles the non-existent final component correctly
        let result = canonicalize(&valid_link, false, MissingHandling::Normal);
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(&valid_link, true, MissingHandling::Normal);
        // FIXME: assert_eq!(result, Err(Errno::ENOENT));
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(&valid_link, false, MissingHandling::Existing);
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(&valid_link, true, MissingHandling::Existing);
        assert_eq!(result, Err(Errno::ENOENT));

        let result = canonicalize(&valid_link, false, MissingHandling::Missing);
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(&valid_link, true, MissingHandling::Missing);
        assert!(result.is_ok(), "{result:?}");

        // Add a trailing slash and retest.
        valid_link.push("");

        let result = canonicalize(&valid_link, false, MissingHandling::Missing);
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(&valid_link, true, MissingHandling::Missing);
        assert!(result.is_ok(), "{result:?}");
    }

    #[test]
    fn test_canonicalize_self_referential_symlink() {
        let tmp_dir = tempfile::tempdir().expect("Failed to create temp dir");
        let mut symlink_path = tmp_dir.path().join("self_link");
        symlink(&symlink_path, &symlink_path).expect("Failed to create self-referential symlink");

        let result = canonicalize(&symlink_path, false, MissingHandling::Normal);
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(&symlink_path, true, MissingHandling::Normal);
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(&symlink_path, false, MissingHandling::Existing);
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(&symlink_path, true, MissingHandling::Existing);
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(&symlink_path, false, MissingHandling::Missing);
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(&symlink_path, true, MissingHandling::Missing);
        assert!(result.is_ok(), "{result:?}");

        // Add a trailing slash and retest.
        symlink_path.push("");

        let result = canonicalize(&symlink_path, false, MissingHandling::Missing);
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(&symlink_path, true, MissingHandling::Missing);
        assert_eq!(result, Err(Errno::ELOOP));
    }

    #[test]
    fn test_canonicalize_broken_symlink() {
        let tmp_dir = tempfile::tempdir().expect("Failed to create temp dir");
        let mut broken_link = tmp_dir.path().join("broken_link");
        let nonexistent_target = tmp_dir.path().join("nonexistent_target");
        symlink(&nonexistent_target, &broken_link).expect("Failed to create broken symlink");

        let result = canonicalize(&broken_link, false, MissingHandling::Normal);
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(&broken_link, true, MissingHandling::Normal);
        // FIXME: assert_eq!(result, Err(Errno::ENOENT));
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(&broken_link, false, MissingHandling::Existing);
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(&broken_link, true, MissingHandling::Existing);
        assert_eq!(result, Err(Errno::ENOENT));

        let result = canonicalize(&broken_link, false, MissingHandling::Missing);
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(&broken_link, true, MissingHandling::Missing);
        assert!(result.is_ok(), "{result:?}");

        // Add a trailing slash and retest.
        broken_link.push("");

        let result = canonicalize(&broken_link, false, MissingHandling::Missing);
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(&broken_link, true, MissingHandling::Missing);
        assert!(result.is_ok(), "{result:?}");
    }

    #[test]
    fn test_canonicalize_symlink_to_directory() {
        let tmp_dir = tempfile::tempdir().expect("Failed to create temp dir");
        let dir = tmp_dir.path().join("dir");
        fs::create_dir(&dir).expect("Failed to create directory");

        let symlink_path = tmp_dir.path().join("dir_link");
        symlink(&dir, &symlink_path).expect("Failed to create symlink to directory");

        let result = canonicalize(&symlink_path, false, MissingHandling::Normal);
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_symlink(), "{result:?}");

        let result = canonicalize(&symlink_path, true, MissingHandling::Normal);
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_dir(), "{result:?}");

        let result = canonicalize(&symlink_path, false, MissingHandling::Existing);
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_symlink(), "{result:?}");

        let result = canonicalize(&symlink_path, true, MissingHandling::Existing);
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_dir(), "{result:?}");

        let result = canonicalize(&symlink_path, false, MissingHandling::Missing);
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_symlink(), "{result:?}");

        let result = canonicalize(&symlink_path, true, MissingHandling::Missing);
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_dir(), "{result:?}");
    }

    #[test]
    fn test_canonicalize_symlink_chain() {
        let tmp_dir = tempfile::tempdir().expect("Failed to create temp dir");
        let link1 = tmp_dir.path().join("link1");
        let link2 = tmp_dir.path().join("link2");
        let link3 = tmp_dir.path().join("link3");
        let file = tmp_dir.path().join("file");
        fs::write(&file, "content").expect("Failed to write file");

        // Create a chain of symlinks: link1 -> link2 -> link3 -> file
        symlink(&link2, &link1).expect("Failed to create link1");
        symlink(&link3, &link2).expect("Failed to create link2");
        symlink(&file, &link3).expect("Failed to create link3");

        let result = canonicalize(&link1, false, MissingHandling::Normal);
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_symlink(), "{result:?}");

        let result = canonicalize(&link1, true, MissingHandling::Normal);
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_file(), "{result:?}");

        let result = canonicalize(&link1, false, MissingHandling::Existing);
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_symlink(), "{result:?}");

        let result = canonicalize(&link1, true, MissingHandling::Existing);
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_file(), "{result:?}");

        let result = canonicalize(&link1, false, MissingHandling::Missing);
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_symlink(), "{result:?}");

        let result = canonicalize(&link1, true, MissingHandling::Missing);
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_file(), "{result:?}");
    }

    #[test]
    fn test_canonicalize_complex_symlink_loop_with_intermediate_components() {
        let tmp_dir = tempfile::tempdir().expect("Failed to create temp dir");

        // Setting up a complex symlink scenario
        setup_symlink_loop(
            &tmp_dir,
            &[("a", "b/c"), ("b/c", "d"), ("b/d", "../e"), ("e", "f/../a")],
        );

        let mut path = tmp_dir.path().join("a");

        let result = canonicalize(&path, false, MissingHandling::Normal);
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_symlink(), "{result:?}");

        let result = canonicalize(&path, true, MissingHandling::Normal);
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(&path, false, MissingHandling::Existing);
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_symlink(), "{result:?}");

        let result = canonicalize(&path, true, MissingHandling::Existing);
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(&path, false, MissingHandling::Missing);
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_symlink(), "{result:?}");

        let result = canonicalize(&path, true, MissingHandling::Missing);
        assert!(result.is_ok(), "{result:?}");
        assert!(!result.clone().unwrap().exists(), "{result:?}");

        // Add a trailing slash and retest.
        path.push("");

        let result = canonicalize(&path, false, MissingHandling::Missing);
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(&path, true, MissingHandling::Missing);
        assert_eq!(result, Err(Errno::ELOOP));

        // Add a final component and retest.
        path.push("foo");

        let result = canonicalize(&path, false, MissingHandling::Missing);
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(&path, true, MissingHandling::Missing);
        assert_eq!(result, Err(Errno::ELOOP));
    }

    #[test]
    fn test_canonicalize_symlinks_with_dot_and_dotdot_components() {
        let tmp_dir = tempfile::tempdir().expect("Failed to create temp dir");

        // Create a complex directory structure with dots and symlinks
        fs::create_dir_all(tmp_dir.path().join("b")).expect("Failed to create directory b");
        symlink("b", tmp_dir.path().join("a")).expect("Failed to create symlink a -> b");
        symlink("..///e", tmp_dir.path().join("b").join("d"))
            .expect("Failed to create symlink b/d -> ../e");
        symlink("b/.///./d", tmp_dir.path().join("e")).expect("Failed to create symlink e -> b/d");

        let mut path = tmp_dir.path().join("a").join(".").join("d");

        let result = canonicalize(&path, false, MissingHandling::Normal);
        assert!(result.is_ok(), "{result:?}");
        assert!(!result.clone().unwrap().exists(), "{result:?}");

        let result = canonicalize(&path, true, MissingHandling::Normal);
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(&path, false, MissingHandling::Existing);
        assert!(result.is_ok(), "{result:?}");
        assert!(!result.clone().unwrap().exists(), "{result:?}");

        let result = canonicalize(&path, true, MissingHandling::Existing);
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(&path, false, MissingHandling::Missing);
        assert!(result.is_ok(), "{result:?}");
        assert!(!result.clone().unwrap().exists(), "{result:?}");

        let result = canonicalize(&path, true, MissingHandling::Missing);
        assert!(result.is_ok(), "{result:?}");
        assert!(!result.clone().unwrap().exists(), "{result:?}");

        // Add a trailing slash and retest.
        path.push("");

        let result = canonicalize(&path, false, MissingHandling::Missing);
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(&path, true, MissingHandling::Missing);
        assert_eq!(result, Err(Errno::ELOOP));

        // Add a final component and retest.
        path.push("foo");

        let result = canonicalize(&path, false, MissingHandling::Missing);
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(&path, true, MissingHandling::Missing);
        assert_eq!(result, Err(Errno::ELOOP));
    }

    #[test]
    fn test_canonicalize_proc_self() -> TestResult {
        let path = canonicalize("/proc/self", false, MissingHandling::Missing)?;
        assert_eq!(path, PathBuf::from("/proc/self"));

        Ok(())
    }

    #[test]
    fn test_glob_proc_fd() -> TestResult {
        assert!(RE_PROC_FD.is_match(b"/proc/1/fd/0"));
        assert!(RE_PROC_FD.is_match(b"/proc/1/fd/1"));
        assert!(RE_PROC_FD.is_match(b"/proc/1234/task/123/fd/0"));
        assert!(RE_PROC_FD.is_match(b"/proc/12345/task/12345/fd/1"));

        assert!(!RE_PROC_FD.is_match(b"/proc/1/fd"));
        assert!(!RE_PROC_FD.is_match(b"/proc/1/fd/n"));
        assert!(!RE_PROC_FD.is_match(b"/proc/1234/task/123/f/0"));
        assert!(!RE_PROC_FD.is_match(b"/proc/12345/12345/fd/1"));

        assert!(RE_PROC_FD_PIPE.is_match(b"pipe:[12345]"));
        assert!(RE_PROC_FD_PIPE.is_match(b"socket:[12345]"));
        assert!(RE_PROC_FD_PIPE.is_match(b"anon_inode:seccomp notify"));
        assert!(RE_PROC_FD_PIPE.is_match(b"anon_inode:[eventpoll]"));

        assert!(!RE_PROC_FD_PIPE.is_match(b"pipe:[n]"));
        assert!(!RE_PROC_FD_PIPE.is_match(b"socket:[12345"));
        assert!(!RE_PROC_FD_PIPE.is_match(b"anon_inode:seccomp:notify"));
        assert!(!RE_PROC_FD_PIPE.is_match(b"anon_inode:foo:bar"));

        Ok(())
    }
}
