//
// SydB☮x: seccomp and landlock based application sandbox with support for namespaces
// src/lib.rs: Common utility functions
//
// Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

//! # SydB☮x: The ☮ther SⒶndbøx
//!
//! [![Shine On You Crazy Diamond!](https://img.shields.io/badge/Shine%20On%20You%20Crazy%20Diamond!-8A2BE2)](https://en.wikipedia.org/wiki/Syd_Barrett)
//! [![license](https://img.shields.io/crates/l/jja.svg)](https://git.sr.ht/~alip/syd/tree/main/item/COPYING)
//! [![msrv](https://img.shields.io/badge/rustc-1.70%2B-green?style=plastic)](https://blog.rust-lang.org/2023/06/01/Rust-1.70.0.html)
//! [![build status](https://builds.sr.ht/~alip/syd.svg)](https://builds.sr.ht/~alip/syd?)
//! [![maintenance-status](https://img.shields.io/badge/maintenance-actively--developed-brightgreen.svg)](https://git.sr.ht/~alip/syd)
//! [![dependency status](https://deps.rs/repo/sourcehut/~alip/syd/status.svg)](https://deps.rs/repo/sourcehut/~alip/syd)
//! [![repology](https://repology.org/badge/latest-versions/sydbox.svg)](https://repology.org/project/sydbox/versions)
//!
//! [![SydB☮x](https://git.sr.ht/~alip/syd/blob/main/data/syd.png)](https://todo.sr.ht/~alip/syd)
//! [![GNU](https://web.archive.org/web/20221222061733if_/https://dev.exherbo.org/~alip/images/gnu.png)](https://www.gnu.org/philosophy/philosophy.html)
//! [![Linux](https://chesswob.org/jja/tux.png)](https://www.kernel.org/category/about.html)
//! [![Exherbo](https://web.archive.org/web/20230518155203if_/https://dev.exherbo.org/~alip/images/zebrapig.png)](https://www.exherbolinux.org/docs/gettingstarted.html)
//! [![musl libc](https://www.chesswob.org/jja/musl-inside.png)](https://www.musl-libc.org/)
//! [![libsecc☮mp](https://web.archive.org/web/20221222061720if_/https://dev.exherbo.org/~alip/images/libseccomp.png)](https://github.com/seccomp/libseccomp)
//! [![Paludis](http://paludis.exherbolinux.org/paludis_270.png)](https://paludis.exherbolinux.org)
//!
//! SydB☮x is a **seccomp**(2) based sandboxing utility for modern Linux\[\>=5.6\]
//! machines to sandbox unwanted process access to filesystem and network resources.
//! SydB☮x requires *no root access* and *no ptrace* rights. All you need is a
//! recent Linux kernel and libsecc☮mp which is available on many different
//! architectures, including **x86**, **x86\_64**, **x32**, **arm**, **aarch64**,
//! **mips**, **mips64**... This makes it very easy for a regular user to use. This is
//! the motto of SydB☮x: *bring easy, simple, flexible and powerful access restriction
//! to the Linux user!*
//!
//! The basic idea of SydB☮x is to run a command under certain restrictions. These
//! restrictions define which system calls the command is permitted to run and which
//! argument values are permitted for the given system call. The restrictions may be
//! applied via two ways.  *seccomp-bpf* can be used to apply simple Secure Computing
//! user filters to run sandboxing fully on kernel space, and *seccomp-notify*
//! functionality can be used to run sandboxing on kernel space and fallback to user
//! space to dereference pointer arguments of system calls (**See
//! [Security](#security) about `TOCTOU` et. al**), which are one of
//! **[pathname](https://en.wikipedia.org/wiki/Path_(computing))**, **[UNIX socket
//! address](https://en.wikipedia.org/wiki/Unix_domain_socket)**,
//! **[IPv4](https://en.wikipedia.org/wiki/IPv4)** or
//! **[IPv6](https://en.wikipedia.org/wiki/IPv6)** network address, and make dynamic
//! decisions using [Unix shell style patterns](https://docs.rs/globset) such as
//! `allowlist/write+/home/sydbox/***`, or `allowlist/write+/run/user/*/pulse` for
//! **[pathnames](https://en.wikipedia.org/wiki/Path_(computing))**, and using
//! **[CIDR](https://docs.rs/ipnetwork)** notation such as
//! `allowlist/net/connect+127.0.0.1/8!9050`, or
//! `allowlist/net/connect+::1/8!9050` for
//! **[IPv4](https://en.wikipedia.org/wiki/IPv4)** and
//! **[IPv6](https://en.wikipedia.org/wiki/IPv6)** addresses and perform an action
//! which is by default denying the system call with an appropriate error, which is
//! usually **access denied**, aka `EACCES`. For default disallowed system calls,
//! such as `ptrace` or `process_vm_writev` (**See [Security](#security) about
//! `TOCTOU` et. al**) SydB☮x returns `EACCES` as well.
//!
//! To be able to use SydB☮x, you need a recent Linux kernel with the system calls
//! **pidfd_getfd**, **pidfd_send_signal**. The Secure Computing facility of the
//! Linux kernel should support the **SECCOMP_USER_NOTIF_FLAG_CONTINUE** operation.
//! It is recommended to have the **CONFIG_CROSS_MEMORY_ATTACH** kernel option
//! enabled, if this option is not enabled, sydbox will fallback to reading/writing
//! from `/proc/$pid/mem`. Linux-5.11 or later is recommended.

// We like clean and simple code with documentation.
// Keep in sync with main.rs.
#![deny(missing_docs)]
#![deny(clippy::allow_attributes_without_reason)]
#![deny(clippy::arithmetic_side_effects)]
#![deny(clippy::as_ptr_cast_mut)]
#![deny(clippy::as_underscore)]
#![deny(clippy::assertions_on_result_states)]
#![deny(clippy::borrow_as_ptr)]
#![deny(clippy::branches_sharing_code)]
#![deny(clippy::case_sensitive_file_extension_comparisons)]
#![deny(clippy::cast_lossless)]
#![deny(clippy::cast_possible_truncation)]
#![deny(clippy::cast_possible_wrap)]
#![deny(clippy::cast_precision_loss)]
#![deny(clippy::cast_ptr_alignment)]
#![deny(clippy::cast_sign_loss)]
#![deny(clippy::checked_conversions)]
#![deny(clippy::clear_with_drain)]
#![deny(clippy::clone_on_ref_ptr)]
#![deny(clippy::cloned_instead_of_copied)]
#![deny(clippy::cognitive_complexity)]
#![deny(clippy::collection_is_never_read)]
#![deny(clippy::copy_iterator)]
#![deny(clippy::create_dir)]
#![deny(clippy::dbg_macro)]
#![deny(clippy::debug_assert_with_mut_call)]
#![deny(clippy::decimal_literal_representation)]
#![deny(clippy::default_trait_access)]
#![deny(clippy::default_union_representation)]
#![deny(clippy::derive_partial_eq_without_eq)]
#![deny(clippy::doc_link_with_quotes)]
//#![deny(clippy::doc_markdown)]
#![deny(clippy::explicit_into_iter_loop)]
#![deny(clippy::explicit_iter_loop)]
#![deny(clippy::fallible_impl_from)]
#![deny(clippy::missing_safety_doc)]
#![deny(clippy::undocumented_unsafe_blocks)]

/// Compatibility code for different libcs
pub(crate) mod compat;
/// Static configuration, edit & recompile!
pub mod config;
/// Filesystem utilities
pub mod fs;
/// Secure computing hooks
#[allow(clippy::as_ptr_cast_mut)]
#[allow(clippy::cast_sign_loss)]
#[allow(clippy::undocumented_unsafe_blocks)]
pub mod hook;
/// Simple logging on standard error using JSON lines
pub mod log;
/// /proc utilities
pub mod proc;
/// Sandbox configuration
pub mod sandbox;

// Vendored crates:
/// Interface to LandLock LSM
#[allow(dead_code)]
#[allow(missing_docs)]
#[allow(clippy::as_underscore)]
#[allow(clippy::cast_possible_truncation)]
#[allow(clippy::decimal_literal_representation)]
#[allow(clippy::missing_safety_doc)]
#[allow(clippy::type_complexity)]
#[allow(clippy::undocumented_unsafe_blocks)]
pub mod landlock;
/// The low-level interface for linux namespaces (containers)
pub mod unshare;

use std::{
    ffi::OsStr,
    fs::File,
    io::{BufRead, BufReader, Result as IOResult},
    os::{fd::RawFd, unix::ffi::OsStrExt},
    path::{Path, PathBuf},
    process::exit,
};

use libseccomp::{error::SeccompErrno, ScmpAction, ScmpArch, ScmpFilterContext, ScmpSyscall};
use nix::{
    errno::Errno,
    fcntl::OFlag,
    sys::{
        signal::{sigaction, SaFlags, SigAction, SigHandler, SigSet, Signal},
        socket::{socket, AddressFamily, SockFlag, SockType},
        stat::{stat, Mode},
        wait::{waitpid, WaitStatus},
    },
    unistd::{fork, ForkResult, Pid, Uid, User},
};
use once_cell::sync::Lazy;
use smallvec::SmallVec;

use crate::landlock::{
    path_beneath_rules, Access, AccessFs, RestrictionStatus, Ruleset, RulesetAttr,
    RulesetCreatedAttr, RulesetError, RulesetStatus, ABI,
};

/* Constants */

/// Number of CPUs on the system.
/// Initialized lazily on startup.
pub static NPROC: Lazy<usize> = Lazy::new(num_cpus::get);

/* Utilities */

/// Checks LandLock ABI v3 is supported.
/// Returns:
/// - 0: Fully enforced
/// - 1: Partially enforced
/// - 2: Not enforced
/// - 127: Unsupported
pub fn lock_enabled() -> u8 {
    let abi = ABI::V3;
    let path_ro = vec!["/".to_string()];
    let path_rw = vec!["/".to_string()];

    // A helper function to wrap the operations and reduce duplication
    fn landlock_operation(
        path_ro: &[String],
        path_rw: &[String],
        abi: ABI,
    ) -> Result<RestrictionStatus, RulesetError> {
        let ruleset = Ruleset::default().handle_access(AccessFs::from_all(abi))?;
        let created_ruleset = ruleset.create()?;
        let ro_rules = path_beneath_rules(path_ro, AccessFs::from_read(abi));
        let updated_ruleset = created_ruleset.add_rules(ro_rules)?;
        let rw_rules = path_beneath_rules(path_rw, AccessFs::from_all(abi));
        let final_ruleset = updated_ruleset.add_rules(rw_rules)?;
        final_ruleset.restrict_self().map_err(RulesetError::from)
    }

    match landlock_operation(&path_ro, &path_rw, abi) {
        Ok(status) => match status.ruleset {
            RulesetStatus::FullyEnforced => 0,
            RulesetStatus::PartiallyEnforced => 1,
            RulesetStatus::NotEnforced => 2,
        },
        Err(_) => 127,
    }
}

/// Returns true if we are running under SydB☮x.
#[allow(clippy::disallowed_methods)]
pub fn syd_enabled() -> bool {
    // This will not work if the sandbox is locked.
    // Path::new("/dev/syd").exists() || Path::new("/dev/sydbox").exists()
    // SAFETY: In libc, we trust.
    match unsafe { fork() } {
        Ok(ForkResult::Parent { child, .. }) => {
            match waitpid(child, None) {
                Ok(WaitStatus::Exited(_, code)) => {
                    // Check the child's exit status.
                    // Exit status of 0 means SydB☮x is enabled.
                    code == 0
                }
                _ => {
                    // If there's an error waiting on the
                    // child, assume SydB☮x is not enabled.
                    false
                }
            }
        }
        Ok(ForkResult::Child) => {
            let mut ctx = match ScmpFilterContext::new_filter(ScmpAction::Allow) {
                Ok(ctx) => ctx,
                Err(_) => exit(1),
            };

            let syscall = ScmpSyscall::new("open");
            if ctx.add_rule(ScmpAction::Notify, syscall).is_err() {
                exit(1);
            }

            if ctx.load().is_err() && Errno::last() == Errno::EBUSY {
                // seccomp filter exists
                // SydB☮x is in business.
                exit(0);
            } else {
                // seccomp filter does not exist
                exit(1);
            }
        }
        Err(_) => {
            // If there's an error forking,
            // assume SydB☮x is not enabled.
            false
        }
    }
}

/// Get the umask of the given `Pid` or current process'.
pub(crate) fn proc_umask(pid: Option<Pid>) -> Result<Mode, Errno> {
    // Construct path to the appropriate status file
    let path = match pid {
        Some(pid_value) => Path::new("/proc")
            .join(pid_value.to_string())
            .join("status"),
        None => Path::new("/proc/self/status").to_path_buf(),
    };

    // Open the file
    let file = File::open(path)
        .map_err(|error| Errno::from_i32(error.raw_os_error().unwrap_or(nix::libc::ENOENT)))?;
    let reader = BufReader::new(file);

    // Search for Umask line
    for line in reader
        .lines()
        .flatten()
        .filter(|line| line.starts_with("Umask:"))
    {
        // Parse the umask value
        let parts: Vec<&str> = line.split_whitespace().collect();
        if parts.len() == 2 {
            if let Ok(umask) = u32::from_str_radix(parts[1], 8) {
                return Ok(Mode::from_bits_truncate(umask));
            } else {
                return Err(Errno::EINVAL);
            }
        }
    }

    Err(Errno::ENOENT)
}

/// Returns the name of the libsecc☮mp native architecture.
pub fn seccomp_arch_native_name() -> Option<&'static str> {
    match ScmpArch::native() {
        ScmpArch::X86 => Some("x86"),
        ScmpArch::X8664 => Some("x86_64"),
        ScmpArch::X32 => Some("x32"),
        ScmpArch::Arm => Some("arm"),
        ScmpArch::Aarch64 => Some("aarch64"),
        ScmpArch::Mips => Some("mips"),
        ScmpArch::Mips64 => Some("mips64"),
        ScmpArch::Mips64N32 => Some("mips64n32"),
        ScmpArch::Mipsel => Some("mipsel"),
        ScmpArch::Mipsel64 => Some("mipsel64"),
        ScmpArch::Mipsel64N32 => Some("mipsel64n32"),
        ScmpArch::Ppc => Some("ppc"),
        ScmpArch::Ppc64 => Some("ppc64"),
        ScmpArch::Ppc64Le => Some("ppc64le"),
        ScmpArch::S390 => Some("s390"),
        ScmpArch::S390X => Some("s390s"),
        ScmpArch::Parisc => Some("parisc"),
        ScmpArch::Parisc64 => Some("parisc64"),
        ScmpArch::Riscv64 => Some("riscv64"),
        _ => None,
    }
}

/// Given a `Uid`, return the user name of the user.
/// On any error conditions, return "nobody".
pub fn get_user_name(uid: Uid) -> String {
    match User::from_uid(uid) {
        Ok(Some(user)) => user.name,
        _ => "nobody".to_string(),
    }
}

/// Given a username, return the home directory of the user.
/// On any error conditions, return "/var/empty".
pub fn get_user_home(username: &str) -> String {
    // Fetch user details.
    match User::from_name(username) {
        Ok(Some(user)) => user.dir.to_string_lossy().to_string(),
        _ => "/var/empty".to_string(),
    }
}

// Sets the specified signal to be ignored.
//
// This function utilizes the `sigaction` system call to set the specified signal's action
// to `SIG_IGN`, effectively causing the process to ignore that signal.
//
// # Arguments
//
// * `signal` - The signal number (e.g., `SIGTSTP`, `SIGTTIN`, `SIGTTOU`).
//
// # Returns
//
// * `Result<(), Error>` - Returns `Ok(())` if successful, or an error if the operation fails.
//
// # Example
//
// ```no_run
// use nix::sys::signal::SIGTSTP;
//
// let result = syd::ignore_signal(SIGTSTP);
// assert!(result.is_ok());
// ```
pub(crate) fn ignore_signal(signal: Signal) -> Result<(), Errno> {
    let sig_action = SigAction::new(
        SigHandler::SigIgn, // Set to ignore
        SaFlags::empty(),
        SigSet::empty(),
    );

    // SAFETY: The unsafe call to `sigaction` is used to set the signal's disposition
    // to "ignore". We're not invoking any handlers or performing any operations that
    // could lead to data races or other undefined behaviors. Hence, it's safe to call
    // in this context.
    unsafe {
        sigaction(signal, &sig_action)
            .map(|_| ())
            .map_err(|_| Errno::last())
    }
}

static PROC_SELF: &[u8] = b"/proc/self";
static PROC_SELF_THREAD: &[u8] = b"/proc/thread-self";
static DEV_FD: &[u8] = b"/dev/fd";
static DEV_STDIN: &[u8] = b"/dev/stdin";
static DEV_STDOUT: &[u8] = b"/dev/stdout";
static DEV_STDERR: &[u8] = b"/dev/stderr";

/// Replace /proc/self with /proc/pid and /dev/fd with /proc/pid/fd
pub(crate) fn replace_self_with_pid<P: AsRef<Path>>(path: P, pid: Pid) -> PathBuf {
    let path_bytes = path.as_ref().as_os_str().as_bytes();

    // Create buffer for PID only once
    let mut buf = itoa::Buffer::new();
    let pid_bytes = buf.format(pid.as_raw()).as_bytes();

    // Define replacements with byte slices
    let replacements = [
        (PROC_SELF, &[b"/proc/", pid_bytes].concat()),
        (PROC_SELF_THREAD, &[b"/proc/", pid_bytes].concat()),
        (DEV_FD, &[b"/proc/", pid_bytes, b"/fd"].concat()),
        (DEV_STDIN, &[b"/proc/", pid_bytes, b"/fd/0"].concat()),
        (DEV_STDOUT, &[b"/proc/", pid_bytes, b"/fd/1"].concat()),
        (DEV_STDERR, &[b"/proc/", pid_bytes, b"/fd/2"].concat()),
    ];

    // Create a new vector with enough capacity
    let mut new_path: SmallVec<[u8; 128]> = SmallVec::with_capacity(path_bytes.len());

    let mut start = 0;
    while start < path_bytes.len() {
        let mut replaced = false;
        for (needle, replacement) in &replacements {
            if let Some(end) = path_bytes[start..]
                .windows(needle.len())
                .position(|window| window == *needle)
            {
                // Check for end of string or followed by '/'
                let match_pos = start.saturating_add(end);
                let match_end = match_pos.saturating_add(needle.len());
                if match_end == path_bytes.len() || path_bytes[match_end] == b'/' {
                    new_path.extend_from_slice(&path_bytes[start..match_pos]);
                    new_path.extend_from_slice(replacement);
                    start = match_end;
                    replaced = true;
                    break;
                }
            }
        }
        if !replaced {
            new_path.extend_from_slice(&path_bytes[start..]);
            break;
        }
    }

    PathBuf::from(OsStr::from_bytes(&new_path))
}

/// Return system call priority by system call name.
#[inline(always)]
pub(crate) fn syscall_priority(name: &str) -> u8 {
    if matches!(name, "brk" | "mmap" | "mmap2") {
        255
    } else if name.starts_with("open") {
        240
    } else if name.contains("stat") {
        225
    } else if name.contains("access") {
        200
    } else if name.starts_with("execve") || name.contains("fork") || name.starts_with("clone") {
        175
    } else if name.starts_with("mkdir") || name.starts_with("rename") || name.starts_with("unlink")
    {
        150
    } else if matches!(
        name,
        "bind" | "connect" | "recvfrom" | "sendto" | "socketcall"
    ) {
        125
    } else if name.starts_with("getdents") {
        100
    } else {
        25
    }
}

/// Conerts an `OFlag` to a `Vec` of `String`s.
#[allow(clippy::cognitive_complexity)]
pub(crate) fn oflag_fmt(flags: OFlag) -> Vec<&'static str> {
    let mut vec = Vec::new();

    match flags & OFlag::O_ACCMODE {
        OFlag::O_RDONLY => {
            vec.push("O_RDONLY");
        }
        OFlag::O_WRONLY => {
            vec.push("O_WRONLY");
        }
        OFlag::O_RDWR => {
            vec.push("O_RDWR");
        }
        _ => {}
    };
    if flags.contains(OFlag::O_APPEND) {
        vec.push("O_APPEND");
    }
    if flags.contains(OFlag::O_ASYNC) {
        vec.push("O_ASYNC");
    }
    if flags.contains(OFlag::O_CLOEXEC) {
        vec.push("O_CLOEXEC");
    }
    if flags.contains(OFlag::O_CREAT) {
        vec.push("O_CREAT");
    }
    if flags.contains(OFlag::O_DIRECTORY) {
        vec.push("O_DIRECTORY");
    }
    if flags.contains(OFlag::O_EXCL) {
        vec.push("O_EXCL");
    }
    if flags.contains(OFlag::O_NOCTTY) {
        vec.push("O_NOCTTY");
    }
    if flags.contains(OFlag::O_NOFOLLOW) {
        vec.push("O_NOFOLLOW");
    }
    if flags.contains(OFlag::O_NONBLOCK) {
        vec.push("O_NONBLOCK");
    }
    if flags.contains(OFlag::O_SYNC) {
        vec.push("O_SYNC");
    }
    if flags.contains(OFlag::O_TRUNC) {
        vec.push("O_TRUNC");
    }

    vec
}

// device id and inode.
type FileId = (nix::libc::dev_t, nix::libc::ino_t);

pub(crate) static DEV_STAT: Lazy<Option<FileId>> = Lazy::new(|| get_stat("/dev"));

fn get_stat(path: &str) -> Option<FileId> {
    match stat(path) {
        Ok(metadata) => Some((metadata.st_dev, metadata.st_ino)),
        Err(_) => None,
    }
}

/// Like `std::path::Path::file_name()` but handles trailing slash transparently.
pub(crate) fn path_filename<P: AsRef<Path>>(path: P) -> PathBuf {
    let path = path.as_ref();
    let mut name = path
        .file_name()
        .map(PathBuf::from)
        .unwrap_or(PathBuf::from("."));

    // Handle trailing slash.
    if path.as_os_str().as_bytes().ends_with(b"/") {
        name.push("");
    }

    name
}

const IOPRIO_CLASS_IDLE: i32 = 3;
const IOPRIO_WHO_PROCESS: i32 = 1;

/// Sets the I/O priority of the current thread to idle.
///
/// This function uses the `ioprio_set` syscall to set the I/O
/// scheduling priority of the current thread to the idle class. The
/// idle I/O class is designed for tasks that should only use disk
/// resources when no other process needs them. When a thread is set to
/// idle, it will not compete with other (non-idle) processes for I/O
/// bandwidth.
///
/// Note that this setting is applied at the thread level in Linux,
/// where each thread is treated as a separate scheduling entity. As a
/// result, calling this function will only affect the I/O priority of
/// the thread from which it is called. If the application is
/// multi-threaded and a global I/O priority change is desired, this
/// function needs to be called from each thread, or specific threads
/// requiring the priority change should be targeted.
///
/// The function does not require any parameters and returns a `Result`:
/// - `Ok(())` on success.
/// - `Err(Errno)` containing Errno.
///
/// # Safety
///
/// This function involves an unsafe block due to the direct system call
/// (`nix::libc::syscall`). The `ioprio_set` syscall is considered
/// unsafe as it directly interfaces with the kernel, bypassing Rust's
/// safety guarantees.  However, the usage in this context is safe given
/// that:
/// - We are specifying `IOPRIO_WHO_PROCESS` with `0`, which correctly
///   targets the current thread.
/// - The `ioprio` value is correctly constructed for the idle I/O
///   class.
///
/// Users of this function do not need to take any special safety precautions.
pub(crate) fn set_io_priority_idle() -> Result<(), Errno> {
    // Set I/O priority: higher bits for the class, lower bits for the priority.
    // IOPRIO_CLASS_IDLE is shifted left by 13 bits to fit the class into higher bits.
    // Priority for idle class is not used, hence set to 0 (lower 13 bits).
    let ioprio = IOPRIO_CLASS_IDLE << 13;

    // SAFETY:
    // The syscall libc::SYS_ioprio_set is used to set the I/O priority
    // of a process. This call is considered unsafe because it involves
    // a direct system call, which bypasses the safety checks and
    // abstractions provided by Rust. However, this usage is safe under
    // the following conditions:
    // 1. The first argument IOPRIO_WHO_PROCESS specifies the target as
    //    a process.
    // 2. The second argument 0 refers to the current process. In the
    //    context of ioprio_set, passing 0 for the 'who' parameter
    //    targets the calling process. This is why getpid() is not
    //    necessary here, as 0 implicitly represents the current
    //    process's PID.
    // 3. The third argument ioprio is correctly constructed with a
    //    valid I/O class and priority, ensuring the syscall behaves as
    //    expected.
    if unsafe { nix::libc::syscall(nix::libc::SYS_ioprio_set, IOPRIO_WHO_PROCESS, 0, ioprio) } == 0
    {
        Ok(())
    } else {
        Err(Errno::last())
    }
}

const SECCOMP_ARCH_LIST: &[ScmpArch] = &[
    ScmpArch::X86,
    ScmpArch::X8664,
    ScmpArch::X32,
    ScmpArch::Arm,
    ScmpArch::Aarch64,
    ScmpArch::Mips,
    ScmpArch::Mips64,
    ScmpArch::Mips64N32,
    ScmpArch::Mipsel,
    ScmpArch::Mipsel64,
    ScmpArch::Mipsel64N32,
    ScmpArch::Ppc,
    ScmpArch::Ppc64,
    ScmpArch::Ppc64Le,
    ScmpArch::S390,
    ScmpArch::S390X,
    ScmpArch::Parisc,
    ScmpArch::Parisc64,
    ScmpArch::Riscv64,
];

/// Print list of libseccomp's supported architectures
/// Used by `syd --arch list`
pub fn print_seccomp_architectures() {
    let native = ScmpArch::native();
    for arch in SECCOMP_ARCH_LIST {
        let mut repr = format!("{arch:?}").to_ascii_lowercase();
        if repr == "x8664" {
            // Fix potential confusion.
            repr = "x86_64".to_string();
        }
        if *arch == native {
            println!("- {repr} [*]")
        } else {
            println!("- {repr}");
        }
    }
}

/// Add all supported architectures to the given filter.
#[allow(clippy::cognitive_complexity)]
pub(crate) fn seccomp_add_architectures(ctx: &mut ScmpFilterContext) -> IOResult<()> {
    // Add architectures based on the current architecture
    if cfg!(target_arch = "x86_64") {
        seccomp_add_arch(ctx, ScmpArch::X8664)?;
        seccomp_add_arch(ctx, ScmpArch::X86)?;
        seccomp_add_arch(ctx, ScmpArch::X32)?;
    } else if cfg!(target_arch = "x86") {
        seccomp_add_arch(ctx, ScmpArch::X86)?;
    } else if cfg!(target_arch = "arm") {
        seccomp_add_arch(ctx, ScmpArch::Arm)?;
    } else if cfg!(target_arch = "aarch64") {
        seccomp_add_arch(ctx, ScmpArch::Aarch64)?;
        seccomp_add_arch(ctx, ScmpArch::Arm)?;
    } else if cfg!(target_arch = "mips") {
        seccomp_add_arch(ctx, ScmpArch::Mips)?;
    } else if cfg!(target_arch = "mips64") {
        seccomp_add_arch(ctx, ScmpArch::Mips64)?;
        seccomp_add_arch(ctx, ScmpArch::Mips64N32)?;
        seccomp_add_arch(ctx, ScmpArch::Mips)?;
    } else if cfg!(target_arch = "mips64n32") {
        seccomp_add_arch(ctx, ScmpArch::Mips64N32)?;
        seccomp_add_arch(ctx, ScmpArch::Mips)?;
        seccomp_add_arch(ctx, ScmpArch::Mips64)?;
    } else if cfg!(target_arch = "mipsel") {
        seccomp_add_arch(ctx, ScmpArch::Mipsel64)?;
        seccomp_add_arch(ctx, ScmpArch::Mipsel64N32)?;
        seccomp_add_arch(ctx, ScmpArch::Mipsel)?;
    } else if cfg!(target_arch = "mipsel64") {
        seccomp_add_arch(ctx, ScmpArch::Mipsel64)?;
        seccomp_add_arch(ctx, ScmpArch::Mipsel)?;
        seccomp_add_arch(ctx, ScmpArch::Mipsel64N32)?;
    } else if cfg!(target_arch = "mipsel64n32") {
        seccomp_add_arch(ctx, ScmpArch::Mipsel64N32)?;
        seccomp_add_arch(ctx, ScmpArch::Mipsel)?;
        seccomp_add_arch(ctx, ScmpArch::Mipsel64)?;
    } else if cfg!(target_arch = "powerpc") {
        seccomp_add_arch(ctx, ScmpArch::Ppc)?;
    } else if cfg!(target_arch = "powerpc64") {
        seccomp_add_arch(ctx, ScmpArch::Ppc64)?;
        seccomp_add_arch(ctx, ScmpArch::Ppc64Le)?;
        seccomp_add_arch(ctx, ScmpArch::Ppc)?;
    } else if cfg!(target_arch = "parisc") {
        seccomp_add_arch(ctx, ScmpArch::Parisc)?;
    } else if cfg!(target_arch = "parisc64") {
        seccomp_add_arch(ctx, ScmpArch::Parisc64)?;
        seccomp_add_arch(ctx, ScmpArch::Parisc)?;
    } else if cfg!(target_arch = "riscv64") {
        seccomp_add_arch(ctx, ScmpArch::Riscv64)?;
    } else if cfg!(target_arch = "s390x") {
        seccomp_add_arch(ctx, ScmpArch::S390X)?;
        seccomp_add_arch(ctx, ScmpArch::S390)?;
    }

    Ok(())
}

fn seccomp2errno(errno: SeccompErrno) -> Errno {
    match errno {
        SeccompErrno::EACCES => Errno::EACCES,
        SeccompErrno::ECANCELED => Errno::ECANCELED,
        SeccompErrno::EDOM => Errno::EDOM,
        SeccompErrno::EEXIST => Errno::EEXIST,
        SeccompErrno::EFAULT => Errno::EFAULT,
        SeccompErrno::EINVAL => Errno::EINVAL,
        SeccompErrno::ENOENT => Errno::ENOENT,
        SeccompErrno::ENOMEM => Errno::ENOMEM,
        SeccompErrno::EOPNOTSUPP => Errno::EOPNOTSUPP,
        SeccompErrno::ERANGE => Errno::ERANGE,
        SeccompErrno::ESRCH => Errno::ESRCH,
        _ => Errno::ECANCELED,
    }
}

fn seccomp_add_arch(ctx: &mut ScmpFilterContext, arch: ScmpArch) -> IOResult<bool> {
    ctx.add_arch(arch).map_err(|e| {
        if let Some(errno) = e.errno() {
            std::io::Error::from_raw_os_error(seccomp2errno(errno) as i32)
        } else {
            std::io::Error::new(std::io::ErrorKind::Other, "Unknown seccomp error")
        }
    })
}

/// Simple human size formatter.
#[allow(clippy::arithmetic_side_effects)]
#[allow(clippy::cast_precision_loss)]
pub fn human_size(bytes: usize) -> String {
    const SIZES: &[&str] = &["Bytes", "KB", "MB", "GB", "TB", "PB", "EB"];
    let factor = 1024usize;

    let mut size = bytes as f64;
    let mut i = 0;

    while size > factor as f64 && i < SIZES.len() - 1 {
        size /= factor as f64;
        i += 1;
    }

    format!("{:.2} {}", size, SIZES[i])
}

/// Parse a FD from a Path.
pub(crate) fn parse_fd<P: AsRef<Path>>(path: P) -> Result<RawFd, Errno> {
    let bytes = path.as_ref().as_os_str().as_bytes();

    // Parsing bytes directly to integer
    let mut fd: RawFd = 0;
    #[allow(clippy::arithmetic_side_effects)]
    for &b in bytes {
        // Check if the byte is a valid digit
        if !b.is_ascii_digit() {
            return Err(Errno::EBADF);
        }
        fd = fd.saturating_mul(10).saturating_add((b - b'0') as RawFd);
    }

    Ok(fd)
}

#[cfg(target_env = "musl")]
pub(crate) type IoctlRequest = nix::libc::c_int;
#[cfg(not(target_env = "musl"))]
pub(crate) type IoctlRequest = nix::libc::c_ulong;

const SIOCGIFFLAGS: IoctlRequest = nix::libc::SIOCGIFFLAGS as IoctlRequest;
const SIOCSIFFLAGS: IoctlRequest = nix::libc::SIOCSIFFLAGS as IoctlRequest;

/// Functionally equivalent to "ifconfig lo up".
pub fn bring_up_loopback() -> Result<(), Errno> {
    // Create a socket
    let sockfd: RawFd = socket(
        AddressFamily::Inet,
        SockType::Stream,
        SockFlag::empty(),
        None,
    )?;

    // Prepare the interface request
    let mut ifreq = nix::libc::ifreq {
        ifr_name: {
            let mut name = [0 as nix::libc::c_char; 16]; // Initialize array with zeros as c_char
            let lo = b"lo\0"; // Interface name as byte array
            #[allow(clippy::cast_possible_wrap)]
            for (dest, &src) in name.iter_mut().zip(lo.iter()) {
                *dest = src as nix::libc::c_char; // Convert u8 to c_char and store in name
            }
            name // Use this initialized array
        },
        // SAFETY: Manually initialize ifr_ifru
        ifr_ifru: unsafe { std::mem::zeroed() },
    };

    // SAFETY: Get the current flags
    if unsafe { nix::libc::ioctl(sockfd, SIOCGIFFLAGS, &mut ifreq) } != 0 {
        return Err(Errno::last());
    }

    // Modify the flags to bring up the interface
    // SAFETY: We're accessing the field of a union here.
    #[allow(clippy::cast_possible_truncation)]
    unsafe {
        ifreq.ifr_ifru.ifru_flags |=
            (nix::libc::IFF_UP | nix::libc::IFF_RUNNING) as nix::libc::c_short
    };

    // SAFETY: Set the new flags
    if unsafe { nix::libc::ioctl(sockfd, SIOCSIFFLAGS, &mut ifreq) } != 0 {
        return Err(Errno::last());
    }

    Ok(())
}

#[cfg(test)]
mod tests {
    use nix::{sys::stat::umask, unistd::Pid};

    use super::*;

    // FIXME: This test is racy and fails randomly.
    #[ignore]
    #[test]
    fn test_self_umask() {
        let my_umask = umask(Mode::empty());
        umask(my_umask); // Restore umask

        let result = proc_umask(None).unwrap();
        assert_eq!(result, my_umask, "{result:o} != {my_umask:o}");
    }

    #[test]
    fn test_invalid_pid() {
        let result = proc_umask(Some(Pid::from_raw(i32::MAX)));
        assert!(result.is_err(), "{result:?}");
    }

    /*
     * FIXME: This fails randomly, however the next test
     * passes consistently, WTF?
     */
    #[ignore]
    #[test]
    fn test_known_pid_umask() {
        // Testing against the current process.
        let my_umask = Mode::from_bits_truncate(0o0077);
        let old_umask = umask(my_umask);
        let result = proc_umask(None).unwrap();
        assert_eq!(result, my_umask, "{result:o} != {my_umask:o}");
        let result = proc_umask(Some(Pid::this())).unwrap();
        assert_eq!(result, my_umask, "{result:o} != {my_umask:o}");
        umask(old_umask); // Restore umask
    }

    #[test]
    fn test_parsing_valid_umask_values() {
        // This test sets various umask values and then checks if our function correctly identifies them.
        let umasks = [
            Mode::from_bits_truncate(0o0000),
            Mode::from_bits_truncate(0o0002),
            Mode::from_bits_truncate(0o0022),
            Mode::from_bits_truncate(0o0077),
            Mode::from_bits_truncate(0o0777),
        ];

        for &my_umask in &umasks {
            umask(my_umask);
            let result = proc_umask(None).unwrap();
            assert_eq!(result, my_umask, "{result:o} != {my_umask:o}");
            let result = proc_umask(Some(Pid::this())).unwrap();
            assert_eq!(result, my_umask, "{result:o} != {my_umask:o}");
        }

        // Resetting the umask to a default value after test
        umask(Mode::from_bits_truncate(0o0022));
    }

    #[test]
    fn test_replace_self_with_pid_only() {
        let result = replace_self_with_pid("/proc/self/cmdline", Pid::from_raw(1234));
        assert_eq!(result, PathBuf::from("/proc/1234/cmdline"));
    }

    #[test]
    fn test_replace_fd_only() {
        let result = replace_self_with_pid("/dev/fd/3", Pid::from_raw(5678));
        assert_eq!(result, PathBuf::from("/proc/5678/fd/3"));
    }

    #[test]
    fn test_no_replacement() {
        let result = replace_self_with_pid("/some/random/path", Pid::from_raw(91011));
        assert_eq!(result, PathBuf::from("/some/random/path"));
    }

    #[test]
    fn test_replace_both_patterns() {
        let result = replace_self_with_pid("/proc/self/cmdline/dev/fd/3", Pid::from_raw(1213));
        assert_eq!(result, PathBuf::from("/proc/1213/cmdline/proc/1213/fd/3"));
    }

    #[test]
    fn test_replace_dev_stdin() {
        let result = replace_self_with_pid("/dev/stdin", Pid::from_raw(1213));
        assert_eq!(result, PathBuf::from("/proc/1213/fd/0"));
    }

    #[test]
    fn test_replace_dev_stdout() {
        let result = replace_self_with_pid("/dev/stdout", Pid::from_raw(1213));
        assert_eq!(result, PathBuf::from("/proc/1213/fd/1"));
    }

    #[test]
    fn test_replace_dev_stderr() {
        let result = replace_self_with_pid("/dev/stderr", Pid::from_raw(1213));
        assert_eq!(result, PathBuf::from("/proc/1213/fd/2"));
    }
}
