//
// SydB☮x: seccomp and landlock based application sandbox with support for namespaces
// src/main.rs: Main entry point
//
// Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

//! syd: seccomp and landlock based application sandbox with support for namespaces
//! Main entry point.

// We like clean and simple code with documentation.
#![deny(missing_docs)]
#![deny(clippy::allow_attributes_without_reason)]
#![deny(clippy::arithmetic_side_effects)]
#![deny(clippy::as_ptr_cast_mut)]
#![deny(clippy::as_underscore)]
#![deny(clippy::assertions_on_result_states)]
#![deny(clippy::borrow_as_ptr)]
#![deny(clippy::branches_sharing_code)]
#![deny(clippy::case_sensitive_file_extension_comparisons)]
#![deny(clippy::cast_lossless)]
#![deny(clippy::cast_possible_truncation)]
#![deny(clippy::cast_possible_wrap)]
#![deny(clippy::cast_precision_loss)]
#![deny(clippy::cast_ptr_alignment)]
#![deny(clippy::cast_sign_loss)]
#![deny(clippy::checked_conversions)]
#![deny(clippy::clear_with_drain)]
#![deny(clippy::clone_on_ref_ptr)]
#![deny(clippy::cloned_instead_of_copied)]
#![deny(clippy::cognitive_complexity)]
#![deny(clippy::collection_is_never_read)]
#![deny(clippy::copy_iterator)]
#![deny(clippy::create_dir)]
#![deny(clippy::dbg_macro)]
#![deny(clippy::debug_assert_with_mut_call)]
#![deny(clippy::decimal_literal_representation)]
#![deny(clippy::default_trait_access)]
#![deny(clippy::default_union_representation)]
#![deny(clippy::derive_partial_eq_without_eq)]
#![deny(clippy::doc_link_with_quotes)]
#![deny(clippy::doc_markdown)]
#![deny(clippy::explicit_into_iter_loop)]
#![deny(clippy::explicit_iter_loop)]
#![deny(clippy::fallible_impl_from)]
#![deny(clippy::missing_safety_doc)]
#![deny(clippy::undocumented_unsafe_blocks)]

use std::{
    collections::HashSet,
    env,
    ffi::{CString, OsStr},
    fs::File,
    io::Write,
    os::unix::{ffi::OsStrExt, process::CommandExt},
    path::{Path, PathBuf},
    process::{Command, ExitCode},
    str::FromStr,
};

use anyhow::{bail, Context, Error};
use getargs::{Opt, Options};
use libseccomp::{ScmpAction, ScmpArch, ScmpFilterContext, ScmpSyscall, ScmpVersion};
use nix::{
    errno::Errno,
    mount::{mount, umount2, MntFlags, MsFlags},
    sched::{unshare, CloneFlags},
    sys::wait::{waitpid, WaitStatus},
    unistd::{chdir, chroot, fork, getgid, getpid, getuid, pivot_root, sethostname, ForkResult},
    NixPath,
};
use once_cell::sync::Lazy;
use parse_size::parse_size;
use syd::{
    bring_up_loopback,
    config::*,
    error,
    hook::{ExportMode, Supervisor},
    info,
    landlock::{
        path_beneath_rules, Access, AccessFs, RestrictionStatus, Ruleset, RulesetAttr,
        RulesetCreatedAttr, RulesetError, RulesetStatus, ABI,
    },
    sandbox,
    unshare::{GidMap, UidMap},
    warn,
};

static SYD_VERSION: Lazy<&'static str> = Lazy::new(|| {
    if env!("SYD_GITHEAD").is_empty() {
        env!("CARGO_PKG_VERSION")
    } else {
        env!("SYD_GITHEAD")
    }
});

// This line will include the content of "esyd.sh" at compile-time
const ESYD: &str = include_str!("esyd.sh");

// Landlock API in use.
const LANDLOCK_ABI: ABI = ABI::V3;
const LANDLOCK_ABI_STR: &str = "v3";

const NONE: Option<&PathBuf> = None::<PathBuf>.as_ref();

// A helper function to wrap the operations and reduce duplication
fn landlock_operation(
    path_ro: &[String],
    path_rw: &[String],
    abi: ABI,
) -> Result<RestrictionStatus, RulesetError> {
    let ruleset = Ruleset::default().handle_access(AccessFs::from_all(abi))?;
    let created_ruleset = ruleset.create()?;
    let ro_rules = path_beneath_rules(path_ro, AccessFs::from_read(abi));
    let updated_ruleset = created_ruleset.add_rules(ro_rules)?;
    let rw_rules = path_beneath_rules(path_rw, AccessFs::from_all(abi));
    let final_ruleset = updated_ruleset.add_rules(rw_rules)?;
    final_ruleset.restrict_self().map_err(RulesetError::from)
}

fn help() {
    let libapi = libseccomp::get_api();
    let libver = ScmpVersion::current().ok();
    let (libmaj, libmin, libmic) = libver
        .map(|v| (v.major, v.minor, v.micro))
        .unwrap_or((0, 0, 0));
    let native = syd::seccomp_arch_native_name().unwrap_or("?");

    println!(
        "SydB☮x {} libsecc☮mp v{}.{}.{} api:{} arch:{}
seccomp and landlock based application sandbox with support for namespaces
Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
SPDX-License-Identifier: GPL-3.0-or-later

# Usage

syd [-hvcl]
    [--lock] [--root] [--chroot=root]
    [-m command...] [-C path...] [-p name...]
    [--unshare-mount,uts,ipc,user,pid,net,cgroup]
    [-A arg0] [--map-root]
    [--] {{command [arg...]}}
syd --check
syd --landlock
syd --export bpf|pfc
syd --list allow|deny|hook|ioctl|prctl
syd --print profile-name
syd --errno number|name-regex
syd --parse size|human-size
syd --read path
syd --syscall number|name-regex [--arch list|native|x86|x86_64|aarch64...]
syd --sh
syd --test [<name-regex>|<number>|<number>..<number>]..
syd exec {{command [arg...]}}
syd log

-h            -- Show usage and exit
-v            -- Show version and exit
-c            -- Causes command to be executed under a shell with the \"user\" profile
                 Shell is /bin/bash by default, override with SYD_SH
                 Login shell compatibility
-l            -- Ignored
                 Login shell compatibility
--lock        -- Lock magic commands
--root        -- Return 0 as user/group id, aka \"fakeroot\" mode
-m command    -- Run a magic command during init, may be repeated
-C path       -- Path to the configuration file to use, may be repeated
-p name       -- Name of the sandbox profile to use, may be repeated

--check       -- Exit with success if the process is running under SydB☮x
--landlock    -- Exit with success if LandLock ABI v3 is fully supported
--export mode -- Export seccomp rules in given format to standard output and exit
--list set    -- Print the names of the system calls which belong to the given set and exit
                 If set is \"ioctl\", print the list of allowlisted ioctl requests
                 If set is \"prctl\", print the list of allowlisted prctl options
--print name  -- Print out the rules of the given sandbox profile and exit
--errno arg   -- Given a number, print the matching errno name and exit
                 Given a regex, print case-insensitively matching errno names and exit
--parse size  -- Given a number, print human-formatted size and exit
                 Given a string, parse human-formatted size into bytes, print and exit
--read path   -- Print the canonicalized path name followed by a newline and exit
--syscall arg -- Given a number, print the matching syscall name and exit
                 Given a regex, print case-insensitively matching syscall names and exit
--sh          -- Output a shell script which defines \"esyd\" the SydB☮x helper function
--test        -- Run integration tests and exit
                 Requires syd-test & syd-test-do programs to be in PATH.

-M, --unshare-mount   -- Unshare mount namespace
-S, --unshare-uts     -- Unshare UTS namespace
-I, --unshare-ipc     -- Unshare IPC namespace
-U, --unshare-user    -- Unshare user namespace
-P, --unshare-pid     -- Unshare pid namespace
-N, --unshare-net     -- Unshare net namespace
-G, --unshare-cgroup  -- Unshare cgroup namespace

-A, --arg0 alias      -- Set alias of the command
                         Passed as `argv[0]` to the program
-0, --map-root        -- Map current user to root in the sandbox
                         Implies --unshare-user
-R, --chroot=root     -- Change directory and chroot to root before starting the process
                         The directory root/proc must exist to mount private procfs
                         Implies --unshare-mount and --unshare-pid

* syd exec may be used to construct a sandbox command to execute a
process outside the sandbox. See the description of cmd/exec
command in the README for more information.
* syd log may be used to access sandbox logs using journalctl.

# Profiles

1. container: Enables Linux namespaces.
              Equivalent to --unshare-mount,uts,ipc,user,pid,net,cgroup
2. landlock: Enables LandLock and allowlists system directories
             for Lock Sandboxing.
3. paludis: Used by the Paludis package mangler.
4. noipv4: Disables IPv4 connectivity.
5. noipv6: Disables Ipv6 connectivity.
6. silent: Silences all access violations.
7. lib: LibSyd helper profile. Turns all sandboxing off.
        Useful to configure SydB☮x in the application using LibSyd.
8. user: Allows user-specific directories and connections,
         and parses the following files if they exist:
         1. /etc/user.syd-3
         2. ~/.user.syd-3

When invoked without arguments, the current shell is executed
under sandbox with the user profile.

# Environment

SYD_LOG: Set log level. See the Enabling Logging¹ section of
 env-logger crate documentation for more information.
SYD_NO_SYSLOG: Disable logging to syslog(3). By default logs of
               severity Warn and higher are logged to syslog(3).
SYD_NO_CROSS_MEMORY_ATTACH: Disable cross memory attach,
                            and fallback to /proc/$pid/mem.
SYD_FAKEROOT: Equivalent to passing --root
SYD_SH: Pick the shell to spawn when invoked as a login shell
        Defaults to /bin/bash
SYD_UNSHARE_MOUNT: Equivalent to passing --unshare-mount
SYD_UNSHARE_UTS: Equivalent to passing --unshare-uts
SYD_UNSHARE_IPC: Equivalent to passing --unshare-ipc
SYD_UNSHARE_PID: Equivalent to passing --unshare-pid
SYD_UNSHARE_NET: Equivalent to passing --unshare-net
SYD_UNSHARE_CGROUP: Equivalent to passing --unshare-cgroup

# Exit Codes

Sydb☮x exits with the same exit code as the sandbox process
itself. If the sandbox process exits with a signal, Sydb☮x exits
with 14 which stands for EFAULT. In case there was an error in
spawning or waiting for the sandbox process, Sydb☮x exits with
errno indicating the error condition. E.g. syd true\" returns 0,
\"syd false\" return 1, and \"syd -- syd true\" returns 16 which
stands for EBUSY which stands for \"Device or resource busy\"
indicating there is already a secure computing filter
loaded. tl;dr Sydb☮x won't run under Sydb☮x, similarly many
process inspection tools such as \"ltrace\", \"strace\", or
\"gdb\" won't work under Sydb☮x. Thus the sandbox process can
either be traced by attaching from outside the sandbox or
running the tracer in follow fork mode, e.g.
\"strace -f syd true\".

¹: https://docs.rs/env_logger/latest/env_logger/#enabling-logging",
        *SYD_VERSION, libmaj, libmin, libmic, libapi, native
    );
}

#[allow(clippy::cognitive_complexity)]
fn main() -> anyhow::Result<ExitCode> {
    // Step 1: Parse CLI arguments.
    let mut argv_iter = argv::iter();
    #[allow(clippy::disallowed_methods)]
    let is_login = *argv_iter.next().unwrap().as_bytes().first().unwrap() == b'-';
    #[allow(clippy::disallowed_methods)]
    let mut opts =
        Options::new(argv_iter.map(|os| os.to_str().expect("Argument is not valid UTF-8!")));

    // Initialize options.
    let sh = env::var(ENV_SH).unwrap_or(SYD_SH.to_string());
    let mut arg0: Option<String> = None;
    let mut argv: Vec<String> = Vec::new();
    let mut test = false;
    let mut arch = ScmpArch::Native;
    let mut call = None;
    let mut root = None;
    let mut fakeroot = env::var(ENV_FAKEROOT).is_ok();
    let mut map_root = false;
    let mut export_mode: Option<ExportMode> = None;
    let mut list_set: Option<String> = None;
    let mut print_profile: Option<String> = None;
    let mut sandbox = sandbox::Sandbox::new();

    // Namespaces handling magic
    const ENV_INIT: &str = "SYD_INIT";
    let init = env::var(ENV_INIT).is_err();

    // Initialize namespace options from the environment.
    sandbox.set_unshare_mount(env::var(ENV_UNSHARE_MOUNT).is_ok());
    sandbox.set_unshare_uts(env::var(ENV_UNSHARE_UTS).is_ok());
    sandbox.set_unshare_ipc(env::var(ENV_UNSHARE_IPC).is_ok());
    sandbox.set_unshare_user(env::var(ENV_UNSHARE_USER).is_ok());
    sandbox.set_unshare_pid(env::var(ENV_UNSHARE_PID).is_ok());
    sandbox.set_unshare_net(env::var(ENV_UNSHARE_NET).is_ok());
    sandbox.set_unshare_cgroup(env::var(ENV_UNSHARE_CGROUP).is_ok());

    while let Some(opt) = opts.next_opt().context("calling Options::next")? {
        match opt {
            /*
             * Basic options
             */
            Opt::Short('h') | Opt::Long("help") => {
                help();
                return Ok(ExitCode::from(0));
            }
            Opt::Short('v') | Opt::Long("version") => {
                println!("SydB☮x {}", *SYD_VERSION);
                println!("Auth☮r: Ali P☮latel <alip@chessw☮b.☮rg>");
                println!("License: GPL-3.0-☮r-later");

                let feat = [
                    #[cfg(debug_assertions)]
                    "+debug",
                    #[cfg(not(debug_assertions))]
                    "-debug",
                    #[cfg(feature = "log")]
                    "+log",
                    #[cfg(not(feature = "log"))]
                    "-log",
                    #[cfg(feature = "uring")]
                    "+uring",
                    #[cfg(not(feature = "uring"))]
                    "-uring",
                ];
                println!("Features: {}", feat.join(", "));

                let state = match syd::lock_enabled() {
                    0 => "Fully enf☮rced",
                    1 => "Partially enf☮rced",
                    2 => "Not enf☮rced",
                    _ => "Unsupp☮rted",
                };
                println!("LandL☮ck: {state}");

                let libapi = libseccomp::get_api();
                let libver =
                    ScmpVersion::current().context("Failed to determine libsecc☮mp version.")?;
                println!(
                    "LibSecc☮mp: v{}.{}.{} api:{}",
                    libver.major, libver.minor, libver.micro, libapi
                );

                let uname = match nix::sys::utsname::uname() {
                    Ok(info) => OsStr::to_str(info.release()).unwrap_or("?").to_string(),
                    Err(_) => "?".to_string(),
                };
                println!(
                    "H☮st: {uname} {}",
                    syd::seccomp_arch_native_name().unwrap_or("?")
                );
                println!("C☮mp: {}", env!("SYD_BUILDHOST"));

                return Ok(ExitCode::from(0));
            }

            /*
             * Sandbox options
             */
            Opt::Long("check") => {
                return Ok(ExitCode::from(if syd::syd_enabled() { 0 } else { 1 }));
            }
            Opt::Long("landlock") => {
                return Ok(ExitCode::from(syd::lock_enabled()));
            }
            Opt::Long("sh") => {
                println!("{ESYD}");
                return Ok(ExitCode::from(0));
            }
            Opt::Long("export") => {
                sandbox
                    .parse_profile("paludis")
                    .context("Unable to parse paludis profile!")?;
                export_mode = Some(ExportMode::from_str(
                    opts.value().context("--export requires an argument!")?,
                )?);
            }
            Opt::Long("list") => {
                list_set = Some(
                    opts.value()
                        .context("--list requires an argument!")?
                        .to_string(),
                );
            }
            Opt::Long("print") => {
                print_profile = Some(
                    opts.value()
                        .context("--print requires an argument!")?
                        .to_string(),
                );
            }
            Opt::Long("errno") => {
                let val = opts.value().context("--errno requires an argument!")?;
                match val.parse::<u16>() {
                    Ok(0) => {
                        return Ok(ExitCode::from(1));
                    }
                    Ok(num) => {
                        // number -> name
                        let errno = Errno::from_i32(i32::from(num));
                        if errno == Errno::UnknownErrno {
                            return Ok(ExitCode::from(1));
                        }
                        let estr = errno.to_string();
                        let mut iter = estr.split(": ");
                        let name = iter.next().unwrap_or("?");
                        let desc = iter.next().unwrap_or("?");
                        println!("{num}\t{name}\t{desc}");
                    }
                    Err(_) => {
                        match regex::RegexBuilder::new(val)
                            .unicode(false)
                            .case_insensitive(true)
                            .build()
                        {
                            Ok(pattern) => {
                                // regex -> [number]
                                let mut ok = false;
                                for errno in (1..u8::MAX).map(|n| Errno::from_i32(i32::from(n))) {
                                    if errno == Errno::UnknownErrno {
                                        continue;
                                    }
                                    let estr = errno.to_string();
                                    let mut iter = estr.split(": ");
                                    let name = iter.next().unwrap_or("?");
                                    let desc = iter.next().unwrap_or("?");
                                    if pattern.is_match(name) {
                                        println!("{}\t{}\t{}", errno as i32, name, desc);
                                        ok = true;
                                    }
                                }
                                if !ok {
                                    return Ok(ExitCode::from(1));
                                }
                            }
                            Err(error) => {
                                eprintln!("Invalid errno regex \"{val}\": {error}");
                                return Ok(ExitCode::from(1));
                            }
                        }
                    }
                }
                return Ok(ExitCode::from(0));
            }
            Opt::Long("parse") => {
                let value = opts.value().context("--parse requires an argument!")?;
                if value.chars().all(|c| c.is_ascii_digit()) {
                    match value.parse::<usize>() {
                        Ok(size) => {
                            println!("{}", syd::human_size(size));
                            return Ok(ExitCode::from(0));
                        }
                        Err(error) => {
                            bail!("Failed to parse: {error}");
                        }
                    }
                } else {
                    match parse_size(value) {
                        Ok(size) => {
                            println!("{size}");
                            return Ok(ExitCode::from(0));
                        }
                        Err(error) => {
                            bail!("Failed to parse: {error}");
                        }
                    }
                }
            }
            Opt::Long("read") => {
                let path = Path::new(opts.value().context("--read requires an argument!")?);
                let path = if path.is_relative() {
                    let cdir = syd::fs::canonicalize(
                        "/proc/self/cwd",
                        true,
                        syd::fs::MissingHandling::Missing,
                    )
                    .context("Failed to canonicalize /proc/self/cwd")?;
                    cdir.join(path)
                } else {
                    path.to_path_buf()
                };
                let path = syd::fs::canonicalize(path, true, syd::fs::MissingHandling::Missing)
                    .context("Failed to canonicalize path!")?;
                let path = path.as_os_str().as_bytes();
                std::io::stdout()
                    .write_all(path)
                    .context("Error writing canonicalized path to standard output!")?;
                println!();
                return Ok(ExitCode::from(0));
            }
            Opt::Long("arch") => {
                let val = opts.value()
                    .context("--arch requires an argument, do '--arch list' to print the list of architectures.")?;
                if matches!(val.to_ascii_lowercase().as_str(), "help" | "list") {
                    syd::print_seccomp_architectures();
                    return Ok(ExitCode::from(0));
                }
                arch = ScmpArch::from_str(&format!("SCMP_ARCH_{}", val.to_ascii_uppercase()))
                    .context(format!("Invalid architecture `{val}'"))?;
            }
            Opt::Long("syscall") => {
                call = Some(opts.value().context("--syscall requires an argument!")?);
            }
            Opt::Long("test") => {
                test = true;
            }
            Opt::Long("lock") => {
                sandbox.lock();
            }
            Opt::Long("root") => {
                fakeroot = true;
            }
            Opt::Short('R') | Opt::Long("chroot") => {
                let val = opts.value().context("--chroot requires an argument!")?;
                if !val.starts_with('/') {
                    bail!("Chroot directory must be an absolute path!");
                }
                root = Some(Path::new(val));
                sandbox.set_unshare_mount(true);
                sandbox.set_unshare_pid(true);
            }
            Opt::Short('m') | Opt::Long("magic") => {
                let cmd = opts.value().context("--magic requires an argument!")?;
                if sandbox.locked() {
                    warn!("ctx": "config", "pid": getpid().as_raw(), "cfg": cmd, "err": "LOCK");
                } else {
                    sandbox
                        .config(cmd)
                        .context(format!("Failed to execute magic command `{cmd}'."))?;
                }
            }
            Opt::Short('l') | Opt::Long("login") => {
                // Ignored, login shell compatibility
            }
            Opt::Short('c') => {
                // syd -c command args... akin to bash -c command args..
                if argv.is_empty() {
                    sandbox
                        .parse_profile("user")
                        .context("Unable to parse user profile!")?;
                    argv.push(sh.clone());
                    argv.push("-c".to_string());
                }
            }
            Opt::Short('C') | Opt::Long("config") => {
                let path = opts.value().context("--config requires an argument!")?;
                if sandbox.locked() {
                    warn!("ctx": "config", "pid": getpid().as_raw(), "path": path, "err": "LOCK");
                } else {
                    sandbox
                        .parse_config_file(path)
                        .context(format!("Failed to parse configuration file `{path}'."))?;
                }
            }
            Opt::Short('p') | Opt::Long("profile") => {
                let profile = opts.value().context("--profile requires an argument!")?;
                if sandbox.locked() {
                    warn!("ctx": "config", "pid": getpid().as_raw(), "prof": profile, "err": "LOCK");
                } else {
                    sandbox.parse_profile(profile).context(format!(
                        "Failed to parse configuration profile `{profile}'."
                    ))?;
                }
            }

            /*
             * Unshare options
             */
            Opt::Short('M') | Opt::Long("unshare-mount") => {
                sandbox.set_unshare_mount(true);
            }
            Opt::Short('S') | Opt::Long("unshare-uts") => {
                sandbox.set_unshare_uts(true);
            }
            Opt::Short('I') | Opt::Long("unshare-ipc") => {
                sandbox.set_unshare_ipc(true);
            }
            Opt::Short('U') | Opt::Long("unshare-user") => {
                sandbox.set_unshare_user(true);
            }
            Opt::Short('P') | Opt::Long("unshare-pid") => {
                sandbox.set_unshare_pid(true);
            }
            Opt::Short('N') | Opt::Long("unshare-net") => {
                sandbox.set_unshare_net(true);
            }
            Opt::Short('G') | Opt::Long("unshare-cgroup") => {
                sandbox.set_unshare_cgroup(true);
            }
            Opt::Short('A') | Opt::Long("arg0") => {
                let val = opts
                    .value()
                    .context("--arg0 requires an argument!")?
                    .to_string();
                arg0 = Some(val);
            }
            Opt::Short('0') | Opt::Long("map-root") => {
                map_root = true;
                sandbox.set_unshare_user(true);
            }

            Opt::Short(c) => {
                bail!("Invalid option `-{c}'!");
            }
            Opt::Long(c) => {
                bail!("Invalid option `--{c}'!");
            }
        }
    }

    if test {
        Command::new("syd-test").args(opts.positionals()).exec();
    } else if fakeroot {
        sandbox.set_root(true);
    }

    // --syscall
    if let Some(val) = call {
        match val.parse::<i32>() {
            Ok(num) => {
                let syscall = ScmpSyscall::from(num);
                if let Ok(name) = syscall.get_name_by_arch(arch) {
                    println!("{num}\t{name}");
                } else {
                    return Ok(ExitCode::from(1));
                }
            }
            Err(_) => {
                match regex::RegexBuilder::new(val)
                    .unicode(false)
                    .case_insensitive(true)
                    .build()
                {
                    Ok(pattern) => {
                        let mut ok = false;
                        for (num, name) in (0..4096)
                            .map(|n| {
                                (
                                    n,
                                    ScmpSyscall::from(n)
                                        .get_name_by_arch(arch)
                                        .unwrap_or_default(),
                                )
                            })
                            .filter(|(_, name)| !name.is_empty())
                        {
                            if pattern.is_match(&name) {
                                println!("{num}\t{name}");
                                ok = true;
                            }
                        }
                        if !ok {
                            return Ok(ExitCode::from(1));
                        }
                    }
                    Err(error) => {
                        eprintln!("Invalid syscall regex \"{val}\": {error}");
                        return Ok(ExitCode::from(1));
                    }
                }
            }
        }
        return Ok(ExitCode::from(0));
    }

    // List set or Print profile if requested, and exit.
    match (list_set.as_deref(), print_profile.as_deref()) {
        (Some(_), Some(_)) => {
            eprintln!("The options --list and --print are mutually exclusive!");
            return Ok(ExitCode::from(1));
        }
        (Some("ioctl"), _) => {
            for (_, name) in ALLOWLIST_IOCTL {
                println!("{name}");
            }
            return Ok(ExitCode::from(0));
        }
        (Some("prctl"), _) => {
            for (_, name) in ALLOWLIST_PRCTL {
                println!("{name}");
            }
            return Ok(ExitCode::from(0));
        }
        (Some("allow"), _) => {
            for name in SAFE_SYSCALLS {
                println!("{name}");
            }
            return Ok(ExitCode::from(0));
        }
        (Some("hook"), _) => {
            for name in HOOK_SYSCALLS {
                println!("{name}");
            }
            return Ok(ExitCode::from(0));
        }
        (Some("deny"), _) => {
            let mut syscall_set: HashSet<_> =
                SAFE_SYSCALLS.iter().map(|&s| String::from(s)).collect();
            for syscall in HOOK_SYSCALLS {
                syscall_set.insert(syscall.to_string());
            }
            for syscall_number in 0..=600 {
                let syscall = ScmpSyscall::from(syscall_number);
                if let Ok(name) = syscall.get_name() {
                    if !syscall_set.contains(&name) {
                        println!("{name}");
                    }
                }
            }
            return Ok(ExitCode::from(0));
        }
        (Some(set), _) => {
            eprintln!("No such set: '{set}'");
            return Ok(ExitCode::from(1));
        }
        (_, Some("container")) => {
            println!("# SydB☮x profile: Container");
            println!("# Number of rules: {}", PROFILE_CONTAINER.len());
            println!("# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in PROFILE_CONTAINER {
                println!("{command}");
            }
            return Ok(ExitCode::from(0));
        }
        (_, Some("landlock")) => {
            println!("# SydB☮x profile: LandLock");
            println!("# Number of rules: {}", PROFILE_LANDLOCK.len());
            println!("# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in PROFILE_LANDLOCK {
                println!("{command}");
            }
            return Ok(ExitCode::from(0));
        }
        (_, Some("paludis")) => {
            println!("# SydB☮x profile: Paludis");
            println!("# Number of rules: {}", PROFILE_PALUDIS.len());
            println!("# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in PROFILE_PALUDIS {
                println!("{command}");
            }
            return Ok(ExitCode::from(0));
        }
        (_, Some("noipv4")) => {
            println!("# SydB☮x profile: NoIpv4");
            println!("# Number of rules: {}", PROFILE_NOIPV4.len());
            println!("# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in PROFILE_NOIPV4 {
                println!("{command}");
            }
            return Ok(ExitCode::from(0));
        }
        (_, Some("noipv6")) => {
            println!("# SydB☮x profile: NoIpv6");
            println!("# Number of rules: {}", PROFILE_NOIPV6.len());
            println!("# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in PROFILE_NOIPV6 {
                println!("{command}");
            }
            return Ok(ExitCode::from(0));
        }
        (_, Some("silent")) => {
            println!("# SydB☮x profile: Silent");
            println!("# Number of rules: {}", PROFILE_SILENT.len());
            println!("# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in PROFILE_SILENT {
                println!("{command}");
            }
            return Ok(ExitCode::from(0));
        }
        (_, Some("lib")) => {
            println!("# SydB☮x profile: LibSyd");
            println!("# Number of rules: {}", PROFILE_SILENT.len());
            println!("# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in PROFILE_LIB {
                println!("{command}");
            }
            return Ok(ExitCode::from(0));
        }
        (_, Some("user")) => {
            let uid = nix::unistd::getuid();
            let name = syd::get_user_name(uid);
            let mut home = syd::get_user_home(&name);

            println!("# SydB☮x profile: User \"{name}\"");
            println!(
                "# Number of rules: {}",
                PROFILE_USER.len().saturating_add(9)
            );
            println!("# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");

            // Step 1: Static configuration defined at compile time.
            for command in PROFILE_USER {
                println!("{command}");
            }

            // Step 2: Dynamic, user-specific configuration.
            // Keep in sync with Sandbox::parse_profile()
            if !home.ends_with('/') {
                home.push('/');
            }

            println!("allowlist/lock/write+{home}");
            println!("allowlist/read+{home}***");
            println!("allowlist/stat+{home}***");
            println!("allowlist/write+{home}**");
            println!("allowlist/exec+{home}**");
            println!("allowlist/net/bind+{home}**");
            println!("allowlist/net/connect+{home}**");
            println!("allowlist/read+/run/user/{uid}/**");
            println!("allowlist/write+/run/user/{uid}/**");

            return Ok(ExitCode::from(0));
        }
        (_, Some(name)) => {
            eprintln!("No such profile: '{name}'");
            return Ok(ExitCode::from(1));
        }
        (None, None) => {
            // Do nothing, fall through.
        }
    };

    // Step 2: Prepare the command to execute, which may be a login shell.
    if argv.is_empty() {
        argv.extend(opts.positionals().map(|s| s.to_string()));
    } else {
        // -c command args...
        argv.push(opts.positionals().collect::<Vec<_>>().join(" "));
    };
    let mut argv = argv.into_iter();
    let argv0 = match (export_mode, argv.next(), is_login) {
        (Some(_), _, false) => "true".to_string(),
        (_, Some(ref cmd), false) if cmd == "log" => {
            Command::new("journalctl")
                .args(["SYSLOG_IDENTIFIER=syd"])
                .exec();
            return Ok(ExitCode::from(1));
        }
        (_, Some(ref cmd), false) if cmd == "exec" => {
            // Split the arguments using the ASCII Unit Separator character
            let args = argv.collect::<Vec<_>>().join("\x1F");

            // Format it using /dev/syd/cmd/exec!<concatenated-path>
            let path = format!("/dev/syd/cmd/exec!{args}");
            print!("{path}");
            return Ok(ExitCode::from(0));
        }
        (_, Some(argv0), false) => argv0,
        (_, None, false) | (_, _, true) => {
            sandbox
                .parse_profile("user")
                .context("Unable to parse user profile!")?;
            arg0 = Some(
                Path::new(&sh)
                    .file_name()
                    .map(|n| format!("-{}", n.to_string_lossy()))
                    .unwrap_or("-".to_string()),
            );
            sh.to_string()
        }
    };

    // Step 3: Set up Linux namespaces if requested. Note,
    // we set it up here before spawning the child so as to
    // include the SydB☮x process into the pid namespace as well
    // such that the sandbox process and SydB☮x have the identical
    // view of /proc.
    let mut namespaces = CloneFlags::empty();
    for namespace in sandbox.namespaces() {
        namespaces |= syd::unshare::namespace::to_clone_flag(namespace);
    }
    if namespaces.contains(CloneFlags::CLONE_NEWPID) {
        // SAFETY: We cannot support NEWPID without NEWNS.
        // ie, pid namespace must have its own private /proc.
        namespaces.insert(CloneFlags::CLONE_NEWNS);
    }
    if init && !namespaces.is_empty() {
        let uid = getuid().as_raw();
        let gid = getgid().as_raw();

        unshare(namespaces).context("Failed to create namespace")?;

        // SAFETY: In libc we trust.
        match unsafe { fork() }.context("Failed to fork")? {
            ForkResult::Parent { child, .. } => loop {
                // SAFETY: Set up a Landlock sandbox to disallow all access.
                let _ = landlock_operation(&[], &[], LANDLOCK_ABI);

                // SAFETY: Set up a seccomp filter which only allows
                // 1. read, write and exit
                // 2. wait4 and waitpid
                // 3. memory allocation functions
                let mut filter = ScmpFilterContext::new_filter(ScmpAction::KillProcess)
                    .context("Failed to create seccomp filter")?;
                let allow_call = [
                    "read",
                    "write",
                    "exit",
                    "exit_group",
                    "wait4",
                    "waitpid",
                    "brk",
                    "madvise",
                    "mmap",
                    "mmap2",
                    "munmap",
                    "sigaltstack",
                ];
                for call in allow_call {
                    filter.add_rule(ScmpAction::Allow, ScmpSyscall::new(call))?;
                }
                filter.load().context("Failed to load seccomp filter")?;

                match waitpid(child, None) {
                    Ok(WaitStatus::Exited(_, code)) => {
                        #[allow(clippy::cast_possible_truncation)]
                        #[allow(clippy::cast_sign_loss)]
                        return Ok(ExitCode::from(code as u8));
                    }
                    Ok(WaitStatus::StillAlive) | Err(Errno::EINTR) => {}
                    Ok(_) => {
                        return Ok(ExitCode::from(127));
                    }
                    Err(error) => {
                        bail!("Failed to wait: {error}");
                    }
                }
            },
            ForkResult::Child => {
                if namespaces.contains(CloneFlags::CLONE_NEWUSER) {
                    // create the UID and GID mappings.
                    let uid_maps = vec![
                        UidMap {
                            inside_uid: if map_root { 0 } else { uid },
                            outside_uid: uid,
                            count: 1,
                        }, // Map the current user.
                    ];
                    let gid_maps = vec![
                        GidMap {
                            inside_gid: if map_root { 0 } else { gid },
                            outside_gid: gid,
                            count: 1,
                        }, // Map the current group.
                    ];

                    // Write "deny" to /proc/self/setgroups before writing to gid_map.
                    File::create("/proc/self/setgroups")
                        .and_then(|mut f| f.write_all(b"deny"))
                        .context("Failed to write to /proc/self/setgroups")?;

                    let mut buf = Vec::new();
                    for map in gid_maps {
                        writeln!(
                            &mut buf,
                            "{} {} {}",
                            map.inside_gid, map.outside_gid, map.count
                        )
                        .context("Failed to format gidmap")?;
                    }
                    File::create("/proc/self/gid_map")
                        .and_then(|mut f| f.write_all(&buf[..]))
                        .context("Failed to write to /proc/self/gid_map")?;

                    let mut buf = Vec::new();
                    for map in uid_maps {
                        writeln!(
                            &mut buf,
                            "{} {} {}",
                            map.inside_uid, map.outside_uid, map.count
                        )
                        .context("Failed to format uidmap")?;
                    }
                    File::create("/proc/self/uid_map")
                        .and_then(|mut f| f.write_all(&buf[..]))
                        .context("Failed to write to /proc/self/uid_map")?;
                }

                if namespaces.contains(CloneFlags::CLONE_NEWUTS) {
                    let _ = sethostname("SydB☮x");
                    // SAFETY: API_VERSION must not contain a nul-byte.
                    #[allow(clippy::disallowed_methods)]
                    let d = CString::new(API_VERSION.to_string()).unwrap();
                    // SAFETY: There's no setdomainname wrapper in nix.
                    let _ = unsafe { nix::libc::setdomainname(d.as_ptr() as *const _, d.len()) };
                }

                // Set mount propagation on the root filesystem for mount namespace.
                if namespaces.contains(CloneFlags::CLONE_NEWNS) {
                    const NONE: Option<&PathBuf> = None::<PathBuf>.as_ref();
                    if let Some(flags) = sandbox.propagation {
                        if !flags.is_empty() {
                            mount(Some("none"), "/", NONE, flags, NONE)
                                .context("Failed to change root filesystem propagation")?;
                        }
                    }

                    if let Some(ref proc) = sandbox.proc {
                        // When not changing root and using default propagation flags
                        // then the recursive propagation change of root will
                        // automatically change that of an existing proc mount.
                        if let Some(flags) = sandbox.propagation {
                            if !flags.is_empty() && flags != MsFlags::MS_PRIVATE | MsFlags::MS_REC {
                                let flags = MsFlags::MS_PRIVATE | MsFlags::MS_REC;
                                match mount(Some("none"), proc, NONE, flags, NONE) {
                                    Ok(_) => {}
                                    Err(Errno::EINVAL) => {
                                        // Custom proc mount means that proc is very likely not mounted,
                                        // causing EINVAL. Ignoring the error in this specific instance
                                        // is considered safe.
                                    }
                                    Err(error) => {
                                        bail!(
                                            "Failed to change proc filesystem propagation: {error}"
                                        )
                                    }
                                }
                            }
                        }

                        let flags = MsFlags::MS_NOSUID | MsFlags::MS_NOEXEC | MsFlags::MS_NODEV;
                        mount(Some("proc"), proc, Some("proc"), flags, NONE)
                            .context("Failed to mount proc filesystem")?;
                    }
                }

                if namespaces.contains(CloneFlags::CLONE_NEWNET) {
                    // Set up the loopback interface.
                    if let Err(errno) = bring_up_loopback() {
                        error!("ctx": "ifup_lo", "err": errno.to_string());
                    }
                }

                // reexecute SydB☮x asserting initialization is done.
                let args: Vec<String> = env::args().skip(1).collect();
                std::process::Command::new(
                    env::current_exe().context("Failed to get current executable")?,
                )
                .args(&args)
                .env(ENV_INIT, "1")
                .exec();
                return Ok(ExitCode::from(127));
            }
        }
    }

    // Step 4: Change root directory if requested.
    if let Some(root) = root {
        let proc = root.join("proc");
        if !proc.is_dir() {
            bail!("Directory proc does not exist under root directory!");
        }
        mount(Some(root), root, NONE, MsFlags::MS_BIND, NONE)
            .context("Failed to bind mount the new root directory")?;

        mount(Some("/proc"), &proc, NONE, MsFlags::MS_BIND, NONE)
            .context("Failed to bind mount the /proc directory")?;

        chdir(root).context("Failed to change to root directory")?;
        pivot_root(".", "./proc").context("Failed to pivot root")?;
        umount2("/proc", MntFlags::MNT_DETACH).context("Failed to unmount old root")?;
        chdir("/").context("Failed to change to new root directory")?;

        chroot(".").context("Failed to change root directory")?;
    }

    // Step 5: Initialize sandbox environment, drop capabilities.
    Supervisor::init_env()?;

    // Step 6: Set up the Landlock sandbox if requested. Note,
    // we set it up here before spawning the child so as to
    // include the SydB☮x sandbox threads into the sandbox as
    // well. This is done for added security.
    // Note, Landlock errors are not fatal.
    if let Some((mut path_ro, mut path_rw)) = sandbox.collect_landlock() {
        path_ro.sort();
        path_rw.sort();

        match landlock_operation(&path_ro, &path_rw, LANDLOCK_ABI) {
            Ok(status) => {
                info!("ctx": "landlock",
                "abi": LANDLOCK_ABI_STR,
                "path_ro": path_ro,
                "path_rw": path_rw,
                "status": match status.ruleset {
                    // The FullyEnforced case must be tested by the developer.
                    RulesetStatus::FullyEnforced => "fully_enforced",
                    RulesetStatus::PartiallyEnforced => "partially_enforced",
                    // Users should be warned that they are not protected.
                    RulesetStatus::NotEnforced => "not_enforced",
                });
            }
            Err(error) => {
                info!("ctx": "landlock",
                    "abi": LANDLOCK_ABI_STR,
                    "path_ro": path_ro,
                    "path_rw": path_rw,
                    "status": "unsupported",
                    "err": error.to_string());
            }
        }
    }

    // Step 7: Initialize sandbox supervisor.
    sandbox
        .build_globsets()
        .context("Failed to build GlobSets.")?;
    let wait_all = sandbox.exit_wait_all();
    let mut supervisor = Supervisor::new(sandbox, *syd::NPROC, export_mode).context(format!(
        "Error creating sandbox with {} threads and export mode {:?}.",
        *syd::NPROC,
        export_mode,
    ))?;
    supervisor.init();

    let mut command = syd::unshare::Command::new(argv0);
    command.args(argv);
    if let Some(arg0) = arg0 {
        command.arg0(arg0);
    }

    // Step 8: Spawn the program under sandbox.
    let (pid, thread_handle, pool) = match supervisor.spawn(command) {
        Ok((pid, thread_handle, pool)) => (pid, thread_handle, pool),
        Err(error) => {
            let errno = Errno::last() as u8;
            let error = error.context("Failed to spawn command under sandbox.");
            eprintln!("{error:?}");
            return Ok(ExitCode::from(errno));
        }
    };

    // Step 9: Wait for the process to exit and return the same error code.
    let code = match Supervisor::wait(pid, thread_handle, pool, wait_all) {
        Ok(code) => code,
        Err(errno) => {
            let error = Error::new(errno).context("Failed to wait for sandboxed process.");
            eprintln!("{error:?}");
            errno as i32
        }
    };
    Ok(ExitCode::from(
        u8::try_from(code).context("Invalid exit code!")?,
    ))
}
