//
// SydB☮x: seccomp and landlock based application sandbox with support for namespaces
// src/proc.rs: /proc utilities
//
// Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

use std::{
    ffi::OsStr,
    fs::{read_dir, read_link, File},
    io::{BufRead, BufReader, Read},
    os::unix::ffi::OsStrExt,
    path::PathBuf,
};

use nix::{errno::Errno, unistd::Pid};
use once_cell::sync::Lazy;
use procfs::process::{MMapPath, Process};
use regex::bytes::Regex;
use smallvec::SmallVec;

static RE_PROC_TGID: Lazy<Regex> = Lazy::new(|| {
    #[allow(clippy::disallowed_methods)]
    Regex::new(r"Tgid:[[:space:]]*([1-9][0-9]*)")
        .expect("invalid proc tgid regex, please report a bug!")
});

/// Retrieves the thread group ID (TGID) for the specified process ID (PID).
///
/// This function reads the `/proc/<pid>/status` file and extracts the
/// TGID from the "Tgid:" line.
///
/// # Arguments
///
/// * `pid` - The process ID for which to retrieve the thread group ID.
///
/// # Returns
///
/// * `Ok(Pid)` - The thread group ID as a `Pid` type if found.
/// * `Err(Errno)` - An error, typically:
///     * `Errno::ENOENT` if the "Tgid:" line is not found in the status file.
///     * `Errno::EINVAL` if the TGID is found but cannot be parsed as an integer.
///
/// # Examples
///
/// ```rust
/// use nix::unistd::Pid;
/// use syd::proc::proc_tgid;
///
/// let tgid = proc_tgid(Pid::this());
/// assert!(tgid.is_ok());
/// ```
pub fn proc_tgid(pid: Pid) -> Result<Pid, Errno> {
    let mut file = File::open(format!("/proc/{pid}/status")).map_err(|_| Errno::last())?;

    let mut buffer = [0; 256];
    let bytes_read = file.read(&mut buffer).map_err(|_| Errno::last())?;

    if let Some(caps) = RE_PROC_TGID.captures(&buffer[..bytes_read]) {
        if let Some(m) = caps.get(1) {
            return parse_bytes_to_pid(m.as_bytes());
        }
    }

    Err(Errno::ESRCH)
}

fn parse_bytes_to_pid(bytes: &[u8]) -> Result<Pid, Errno> {
    let mut tgid = 0;
    for &b in bytes {
        #[allow(clippy::arithmetic_side_effects)]
        #[allow(clippy::cast_lossless)]
        if b.is_ascii_digit() {
            tgid = tgid * 10 + (b - b'0') as nix::libc::pid_t;
        } else {
            return Err(Errno::EINVAL);
        }
    }
    Ok(Pid::from_raw(tgid))
}

/// Retrieves the command line of the specified process ID (PID)
/// concatenated as a single string.
///
/// This function reads the `/proc/<pid>/cmdline` file and concatenates
/// the arguments using spaces. The function takes care of replacing null
/// bytes (`'\0'`) with spaces to format the command line as a readable string.
///
/// # Arguments
///
/// * `pid` - The process ID for which to retrieve the command line.
///
/// # Returns
///
/// * `Ok(String)` - The command line of the process as a single string.
/// * `Err(Errno)` - An error, depending on the issue encountered while reading
///                  or processing the cmdline file.
///
/// # Examples
///
/// ```rust
/// use nix::unistd::Pid;
/// use syd::proc::proc_cmdline;
///
/// let cmdline = proc_cmdline(Pid::this());
/// assert!(cmdline.is_ok());
/// ```
#[cfg(feature = "log")]
pub fn proc_cmdline(pid: Pid) -> Result<String, Errno> {
    const LIMIT: usize = 256;
    let path = format!("/proc/{pid}/cmdline");
    let file = File::open(path).map_err(|_| Errno::last())?;

    let mut data = Vec::with_capacity(LIMIT + 1 /* sentinel */);
    let reader = BufReader::new(file);

    // Read up to LIMIT+1 characters or until EOF
    for byte in reader.bytes().take(LIMIT + 1 /* sentinel */) {
        let byte = byte.map_err(|_| Errno::last())?;
        if byte == 0 {
            // Replace null bytes with spaces
            data.push(b' ');
        } else {
            data.push(byte);
        }
    }

    let cmdline = String::from_utf8_lossy(&data);
    let cmdline = if cmdline.len() > LIMIT {
        let cutoff = cmdline[..LIMIT].rfind(' ').unwrap_or(LIMIT);
        format!("{}…", &cmdline[..cutoff].trim_end_matches(' '))
    } else {
        cmdline.trim_end_matches(' ').to_string()
    };

    Ok(cmdline)
}

/// Retrieves the command name (comm) of the specified process ID (PID)
/// as a single string.
///
/// This function reads the `/proc/<pid>/comm` file. It reads up to 16 characters,
/// which is typically sufficient for process names.
///
/// # Arguments
///
/// * `pid` - The process ID for which to retrieve the command name.
///
/// # Returns
///
/// * `Ok(String)` - The command name of the process as a single string.
/// * `Err(Errno)` - An error, depending on the issue encountered while reading
///                  the comm file.
///
/// # Examples
///
/// ```rust
/// use nix::unistd::Pid;
/// use syd::proc::proc_comm;
///
/// let comm = proc_comm(Pid::this());
/// assert!(comm.is_ok());
/// ```
#[cfg(not(feature = "log"))]
pub fn proc_comm(pid: Pid) -> Result<String, Errno> {
    const LIMIT: usize = 16;
    let mut buf = itoa::Buffer::new();
    let mut path = PathBuf::from("/proc");
    path.push(buf.format(pid.as_raw()));
    path.push("comm");

    let mut file = File::open(path).map_err(|_| Errno::last())?;
    let mut data = Vec::with_capacity(LIMIT);

    // Read up to LIMIT characters or until EOF
    file.read_to_end(&mut data).map_err(|_| Errno::last())?;

    let comm = String::from_utf8_lossy(&data);
    Ok(comm.trim_end().to_string())
}

/// Retrieves the current working directory (CWD) of the specified process ID (PID).
///
/// This function reads the symbolic link `/proc/<pid>/cwd` to determine the CWD.
///
/// # Arguments
///
/// * `pid` - The process ID for which to retrieve the current working directory.
///
/// # Returns
///
/// * `Ok(PathBuf)` - The current working directory of the process.
/// * `Err(Errno)` - An error, depending on the issue encountered while reading
///                  or resolving the `cwd` symlink.
///
/// # Examples
///
/// ```rust
/// use nix::unistd::Pid;
/// use syd::proc::proc_cwd;
///
/// let cwd = proc_cwd(Pid::this());
/// assert!(cwd.is_ok());
/// ```
pub fn proc_cwd(pid: Pid) -> Result<PathBuf, Errno> {
    let mut buf = itoa::Buffer::new();
    let mut path = PathBuf::from("/proc");
    path.push(buf.format(pid.as_raw()));
    path.push("cwd");
    read_link(path)
        .map(|p| p.to_path_buf())
        .map_err(|_| Errno::last())
}

/// Reads the tty number from /proc/[pid]/stat and figures out the corresponding /dev/tty device node path.
pub fn proc_tty(pid: Pid) -> Result<PathBuf, Errno> {
    let mut buf = itoa::Buffer::new();
    let mut path = PathBuf::from("/proc");
    path.push(buf.format(pid.as_raw()));
    path.push("stat");
    let file = File::open(path).map_err(|_| Errno::ENOENT)?;
    let reader = BufReader::new(file);

    // Read the tty_nr (terminal number) from the stat file.
    let tty_nr = reader.lines()
        .next()
        .ok_or(Errno::ENOENT)?
        .map_err(|_| Errno::EINVAL)?
        .split_whitespace()
        .nth(6) // tty_nr is typically the 7th field in the /proc/[pid]/stat file.
        .ok_or(Errno::EINVAL)?
        .parse::<i32>()
        .map_err(|_| Errno::EINVAL)?;

    // Convert tty_nr to the corresponding /dev/tty device node path.
    if tty_nr <= 0 {
        // Process has no controlling terminal
        Err(Errno::ENXIO)
    } else {
        // Construct the path based on the major and minor device numbers.
        let major = (tty_nr >> 8) & 0xfff;
        let minor = (tty_nr & 0xff) | ((tty_nr >> 12) & 0xfff00);

        // Determine the correct device path.
        match major {
            // Unix 98 PTYs (e.g., /dev/pts/N)
            136 => {
                let mut buf = itoa::Buffer::new();
                let mut path = PathBuf::from("/dev/pts");
                path.push(buf.format(minor));
                Ok(path)
            }
            // Standard TTYs (e.g., /dev/ttyN)
            14 => {
                let mut buf = itoa::Buffer::new();
                Ok(PathBuf::from("/dev/tty".to_owned() + buf.format(minor)))
            }
            // Other cases: unknown or unsupported major number
            _ => Err(Errno::ENXIO),
        }
    }
}

/// Checks if the number of tasks across all processes in the system
/// exceeds the given limit by inspecting the `/proc` filesystem.
///
/// This function iterates over each entry in the `/proc` directory,
/// assuming each numeric directory name represents a process ID. For
/// each process, it counts the number of tasks (threads) by counting
/// entries in the `/proc/[pid]/task` directory. The function skips
/// over the initial entry `/proc/1`.
///
/// # Returns
/// `Ok(true)`: Number of tasks reach the given limit.
/// `Err(Errno)`: An error occurred during the operation, excluding
/// ENOENT (file not found), EACCES (access denied), EPERM (permission
/// denied), and ESRCH (process not found) errors, which are ignored.
///
/// # Errors
/// This function returns an error for any issue encountered while
/// reading the `/proc` directory or task subdirectories, except for
/// ENOENT, EACCES, EPERM, and ESRCH errors, which are ignored. The
/// error is wrapped in an `Errno` type.
pub fn proc_task_limit(max: usize) -> Result<bool, Errno> {
    let proc_entries = read_dir("/proc")
        .map_err(|e| Errno::from_i32(e.raw_os_error().unwrap_or(nix::libc::EINVAL)))?;

    let result = proc_entries
        .filter_map(Result::ok)
        .filter(|entry| entry.file_name().as_bytes() != b"1" && is_numeric(&entry.file_name()))
        .try_fold(
            (0usize, SmallVec::<[PathBuf; 256]>::new()),
            |(count, mut entries), entry| {
                let count = count.saturating_add(1);
                if count >= max {
                    // Early termination with Err to break out of the try_fold
                    return Err(());
                }
                entries.push(entry.path());
                Ok((count, entries))
            },
        );

    // Unwrap the result of try_fold, handling early termination
    let (mut pid_count, eligible_entries) = match result {
        Ok(data) => data,
        Err(_) => return Ok(true),
    };

    // Step 2: Count tasks in each PID
    for path in eligible_entries {
        match read_dir(&path.join("task")) {
            Ok(tasks) => {
                pid_count = pid_count.saturating_add(tasks.count()).saturating_sub(1);
                if pid_count >= max {
                    return Ok(true);
                }
            }
            Err(error) => match error.raw_os_error() {
                Some(
                    nix::libc::ENOENT | nix::libc::EPERM | nix::libc::EACCES | nix::libc::ESRCH,
                ) => {}
                Some(error) => return Err(Errno::from_i32(error)),
                None => return Err(Errno::EINVAL),
            },
        }
    }

    Ok(false)
}

#[inline]
fn is_numeric(name: &OsStr) -> bool {
    name.as_bytes().iter().all(|&b| b.is_ascii_digit())
}

/// Checks whether process memory usage is within the give maximum.
///
/// This function uses the `procfs` crate to obtain detailed memory maps
/// from `/proc/[pid]/smaps`. It sums multiple memory usage values reported in these maps
/// to calculate a more comprehensive total memory usage.
///
/// # Arguments
///
/// * `process` - `Process` instance representing the process.
///
/// # Returns
///
/// This function returns a `Result<bool, Errno>`. It returns Ok(true)
/// if the limit was exceeded Ok(false) otherwise. On failure, it
/// returns `Errno`.
///
/// # Errors
///
/// This function returns an error if it fails to retrieve the process's memory maps,
/// typically due to insufficient permissions or an invalid process ID.
pub fn proc_mem_limit(process: &Process, max: u64) -> Result<bool, Errno> {
    match process.smaps() {
        Ok(maps) => {
            let mut total_size: u64 = 0;
            for map in &maps.memory_maps {
                match &map.pathname {
                    MMapPath::Path(_)
                    | MMapPath::Anonymous
                    | MMapPath::Stack
                    | MMapPath::Other(_) => {
                        let pss = map.extension.map.get("Pss").copied().unwrap_or(0);
                        let private_dirty =
                            map.extension.map.get("Private_Dirty").copied().unwrap_or(0);
                        let shared_dirty =
                            map.extension.map.get("Shared_Dirty").copied().unwrap_or(0);

                        total_size = total_size.saturating_add(
                            pss.saturating_add(private_dirty)
                                .saturating_add(shared_dirty),
                        );

                        // Stop processing if total size exceeds or equals max
                        if total_size >= max {
                            return Ok(true);
                        }
                    }
                    _ => (),
                }
            }

            // If we're at this point, we did not hit the limit.
            Ok(false)
        }
        Err(_) => Err(Errno::last()),
    }
}
