#![allow(clippy::undocumented_unsafe_blocks)]

use std::{mem, ptr};

use nix::{
    self,
    errno::Errno,
    libc::{self, signal, sigset_t, SIG_DFL, SIG_SETMASK},
    unistd::close,
};
use sendfd::SendWithFd;

use crate::unshare::{error::ErrorCode as Err, run::ChildInfo};

unsafe fn fail(code: Err) -> ! {
    fail_errno(code, nix::errno::errno())
}
unsafe fn fail_errno(code: Err, errno: i32) -> ! {
    let errnox = Errno::from_i32(errno);
    eprintln!("exec error: {code:?} errno:{errnox}");
    libc::_exit(errno);
}

macro_rules! fail_safe {
    ($child:expr, $error:expr) => {
        drop($child);
        unsafe { fail($error) };
    };
}

macro_rules! fail_errno_safe {
    ($child:expr, $error:expr, $errno:expr) => {
        drop($child);
        unsafe { fail_errno($error, $errno) };
    };
}

#[allow(clippy::cognitive_complexity)]
pub fn child_after_clone(child: ChildInfo) -> ! {
    if let Some(&sig) = child.cfg.death_sig.as_ref() {
        if unsafe { libc::prctl(libc::PR_SET_PDEATHSIG, sig as libc::c_ulong, 0, 0, 0) } != 0 {
            fail_safe!(child, Err::ParentDeathSignal);
        }
    }

    if child.cfg.restore_sigmask {
        unsafe {
            let mut sigmask: sigset_t = mem::zeroed();
            libc::sigemptyset(&mut sigmask);
            libc::pthread_sigmask(SIG_SETMASK, &sigmask, ptr::null_mut());
            for sig in 1..32 {
                signal(sig, SIG_DFL);
            }
        }
    }

    if let Some(callback) = &child.pre_exec {
        if let Err(e) = callback() {
            fail_errno_safe!(child, Err::PreExec, e as i32);
        }
    }

    if let Some(seccomp_filter) = &child.seccomp_filter {
        if seccomp_filter.load().is_err() {
            fail_safe!(child, Err::Seccomp);
        } else if let Some(seccomp_sender) = &child.seccomp_sender {
            let fd = match seccomp_filter.get_notify_fd() {
                Ok(fd) => fd,
                Err(_) => {
                    fail_safe!(child, Err::Seccomp);
                }
            };
            if seccomp_sender.send_with_fd(&[0; 1], &[fd]).is_err() {
                fail_safe!(child, Err::SeccompNotify);
            }
            let _ = close(fd);
        }
    }

    unsafe { libc::execvp(child.filename, child.args.as_ptr()) };
    fail_safe!(child, Err::Exec);
}
