use std::{
    ffi::CString,
    os::unix::{io::RawFd, net::UnixStream},
    ptr,
};

use libseccomp::ScmpFilterContext;
use nix::{
    self,
    errno::Errno,
    libc::c_char,
    sys::wait::{waitpid, WaitStatus},
    unistd::{fork, setpgid, ForkResult, Pid},
};
use sendfd::RecvWithFd;

use crate::unshare::{child, config::Config, Child, Command};

type ChildPreExecFunc = Box<dyn Fn() -> Result<(), Errno>>;

pub struct ChildInfo<'a> {
    pub filename: *const c_char,
    pub args: &'a [*const c_char],
    pub cfg: Config,
    pub pre_exec: Option<ChildPreExecFunc>,
    pub seccomp_filter: Option<ScmpFilterContext>,
    pub seccomp_sender: Option<UnixStream>,
}

fn raw_with_null(arr: &Vec<CString>) -> Vec<*const c_char> {
    let mut vec = Vec::with_capacity(arr.len().saturating_add(1));
    for i in arr {
        vec.push(i.as_ptr());
    }
    vec.push(ptr::null());
    vec
}

impl Command {
    /// Spawn the command and return a handle that can be waited for
    pub fn spawn(self) -> Result<Child, Errno> {
        let c_args = raw_with_null(&self.args);

        // Clear the environment.
        std::env::remove_var("SYD_INIT");
        std::env::remove_var(crate::config::ENV_LOG);
        std::env::remove_var(crate::config::ENV_NO_SYSLOG);
        std::env::remove_var(crate::config::ENV_NO_CROSS_MEMORY_ATTACH);
        std::env::remove_var(crate::config::ENV_FAKEROOT);
        std::env::remove_var(crate::config::ENV_SH);
        std::env::remove_var(crate::config::ENV_UNSHARE_MOUNT);
        std::env::remove_var(crate::config::ENV_UNSHARE_UTS);
        std::env::remove_var(crate::config::ENV_UNSHARE_IPC);
        std::env::remove_var(crate::config::ENV_UNSHARE_USER);
        std::env::remove_var(crate::config::ENV_UNSHARE_PID);
        std::env::remove_var(crate::config::ENV_UNSHARE_NET);
        std::env::remove_var(crate::config::ENV_UNSHARE_CGROUP);

        // SAFETY: In libc we trust.
        match unsafe { fork() }? {
            ForkResult::Parent { child, .. } => {
                let seccomp_fd = match self.after_start(child) {
                    Ok(seccomp_fd) => seccomp_fd,
                    Err(e) => loop {
                        match waitpid(child, None) {
                            Ok(WaitStatus::Exited(_, errno)) => return Err(Errno::from_i32(errno)),
                            Err(Errno::EINTR) => {}
                            _ => return Err(e),
                        }
                    },
                };

                Ok(Child {
                    pid: child.into(),
                    status: None,
                    seccomp_fd,
                })
            }
            ForkResult::Child => {
                let mut child_info = ChildInfo {
                    filename: self.filename.as_ptr(),
                    args: &c_args[..],
                    cfg: self.config,
                    pre_exec: self.pre_exec,
                    seccomp_filter: self.seccomp_filter,
                    seccomp_sender: None,
                };
                if let Some((sender, receiver)) = self.seccomp_channel {
                    drop(receiver);
                    child_info.seccomp_sender = Some(sender);
                }
                child::child_after_clone(child_info);
            }
        }
    }

    #[allow(clippy::cognitive_complexity)]
    fn after_start(mut self, pid: Pid) -> Result<Option<RawFd>, Errno> {
        // Receive the seccomp notification file descriptor.
        let seccomp_fd = if let Some((sender, receiver)) = self.seccomp_channel {
            drop(sender);
            let mut buf = [0; 1];
            let mut fds = [0; 1];
            match receiver.recv_with_fd(&mut buf, &mut fds) {
                Ok(_) => {
                    drop(receiver);
                    Some(fds[0])
                }
                Err(error) => {
                    return Err(Errno::from_i32(
                        error.raw_os_error().unwrap_or(nix::libc::EINVAL),
                    ));
                }
            }
        } else {
            None
        };

        if self.config.make_group_leader {
            setpgid(pid, pid)?;
        }

        if let Some(ref mut callback) = self.before_unfreeze {
            #[allow(clippy::cast_sign_loss)]
            callback(i32::from(pid) as u32)?;
        }

        Ok(seccomp_fd)
    }
}
