//
// SydB☮x: seccomp and landlock based application sandbox with support for namespaces
// src/hook.rs: Secure computing hooks
//
// Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
// Based in part upon greenhook which is under public domain.
//
// SPDX-License-Identifier: GPL-3.0-or-later

use std::{
    collections::{HashMap, HashSet},
    ffi::{CStr, CString, OsStr, OsString},
    fs::{File, OpenOptions},
    io::{self, BufReader, IoSlice, IoSliceMut, Read, Result as IOResult, Seek, SeekFrom, Write},
    mem::MaybeUninit,
    net::{IpAddr, Ipv4Addr},
    os::{
        fd::{AsRawFd, FromRawFd, OwnedFd, RawFd},
        unix::ffi::{OsStrExt, OsStringExt},
    },
    path::{Path, PathBuf},
    process::exit,
    str::FromStr,
    sync::Arc,
    thread::{Builder, JoinHandle},
};

use anyhow::{bail, Context};
use globset::{Glob, GlobMatcher};
use libseccomp::{
    ScmpAction, ScmpArgCompare, ScmpCompareOp, ScmpFilterContext, ScmpNotifReq, ScmpNotifResp,
    ScmpNotifRespFlags, ScmpSyscall, ScmpVersion,
};
use log::{log_enabled, Level};
use nix::{
    errno::Errno,
    fcntl::{open, openat, renameat, AtFlags, OFlag},
    libc::ioctl,
    mount::{mount, umount, umount2, MntFlags, MsFlags},
    sched::{clone, CloneFlags},
    sys::{
        epoll::{
            epoll_create1, epoll_ctl, epoll_wait, EpollCreateFlags, EpollEvent, EpollFlags, EpollOp,
        },
        signal::{kill, SIGCHLD, SIGHUP, SIGKILL, SIGTSTP, SIGTTIN, SIGTTOU},
        socket::{
            bind, connect, getsockname, recvfrom, sendto, AddressFamily, MsgFlags, SockaddrLike,
            SockaddrStorage,
        },
        stat::{
            fchmod, fchmodat, fstat, fstatat, lstat, mkdirat, mknodat, umask, FchmodatFlags, Mode,
            SFlag,
        },
        uio::{process_vm_readv, process_vm_writev, RemoteIoVec},
        utsname::uname,
        wait::{waitpid, WaitPidFlag, WaitStatus},
    },
    unistd::{
        access, close, faccessat, fchown, fchownat, ftruncate, linkat, lseek, mkstemp, symlinkat,
        sysconf, truncate, unlink, unlinkat, write, AccessFlags, FchownatFlags, Gid, LinkatFlags,
        Pid, SysconfVar, Uid, UnlinkatFlags, Whence,
    },
    NixPath,
};
use nonempty::{nonempty, NonEmpty};
use once_cell::sync::Lazy;
use parking_lot::RwLock;
use procfs::process::Process;
use regex::bytes::{Regex, RegexBuilder};
use rusty_pool::ThreadPool;
use serde::{ser::SerializeMap, Serialize};

use crate::{
    compat::getdents,
    config::*,
    debug, error,
    fs::normalize_path,
    info,
    log::JsonLinesLogger,
    oflag_fmt, parse_fd,
    proc::{proc_mem_limit, proc_task_limit, proc_tgid, proc_tty},
    proc_umask, replace_self_with_pid,
    sandbox::{Action, Capability, Sandbox, SandboxGuard},
    seccomp_add_architectures, set_io_priority_idle, trace, warn, IoctlRequest,
};

/// Match the special file paths under /proc
/// 1. /proc/pid/fd/pipe:[100]
/// 2. /proc/pid/fd/socket:[100]
/// 3. /proc/pid/fd/anon_inode:seccomp notify
/// 4. /proc/pid/fd/anon_inode:[eventpoll]
/// 5. /proc/pid/fd/foo (deleted)
/// 6. /proc/pid/task/pid/fd/{1..5}
static GLOB_PROC: Lazy<GlobMatcher> = Lazy::new(|| {
    #[allow(clippy::disallowed_methods)]
    Glob::new(r"{* (deleted),/proc/[0-9]*/fd/[!:]*:[!:]*,/proc/[0-9]*/task/[0-9]*/fd/[!:]*:[!:]*}")
        .expect("Invalid proc glob, please file a bug!")
        .compile_matcher()
});

/// Match the special file paths /dev/tty* /dev/pts/*
static GLOB_TTY: Lazy<GlobMatcher> = Lazy::new(|| {
    #[allow(clippy::disallowed_methods)]
    Glob::new(r"/dev/tty*")
        .expect("Invalid dev/tty glob, please file a bug!")
        .compile_matcher()
});

/// Match /proc/pid/fd/n and /proc/pid/task/pid/fd/n
static RE_PROC_FD: Lazy<Regex> = Lazy::new(|| {
    #[allow(clippy::disallowed_methods)]
    RegexBuilder::new(
        r"
        \A
        /proc/(?P<pid>[1-9][0-9]*)/
            (task/(?P<tid>[1-9][0-9]*)/)?
        fd/(?P<fd>[0-9]+)
        \z
    ",
    )
    .ignore_whitespace(true)
    .build()
    .expect("Invalid proc regex, please file a bug!")
});

/*
 * Seccomp constants
 */

const SECCOMP_IOCTL_NOTIF_ADDFD: IoctlRequest = 0x40182103;
/*
const SECCOMP_IOCTL_NOTIF_ID_VALID: IoctlRequest = 0x40082102;
const SECCOMP_IOCTL_NOTIF_RECV: IoctlRequest = 0xc0502100;
const SECCOMP_IOCTL_NOTIF_SEND: IoctlRequest = 0xc0182101;
*/

/*
 * Thread local variables
 */
thread_local! {
    static PRIORITY_SET: std::cell::Cell<bool> = std::cell::Cell::new(false);
}

/*
 * Utilities
 */

/// Converts a reference to `T` into a slice of bytes.
///
/// # Safety
///
/// This function is unsafe because it may produce a slice that does not
/// respect the original alignment requirements of `T`. The caller must ensure
/// that the resultant slice of bytes is only used in ways that are consistent
/// with the memory layout of `T`.
///
/// It's also crucial to ensure that the lifetime of the returned slice does not
/// exceed the lifetime of the input reference, `p`.
unsafe fn any_as_u8_slice<T: Sized>(p: &T) -> &[u8] {
    ::core::slice::from_raw_parts((p as *const T) as *const u8, ::core::mem::size_of::<T>())
}

/// Converts a mutable reference to `T` into a mutable slice of bytes.
///
/// # Safety
///
/// This function is unsafe for the same reasons as `any_as_u8_slice`. It can
/// produce a slice that does not respect the original alignment requirements
/// of `T`. The caller must ensure the resultant slice of bytes is only used in
/// ways that are consistent with the memory layout of `T`.
///
/// Also, the caller should ensure that no other references to `p` exist when
/// this function is called, as this could lead to aliasing mutable references,
/// which is undefined behavior in Rust.
unsafe fn any_as_u8_mut_slice<T: Sized>(p: &mut T) -> &mut [u8] {
    ::core::slice::from_raw_parts_mut((p as *mut T) as *mut u8, ::core::mem::size_of::<T>())
}

/*
 * Macros
 */
macro_rules! syscall_handler {
    ($request:expr, $body:expr) => {{
        let inner = |request: &UNotifyEventRequest| -> Result<libseccomp::ScmpNotifResp, Errno> {
            let (req, proc) = request.prepare();
            $body(req, &proc)
        };

        match inner($request) {
            Ok(result) => result,
            Err(error) => $request.fail_syscall(error as i32),
        }
    }};
}

macro_rules! remote_path_n {
    ($remote_process:expr, $request:expr, $n:expr, $event_request:expr) => {
        $remote_process.remote_path($request.data.args[$n] as usize, &$event_request)
    };
}

/// Seccomp sandbox profile export modes.
#[derive(Debug, Clone, Copy, Eq, PartialEq)]
pub enum ExportMode {
    /// Berkeley Packet Filter (binary, machine readable)
    BerkeleyPacketFilter,
    /// Pseudo Filter Code (text, human readable)
    PseudoFiltercode,
}

impl FromStr for ExportMode {
    type Err = anyhow::Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s {
            "bpf" => Ok(Self::BerkeleyPacketFilter),
            "pfc" => Ok(Self::PseudoFiltercode),
            _ => bail!("Invalid export mode, expected one of bpf, or pfc!"),
        }
    }
}

/// `SyscallPathArgument` represents a system call path argument,
/// coupled with a directory file descriptor as necessary.
pub(crate) struct SyscallPathArgument {
    /// DirFd index in syscall args, if applicable.
    pub dirfd: Option<usize>,
    /// Path index in syscall args, if applicable.
    pub path: Option<usize>,
    /// Whether if it's ok for the path to be a NULL pointer.
    pub null: bool,
    /// Whether if it's ok for the path to be empty.
    pub empty: bool,
    /// Whether symbolic links should be resolved.
    pub resolve: bool,
    /// Whether dot as final component must return the given `Errno`.
    pub dotlast: Option<Errno>,
}

impl Serialize for SyscallPathArgument {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        let mut map = serializer.serialize_map(Some(2))?;
        map.serialize_entry("dirfd", &self.dirfd)?;
        map.serialize_entry("path", &self.path)?;
        map.serialize_entry("null", &self.null)?;
        map.serialize_entry("empty", &self.empty)?;
        map.serialize_entry("resolve", &self.resolve)?;
        map.serialize_entry("dotlast", &self.dotlast.map(|e| e.to_string()))?;
        map.end()
    }
}

/// `UNotifyEventRequest` is the type of parameter that user's function
/// would get.
#[derive(Debug)]
pub struct UNotifyEventRequest {
    request: libseccomp::ScmpNotifReq,
    notify_fd: RawFd,
    sandbox: Arc<RwLock<Sandbox>>,
}

type RequestWithProcess<'a> = (&'a ScmpNotifReq, RemoteProcess);

impl UNotifyEventRequest {
    fn new(
        request: libseccomp::ScmpNotifReq,
        notify_fd: RawFd,
        sandbox: Arc<RwLock<Sandbox>>,
    ) -> Self {
        UNotifyEventRequest {
            request,
            notify_fd,
            sandbox,
        }
    }

    /// Given a `UNotifyEventRequest` return the seccomp request and a corresponding `RemoteProcess` instance.
    pub fn prepare(&self) -> RequestWithProcess {
        let req = self.get_request();
        #[allow(clippy::cast_possible_wrap)]
        let pid = Pid::from_raw(req.pid as i32);
        let proc = RemoteProcess::new(pid);
        (req, proc)
    }

    /// Returns the unotify request (`libseccomp::ScmpNotifReq`) of
    /// this event.
    pub fn get_request(&self) -> &libseccomp::ScmpNotifReq {
        &self.request
    }

    /// Returns the internal `Sandbox` object locking it as necessary,
    /// and wrapped in a `SandboxGuard`.
    pub fn get_sandbox(&self, write: bool) -> SandboxGuard {
        if write {
            SandboxGuard::Write(self.sandbox.write())
        } else {
            SandboxGuard::Read(self.sandbox.read())
        }
    }

    /// Let the kernel continue the syscall.
    ///
    /// # Safety
    /// CAUTION! This method is unsafe because it may suffer TOCTOU attack.
    /// Please read `seccomp_unotify(2)` "NOTES/Design goals; use of `SECCOMP_USER_NOTIF_FLAG_CONTINUE`"
    /// before using this method.
    pub unsafe fn continue_syscall(&self) -> libseccomp::ScmpNotifResp {
        libseccomp::ScmpNotifResp::new(self.request.id, 0, 0, ScmpNotifRespFlags::CONTINUE.bits())
    }

    /// Returns error to supervised process.
    /// `err` parameter should be a number larger than 0.
    pub fn fail_syscall(&self, err: i32) -> libseccomp::ScmpNotifResp {
        debug_assert!(err > 0);
        #[allow(clippy::arithmetic_side_effects)]
        libseccomp::ScmpNotifResp::new(self.request.id, 0, -err, 0)
    }

    /// Returns value to supervised process.
    pub fn return_syscall(&self, val: i64) -> libseccomp::ScmpNotifResp {
        libseccomp::ScmpNotifResp::new(self.request.id, val, 0, 0)
    }

    /// Check if this event is still valid.
    /// In some cases this is necessary, please check `seccomp_unotify(2)` for more information.
    pub fn is_valid(&self) -> bool {
        libseccomp::notify_id_valid(self.notify_fd, self.request.id).is_ok()
    }

    /// Add a file descriptor to the supervised process.
    /// This could help avoid TOCTOU attack in some cases.
    pub fn add_fd(&self, src_fd: RawFd, close_on_exec: bool) -> Result<RawFd, Errno> {
        #[allow(clippy::cast_sign_loss)]
        let src_fd = if src_fd >= 0 {
            src_fd as u32
        } else {
            return Err(Errno::EBADF);
        };
        let newfd_flags = if close_on_exec {
            nix::libc::O_CLOEXEC as u32
        } else {
            0
        };
        let addfd: libseccomp_sys::seccomp_notif_addfd = libseccomp_sys::seccomp_notif_addfd {
            id: self.request.id,
            flags: 0,
            srcfd: src_fd,
            newfd: 0,
            newfd_flags,
        };

        // SAFETY: The 'ioctl' function is a low-level interface to the
        // kernel, and its safety depends on the correctness of its
        // arguments.  Here, we ensure that 'self.notify_fd' is a valid
        // file descriptor and 'addr_of!(addfd)' provides a valid
        // pointer to 'addfd'.  The usage of ioctl is considered safe
        // under these conditions, as it does not lead to undefined
        // behavior.
        let new_fd = unsafe {
            ioctl(
                self.notify_fd,
                SECCOMP_IOCTL_NOTIF_ADDFD,
                std::ptr::addr_of!(addfd),
            )
        };
        if new_fd < 0 {
            Err(Errno::last())
        } else {
            Ok(new_fd as RawFd)
        }
    }
}

/// By using `RemoteProcess`, you can get some information about the supervised process.
#[derive(Debug)]
pub struct RemoteProcess {
    pid: Pid,
}

impl RemoteProcess {
    /// Create a `RemoteProcess` object from a `Pid`.
    ///
    /// # Examples
    ///
    /// ```ignore
    /// let remote = RemoteProcess::new(Pid::from_raw(req.request.pid as i32));
    /// ```
    pub fn new(pid: Pid) -> Self {
        RemoteProcess { pid }
    }

    /// Get file descriptor from remote process with `pidfd_getfd()`.
    /// This function requires Linux 5.6+.
    pub fn get_fd(&self, remote_fd: RawFd) -> Result<OwnedFd, Errno> {
        // get TGID of given pid (TID)
        let tgid = proc_tgid(self.pid)?;
        // SAFETY: The call to `libc::syscall` for `SYS_pidfd_getfd` is
        // safe provided the arguments are correct.  Here, `self.fd` is
        // assumed to be a valid file descriptor, and `remote_fd` is a
        // valid RawFd. The third argument, `0`, is a valid flag for
        // the syscall. Assuming these conditions, the syscall does not
        // lead to undefined behavior.
        #[allow(clippy::cast_possible_truncation)]
        let pid_fd = match unsafe { libc::syscall(libc::SYS_pidfd_open, tgid, 0) } {
            e if e < 0 => return Err(Errno::last()),
            fd => unsafe { OwnedFd::from_raw_fd(fd as RawFd) },
        };
        let local_fd =
            unsafe { libc::syscall(libc::SYS_pidfd_getfd, pid_fd.as_raw_fd(), remote_fd, 0) };
        if local_fd < 0 {
            Err(Errno::last())
        } else {
            // SAFETY: Valid FD
            Ok(unsafe { OwnedFd::from_raw_fd(local_fd as RawFd) })
        }
    }

    /// Read data from remote process's memory with `process_vm_readv()`.
    pub fn read_mem(
        &self,
        local_buffer: &mut [u8],
        remote_addr: usize,
        request: &UNotifyEventRequest,
    ) -> Result<usize, Errno> {
        static FORCE_PROC: Lazy<bool> =
            Lazy::new(|| std::env::var(ENV_NO_CROSS_MEMORY_ATTACH).is_ok());
        if *FORCE_PROC {
            return self.read_mem_proc(local_buffer, remote_addr, request);
        }

        if remote_addr == 0 {
            return Err(Errno::EFAULT);
        } else if !request.is_valid() {
            return Err(Errno::ESRCH);
        }
        let len = local_buffer.len();
        match process_vm_readv(
            self.pid,
            &mut [IoSliceMut::new(local_buffer)],
            &[RemoteIoVec {
                len,
                base: remote_addr,
            }],
        ) {
            Ok(n) => {
                if !request.is_valid() {
                    Err(Errno::ESRCH)
                } else {
                    Ok(n)
                }
            }
            Err(Errno::ENOSYS | Errno::EPERM) => {
                self.read_mem_proc(local_buffer, remote_addr, request)
            }
            Err(e) => Err(e),
        }
    }

    /// Write data to remote process's memory with `process_vm_writev()`.
    pub fn write_mem(
        &self,
        local_buffer: &[u8],
        remote_addr: usize,
        request: &UNotifyEventRequest,
    ) -> Result<usize, Errno> {
        static FORCE_PROC: Lazy<bool> =
            Lazy::new(|| std::env::var(ENV_NO_CROSS_MEMORY_ATTACH).is_ok());
        if *FORCE_PROC {
            return self.write_mem_proc(local_buffer, remote_addr, request);
        }

        if remote_addr == 0 {
            return Err(Errno::EFAULT);
        } else if !request.is_valid() {
            return Err(Errno::ESRCH);
        }
        let len = local_buffer.len();
        match process_vm_writev(
            self.pid,
            &[IoSlice::new(local_buffer)],
            &[RemoteIoVec {
                len,
                base: remote_addr,
            }],
        ) {
            Err(Errno::ENOSYS | Errno::EPERM) => {
                self.write_mem_proc(local_buffer, remote_addr, request)
            }
            Ok(n) => Ok(n),
            Err(e) => Err(e),
        }
    }

    /// Fallback method to read data from `/proc/$pid/mem` when `process_vm_readv()` is unavailable.
    pub fn read_mem_proc(
        &self,
        local_buffer: &mut [u8],
        remote_addr: usize,
        request: &UNotifyEventRequest,
    ) -> Result<usize, Errno> {
        if remote_addr == 0 {
            return Err(Errno::EFAULT);
        }
        let mut buf = itoa::Buffer::new();
        let mut path = PathBuf::from("/proc");
        path.push(buf.format(self.pid.as_raw()));
        path.push("mem");
        let mut file = File::open(path).map_err(|_| Errno::last())?;
        if !request.is_valid() {
            return Err(Errno::ESRCH);
        }
        file.seek(SeekFrom::Start(remote_addr as u64))
            .map_err(|_| Errno::last())?;
        let nread = file.read(local_buffer).map_err(|_| Errno::last())?;
        if request.is_valid() {
            Ok(nread)
        } else {
            Err(Errno::ESRCH)
        }
    }

    /// Fallback method to write data to `/proc/$pid/mem` when `process_vm_writev()` is unavailable.
    fn write_mem_proc(
        &self,
        local_buffer: &[u8],
        remote_addr: usize,
        request: &UNotifyEventRequest,
    ) -> Result<usize, Errno> {
        if remote_addr == 0 {
            return Err(Errno::EFAULT);
        }
        let mut buf = itoa::Buffer::new();
        let mut path = PathBuf::from("/proc");
        path.push(buf.format(self.pid.as_raw()));
        path.push("mem");
        let mut file = OpenOptions::new()
            .write(true)
            .open(path)
            .map_err(|_| Errno::last())?;
        if !request.is_valid() {
            return Err(Errno::ESRCH);
        }
        file.seek(SeekFrom::Start(remote_addr as u64))
            .map_err(|_| Errno::last())?;
        file.write(local_buffer).map_err(|_| Errno::last())
    }

    /// Read path from the given system call argument with the given request,
    /// returning together the optional raw system call argument.
    #[allow(clippy::type_complexity)]
    fn read_path_opt(
        &self,
        request: &UNotifyEventRequest,
        arg: &SyscallPathArgument,
        replace_self: bool,
    ) -> Result<(PathBuf, Option<PathBuf>), Errno> {
        let req = request.get_request();
        let path = match arg.path {
            Some(idx) => {
                if req.data.args[idx] == 0 {
                    if arg.null {
                        // NULL is ok, use fd
                        None
                    } else {
                        return Err(Errno::EFAULT);
                    }
                } else {
                    Some(remote_path_n!(self, req, idx, request)?)
                }
            }
            None => None,
        };
        let mut doterr = false;
        let (path, path_raw) = if let Some(ref path) = path {
            if !arg.empty && path.is_empty() {
                return Err(Errno::ENOENT);
            }
            let ends_with_dot = crate::fs::path_ends_with_dot(path);
            if ends_with_dot {
                match arg.dotlast {
                    Some(Errno::ENOENT) => {
                        // This will be handled later,
                        // as we may need to return EEXIST instead
                        // of ENOENT if the path exists.
                        doterr = true;
                    }
                    Some(errno) => {
                        return Err(errno);
                    }
                    _ => {}
                }
            }
            #[allow(clippy::cast_possible_wrap)]
            let pid = Pid::from_raw(req.pid as i32);
            let path = OsStr::from_bytes(path.to_bytes());
            let path_raw = if replace_self {
                Some(replace_self_with_pid(PathBuf::from(path), pid))
            } else {
                Some(PathBuf::from(path))
            };
            let mut path = normalize_path(path);
            path = if path.is_absolute() {
                path
            } else if path.is_empty() && !ends_with_dot {
                // SAFETY: The ends_with_dot check above
                // ensures we return ENOTDIR when e.g. path is
                // a dot and the file descriptor argument is a
                // regular file. This happens because in this
                // case, joining the directory with an empty
                // path on the next branch essentially adds a
                // trailing slash to the path, making the
                // system call emulator fail with ENOTDIR if
                // the argument is not a directory. This way,
                // we avoid stat'ing the path here to
                // determine whether it's a directory or not.
                self.read_directory(request, arg)?
            } else if path == PathBuf::from(".") {
                // SAFETY: Add a trailing slash to the directory
                // to assert it must be a directory.
                self.read_directory(request, arg)?.join("")
            } else {
                self.read_directory(request, arg)?.join(&path)
            };
            (replace_self_with_pid(path, pid), path_raw)
        } else {
            (self.read_directory(request, arg)?, None)
        };

        // SAFETY: For system calls that don't resolve
        // symbolic links we resolve symbolic links
        // in all parts of the path except last to
        // avoid TOCTOU between check and handle.
        // This is handled transparently by the
        // syd::fs::canonicalize function.
        let path = if path.starts_with(MAGIC_PREFIX) {
            path
        } else {
            crate::fs::canonicalize(path, arg.resolve, crate::fs::MissingHandling::Missing)?
        };
        if doterr {
            // Delayed dotlast Errno::ENOENT handler, see above for the rationale.
            if access(&path, AccessFlags::F_OK).is_ok() {
                return Err(Errno::EEXIST);
            } else {
                return Err(Errno::ENOENT);
            }
        }
        Ok((path, path_raw))
    }

    /// Read path from the given system call argument with the given request.
    #[allow(clippy::cognitive_complexity)]
    fn read_path(
        &self,
        request: &UNotifyEventRequest,
        arg: &SyscallPathArgument,
    ) -> Result<PathBuf, Errno> {
        let req = request.get_request();
        let path = match arg.path {
            Some(idx) => {
                if req.data.args[idx] == 0 {
                    if arg.null {
                        // NULL is ok, use fd
                        None
                    } else {
                        return Err(Errno::EFAULT);
                    }
                } else {
                    Some(remote_path_n!(self, req, idx, request)?)
                }
            }
            None => None,
        };
        let mut doterr = false;
        let path = if let Some(ref path) = path {
            if !arg.empty && path.is_empty() {
                return Err(Errno::ENOENT);
            }
            let ends_with_dot = crate::fs::path_ends_with_dot(path);
            if ends_with_dot {
                match arg.dotlast {
                    Some(Errno::ENOENT) => {
                        // This will be handled later,
                        // as we may need to return EEXIST instead
                        // of ENOENT if the path exists.
                        doterr = true;
                    }
                    Some(errno) => {
                        return Err(errno);
                    }
                    _ => {}
                }
            }
            let path = OsStr::from_bytes(path.to_bytes());
            let mut path = normalize_path(path);
            path = if path.is_absolute() {
                path
            } else if path.is_empty() && !ends_with_dot {
                // SAFETY: The ends_with_dot check above
                // ensures we return ENOTDIR when e.g. path is
                // a dot and the file descriptor argument is a
                // regular file. This happens because in this
                // case, joining the directory with an empty
                // path on the next branch essentially adds a
                // trailing slash to the path, making the
                // system call emulator fail with ENOTDIR if
                // the argument is not a directory. This way,
                // we avoid stat'ing the path here to
                // determine whether it's a directory or not.
                self.read_directory(request, arg)?
            } else if path == PathBuf::from(".") {
                // SAFETY: Add a trailing slash to the directory
                // to assert it must be a directory.
                self.read_directory(request, arg)?.join("")
            } else {
                self.read_directory(request, arg)?.join(&path)
            };
            #[allow(clippy::cast_possible_wrap)]
            replace_self_with_pid(path, Pid::from_raw(req.pid as i32))
        } else {
            self.read_directory(request, arg)?
        };

        // SAFETY: For system calls that don't resolve
        // symbolic links we resolve symbolic links
        // in all parts of the path except last to
        // avoid TOCTOU between check and handle.
        // This is handled transparently by the
        // syd::fs::canonicalize function.
        let path = if path.starts_with(MAGIC_PREFIX) {
            path
        } else {
            crate::fs::canonicalize(path, arg.resolve, crate::fs::MissingHandling::Missing)?
        };
        if doterr {
            // Delayed dotlast Errno::ENOENT handler, see above for the rationale.
            if access(&path, AccessFlags::F_OK).is_ok() {
                return Err(Errno::EEXIST);
            } else {
                return Err(Errno::ENOENT);
            }
        }
        Ok(path)
    }

    /// Read directory from the given system call argument with the given request.
    fn read_directory(
        &self,
        request: &UNotifyEventRequest,
        arg: &SyscallPathArgument,
    ) -> Result<PathBuf, Errno> {
        let req = request.get_request();
        let mut is_dirfd: bool = false;
        let dir = match arg.dirfd {
            Some(idx) => {
                let fd = Self::remote_fd(req, idx);
                if let Some(fd) = fd {
                    is_dirfd = true;
                    if fd < 0 {
                        return Err(Errno::EBADF);
                    }
                }
                let dir = Self::remote_dirfd(req.pid, fd);
                if is_dirfd
                    && faccessat(None, &dir, AccessFlags::F_OK, AtFlags::AT_SYMLINK_NOFOLLOW)
                        == Err(Errno::ENOENT)
                {
                    return Err(Errno::EBADF);
                }
                dir
            }
            None => Self::remote_dirfd(req.pid, None),
        };
        match crate::fs::canonicalize(dir, true, crate::fs::MissingHandling::Missing) {
            Ok(dir) => {
                if !request.is_valid() {
                    Err(Errno::ESRCH)
                } else {
                    Ok(dir)
                }
            }
            Err(error) => {
                warn!("ctx": "read_path",
                    "arg": arg,
                    "dir": arg.dirfd.unwrap_or(-nix::libc::AT_FDCWD as usize),
                    "err": error.to_string(),
                    "pid": req.pid);
                Err(if is_dirfd { Errno::EBADF } else { error })
            }
        }
    }

    /// Convert the file descriptor argument into a `RawFd`.
    /// If the argument refers to the current working directory,
    /// this function returns `None`.
    pub fn remote_fd(req: &ScmpNotifReq, idx: usize) -> Option<RawFd> {
        #[allow(clippy::cast_possible_truncation)]
        let fd = req.data.args[idx] as i32;
        if fd == nix::libc::AT_FDCWD {
            None
        } else {
            Some(fd as RawFd)
        }
    }

    /// Returns the file descriptor or current working directory path for the given `Pid`.
    pub fn remote_dirfd(pid: u32, fd: Option<RawFd>) -> PathBuf {
        let mut buf0 = itoa::Buffer::new();
        let mut path = PathBuf::from("/proc");
        path.push(buf0.format(pid));

        if let Some(fd) = fd {
            let mut buf1 = itoa::Buffer::new();
            path.push("fd");
            path.push(buf1.format(fd));
        } else {
            path.push("cwd");
        }

        path
    }

    /// Read the `nix::libc::open_how` struct from process memory
    /// at the given address and size.
    pub fn remote_ohow(
        &self,
        addr: usize,
        size: usize,
        request: &UNotifyEventRequest,
    ) -> Result<nix::libc::open_how, Errno> {
        if size != std::mem::size_of::<nix::libc::open_how>() {
            return Err(Errno::EINVAL);
        }

        let mut buf = [0u8; std::mem::size_of::<nix::libc::open_how>()];
        self.read_mem(&mut buf, addr, request)?;

        // SAFETY: The following unsafe block assumes that:
        // 1. The memory layout of open_how in our Rust environment matches that of the target process.
        // 2. The proc.read_mem call has populated buf with valid data of the appropriate size (ensured by the size check above).
        // 3. The buffer is appropriately aligned for reading an open_how struct. If the remote process's representation of open_how
        //    was correctly aligned, our local buffer should be too, since it's an array on the stack.
        Ok(unsafe { std::ptr::read_unaligned(buf.as_ptr() as *const _) })
    }

    /// Read the `nix::libc::utimbuf` struct from process memory at the given address.
    /// Convert it to a `nix::libc::timespec[2]` for easy interoperability.
    fn remote_utimbuf(
        &self,
        addr: usize,
        request: &UNotifyEventRequest,
    ) -> Result<Option<[nix::libc::timespec; 2]>, Errno> {
        if addr == 0 {
            return Ok(None);
        }

        let mut buf = [0u8; std::mem::size_of::<nix::libc::utimbuf>()];
        self.read_mem(&mut buf, addr, request)?;

        // SAFETY: The following unsafe block assumes that:
        // 1. The memory layout of utimbuf in our Rust environment matches that of the target process.
        // 2. The proc.read_mem call has populated buf with valid data of the appropriate size (ensured by the size check above).
        // 3. The buffer is appropriately aligned for reading a utimbuf struct. If the remote process's representation of utimbuf
        //    was correctly aligned, our local buffer should be too, since it's an array on the stack.
        let t: nix::libc::utimbuf = unsafe { std::ptr::read_unaligned(buf.as_ptr() as *const _) };
        Ok(Some([
            nix::libc::timespec {
                tv_sec: t.actime,
                tv_nsec: 0,
            },
            nix::libc::timespec {
                tv_sec: t.modtime,
                tv_nsec: 0,
            },
        ]))
    }

    /// Read the `nix::libc::timeval[2]` struct from process memory at the given address.
    /// Convert it to a `nix::libc::timespec[2]` for easy interoperability.
    fn remote_timeval(
        &self,
        addr: usize,
        request: &UNotifyEventRequest,
    ) -> Result<Option<[nix::libc::timespec; 2]>, Errno> {
        if addr == 0 {
            return Ok(None);
        }

        let mut buf = [0u8; std::mem::size_of::<nix::libc::timeval>() * 2];
        self.read_mem(&mut buf, addr, request)?;

        // SAFETY: The following unsafe block assumes that:
        // 1. The memory layout of timeval in our Rust environment matches that of the target process.
        // 2. The proc.read_mem call has populated buf with valid data of the appropriate size (ensured by the size check above).
        // 3. The buffer is appropriately aligned for reading a timeval struct. If the remote process's representation of timeval
        //    was correctly aligned, our local buffer should be too, since it's an array on the stack.
        let t: [nix::libc::timeval; 2] =
            unsafe { std::ptr::read_unaligned(buf.as_ptr() as *const _) };
        Ok(Some([
            nix::libc::timespec {
                tv_sec: t[0].tv_sec,
                tv_nsec: (t[0].tv_usec as nix::libc::c_long).saturating_mul(1_000), /* ms->ns */
            },
            nix::libc::timespec {
                tv_sec: t[1].tv_sec,
                tv_nsec: (t[1].tv_usec as nix::libc::c_long).saturating_mul(1_000), /* ms->ns */
            },
        ]))
    }

    /// Read the `nix::libc::timespec[2]` struct from process memory at the given address.
    fn remote_timespec(
        &self,
        addr: usize,
        request: &UNotifyEventRequest,
    ) -> Result<Option<[nix::libc::timespec; 2]>, Errno> {
        if addr == 0 {
            return Ok(None);
        }

        let mut buf = [0u8; std::mem::size_of::<nix::libc::timespec>() * 2];
        self.read_mem(&mut buf, addr, request)?;

        // SAFETY: The following unsafe block assumes that:
        // 1. The memory layout of timespec in our Rust environment matches that of the target process.
        // 2. The proc.read_mem call has populated buf with valid data of the appropriate size (ensured by the size check above).
        // 3. The buffer is appropriately aligned for reading a timespec struct. If the remote process's representation of timespec
        //    was correctly aligned, our local buffer should be too, since it's an array on the stack.
        Ok(Some(unsafe {
            std::ptr::read_unaligned(buf.as_ptr() as *const _)
        }))
    }

    /// Read the path from memory of the process with the given `Pid` with the given address.
    fn remote_path(&self, addr: usize, request: &UNotifyEventRequest) -> Result<Box<CStr>, Errno> {
        let mut buf = [0u8; nix::libc::PATH_MAX as usize];
        self.read_mem(&mut buf, addr, request)?;

        // SAFETY: If buffer has no null byte,
        // return ENAMETOOLONG as the path is too
        // long for us to handle.
        Ok(Box::from(
            CStr::from_bytes_until_nul(&buf)
                .map_err(|_| Errno::ENAMETOOLONG)?
                .to_owned(),
        ))
    }
}

type ChildHandle = JoinHandle<Result<(), Errno>>;
type Child = (Pid, ChildHandle, ThreadPool);
type UserHookFunc = Box<dyn Fn(&UNotifyEventRequest) -> libseccomp::ScmpNotifResp + Send + Sync>;

/// The main component of greenhook.
pub struct Supervisor {
    export: Option<ExportMode>,
    handlers: HashMap<String, Arc<UserHookFunc>>,
    sysallow: HashSet<ScmpSyscall>,
    thread_pool: ThreadPool,
    sandbox: Arc<RwLock<Sandbox>>,
}

impl Supervisor {
    /// Create a new `Supervisor` object. You can specify the number of threads in the thread pool.
    /// This function will also check your kernel version and show warning or return error if necessary.
    ///
    /// # Examples
    /// ```
    /// use syd::{hook::Supervisor, sandbox::Sandbox};
    /// let supervisor = Supervisor::new(Sandbox::default(), num_cpus::get(), None).unwrap();
    /// ```
    pub fn new(
        sandbox: Sandbox,
        thread_num: usize,
        export_mode: Option<ExportMode>,
    ) -> Result<Self, io::Error> {
        if thread_num == 0 {
            return Err(io::Error::new(
                io::ErrorKind::InvalidInput,
                "thread_num should be greater than 0",
            ));
        }
        // detect kernel version and show warning
        let version = uname().map_err(|e| io::Error::from_raw_os_error(e as i32))?;
        let version = version.release();

        macro_rules! parse_error {
            () => {
                io::Error::new(io::ErrorKind::Other, "unknown version")
            };
        }

        let (major, minor) = {
            let mut iter = version.to_str().ok_or_else(|| parse_error!())?.split('.');
            let major = iter
                .next()
                .ok_or_else(|| parse_error!())?
                .parse::<u32>()
                .map_err(|_| parse_error!())?;
            let minor = iter
                .next()
                .ok_or_else(|| parse_error!())?
                .parse::<u32>()
                .map_err(|_| parse_error!())?;
            (major, minor)
        };
        if major < 5 {
            error!("ctx": "check_kernel", "err": "Your kernel version is too old.");
            return Err(io::Error::new(io::ErrorKind::Other, "kernel too old"));
        } else if major == 5 && minor < 5 {
            error!("ctx": "check_kernel", "err": "Your kernel version is too old (Does not support SECCOMP_USER_NOTIF_FLAG_CONTINUE, etc.).");
            return Err(io::Error::new(io::ErrorKind::Other, "kernel too old"));
        } else if major == 5 && minor < 6 {
            error!("ctx": "check_kernel", "err": "Your kernel version is too old (Does not support pidfd_getfd() and SECCOMP_IOCTL_NOTIF_ADDFD).");
            return Err(io::Error::new(io::ErrorKind::Other, "kernel too old"));
        } else if major == 5 && minor < 9 {
            error!("ctx": "check_kernel", "err": "Your kernel version is too old (Does not support SECCOMP_IOCTL_NOTIF_ADDFD).");
            return Err(io::Error::new(io::ErrorKind::Other, "kernel too old"));
        }
        Ok(Supervisor {
            export: export_mode,
            handlers: HashMap::new(),
            sysallow: HashSet::new(),
            thread_pool: rusty_pool::Builder::new()
                .name(module_path!().to_string().replace('o', "☮"))
                .core_size(thread_num)
                .max_size(rusty_pool::MAX_SIZE)
                .keep_alive(MON_KEEPALIVE)
                .build(),
            sandbox: Arc::new(RwLock::new(sandbox)),
        })
    }

    /// Initialize the environment for the sandbox.
    /// Call this before `init`.
    pub fn init_env() -> anyhow::Result<()> {
        // SAFETY: The nix::sys::signal::signal function is unsafe because it affects the global state
        // of the program by changing how a signal (SIGPIPE in this case) is handled. It's safe to call
        // here because changing the SIGPIPE signal to its default behavior will not interfere with any
        // other part of this program that could be relying on a custom SIGPIPE signal handler.
        unsafe {
            nix::sys::signal::signal(
                nix::sys::signal::Signal::SIGPIPE,
                nix::sys::signal::SigHandler::SigDfl,
            )
        }
        .context("Failed to set SIGPIPE signal handler to default.")?;

        // Initialize logging.
        JsonLinesLogger::init().context("Failed to initialize logging.")?;

        // SAFETY: Set the no_new_privs attribute.
        if unsafe { nix::libc::prctl(nix::libc::PR_SET_NO_NEW_PRIVS, 1, 0, 0, 0) } == -1 {
            let errno = Errno::last();
            bail!("Error setting no_new_privs attribute: {errno}");
        }
        debug!("ctx": "init", "no_new_privs": true);

        // SAFETY: Set parent-death signal
        if unsafe { nix::libc::prctl(nix::libc::PR_SET_PDEATHSIG, nix::libc::SIGKILL, 0, 0, 0) }
            != 0
        {
            let errno = Errno::last();
            bail!("Error setting parent-death signal: {errno}");
        }
        debug!("ctx": "init", "parent_death": nix::libc::SIGKILL);

        for cap in CAPS_DROP {
            // SAFETY: Drop capabilities as early as possible.
            let val = *cap as nix::libc::c_ulong;
            if unsafe { nix::libc::prctl(nix::libc::PR_CAPBSET_DROP, val, 0, 0, 0) } != 0 {
                match Errno::last() {
                    Errno::EINVAL | Errno::EPERM => {
                        // We either don't have CAP_SETPCAP capability,
                        // or capabilities are not enabled in the kernel.
                        // Let's move on.
                    }
                    errno => {
                        // Genuine error condition, let's fail.
                        bail!("Error dropping capability {cap}: {errno}");
                    }
                };
            } else {
                info!("ctx": "init", "cap_drop" : cap);
            }
        }

        // Apply seccomp strengtening for the Sydb☮x process itself.
        Self::setup_seccomp_parent()
            .context("Failed to create parent seccomp filter")?
            .load()
            .context("Failed to load parent seccomp filter")?;

        Ok(())
    }

    /// Initilizes the supervisor by adding the system call handlers.
    pub fn init(&mut self) -> &mut Self {
        // memory sandboxing
        self.insert_handler("brk", sys_brk);
        self.insert_handler("mmap", sys_mmap);
        self.insert_handler("mmap2", sys_mmap2);
        self.insert_handler("mremap", sys_mremap);

        // PID sandboxing
        self.insert_handler("fork", sys_fork);
        self.insert_handler("vfork", sys_vfork);
        self.insert_handler("clone", sys_clone);
        self.insert_handler("clone3", sys_clone3);

        // signal protection
        self.insert_handler("kill", sys_kill);
        self.insert_handler("tkill", sys_tkill);
        self.insert_handler("tgkill", sys_tgkill);
        self.insert_handler("pidfd_open", sys_pidfd_open);

        // network sandboxing
        self.insert_handler("socketcall", sys_socketcall);
        self.insert_handler("bind", sys_bind);
        self.insert_handler("connect", sys_connect);
        self.insert_handler("recvfrom", sys_recvfrom);
        self.insert_handler("sendto", sys_sendto);

        // exec sandboxing
        self.insert_handler("execve", sys_execve);
        self.insert_handler("execveat", sys_execveat);

        // stat sandboxing
        self.insert_handler("chdir", sys_chdir);
        self.insert_handler("fchdir", sys_fchdir);
        self.insert_handler("getdents", sys_getdents);
        self.insert_handler("getdents64", sys_getdents);
        self.insert_handler("stat", sys_stat);
        self.insert_handler("stat64", sys_stat);
        self.insert_handler("fstat", sys_fstat);
        self.insert_handler("fstat64", sys_fstat);
        self.insert_handler("lstat", sys_lstat);
        self.insert_handler("lstat64", sys_lstat);
        self.insert_handler("statx", sys_statx);
        self.insert_handler("fstatat64", sys_newfstatat);
        self.insert_handler("newfstatat", sys_newfstatat);

        // read/write sandboxing
        self.insert_handler("access", sys_access);
        self.insert_handler("faccessat", sys_faccessat);
        self.insert_handler("faccessat2", sys_faccessat2);
        self.insert_handler("chmod", sys_chmod);
        self.insert_handler("fchmod", sys_fchmod);
        self.insert_handler("fchmodat", sys_fchmodat);
        self.insert_handler("chown", sys_chown);
        self.insert_handler("fchown", sys_fchown);
        self.insert_handler("fchown32", sys_fchown);
        self.insert_handler("lchown", sys_lchown);
        self.insert_handler("lchown32", sys_lchown);
        self.insert_handler("fchownat", sys_fchownat);
        self.insert_handler("creat", sys_creat);
        self.insert_handler("link", sys_link);
        self.insert_handler("symlink", sys_symlink);
        self.insert_handler("unlink", sys_unlink);
        self.insert_handler("linkat", sys_linkat);
        self.insert_handler("symlinkat", sys_symlinkat);
        self.insert_handler("unlinkat", sys_unlinkat);
        self.insert_handler("mkdir", sys_mkdir);
        self.insert_handler("rmdir", sys_rmdir);
        self.insert_handler("mkdirat", sys_mkdirat);
        self.insert_handler("mknod", sys_mknod);
        self.insert_handler("mknodat", sys_mknodat);
        self.insert_handler("mount", sys_mount);
        self.insert_handler("umount", sys_umount);
        self.insert_handler("umount2", sys_umount2);
        self.insert_handler("open", sys_open);
        self.insert_handler("openat", sys_openat);
        self.insert_handler("openat2", sys_openat2);
        self.insert_handler("rename", sys_rename);
        self.insert_handler("renameat", sys_renameat);
        self.insert_handler("renameat2", sys_renameat2);
        self.insert_handler("utime", sys_utime);
        self.insert_handler("utimes", sys_utimes);
        self.insert_handler("futimesat", sys_futimesat);
        self.insert_handler("utimensat", sys_utimensat);
        self.insert_handler("truncate", sys_truncate);
        self.insert_handler("truncate64", sys_truncate);
        self.insert_handler("ftruncate", sys_ftruncate);
        self.insert_handler("getxattr", sys_getxattr);
        self.insert_handler("fgetxattr", sys_fgetxattr);
        self.insert_handler("lgetxattr", sys_lgetxattr);
        self.insert_handler("setxattr", sys_setxattr);
        self.insert_handler("fsetxattr", sys_fsetxattr);
        self.insert_handler("lsetxattr", sys_lsetxattr);
        self.insert_handler("listxattr", sys_listxattr);
        self.insert_handler("flistxattr", sys_flistxattr);
        self.insert_handler("llistxattr", sys_llistxattr);
        self.insert_handler("removexattr", sys_removexattr);
        self.insert_handler("fremovexattr", sys_fremovexattr);
        self.insert_handler("lremovexattr", sys_lremovexattr);

        // Allowlist safe system calls.
        for sysname in SAFE_SYSCALLS {
            let syscall = ScmpSyscall::new(sysname);
            if i32::from(syscall) == libseccomp_sys::__NR_SCMP_ERROR {
                error!("ctx": "allowlist_safe_syscall", "err": "scmp_error", "sys": sysname);
                continue;
            }
            self.allow_syscall(syscall);
        }

        self
    }

    /// Insert this system call to the list of allowed system calls.
    /// No filtering is done one these system calls and they're allowed at the kernel level.
    pub fn allow_syscall(&mut self, syscall: ScmpSyscall) {
        self.sysallow.insert(syscall);
    }

    ///
    /// # Examples
    ///
    /// ```no_run
    /// use libseccomp::ScmpSyscall;
    /// use syd::{
    ///     hook::{Supervisor, UNotifyEventRequest},
    ///     sandbox::Sandbox,
    /// };
    ///
    /// fn close_handler(req: &UNotifyEventRequest) -> libseccomp::ScmpNotifResp {
    ///     println!("close");
    ///     unsafe { req.continue_syscall() }
    /// }
    ///
    /// let mut supervisor = Supervisor::new(Sandbox::default(), num_cpus::get(), None).unwrap();
    /// supervisor.insert_handler("open", |req| {
    ///     println!("open: {}", req.get_request().data.args[0]);
    ///     unsafe { req.continue_syscall() }
    /// });
    /// supervisor.insert_handler("close", close_handler);
    /// ```
    pub fn insert_handler(
        &mut self,
        syscall_name: &str,
        handler: impl Fn(&UNotifyEventRequest) -> libseccomp::ScmpNotifResp + Send + Sync + 'static,
    ) {
        self.handlers
            .insert(syscall_name.to_string(), Arc::new(Box::new(handler)));
    }

    /// Run a command with seccomp filter.
    /// This method will fork a child process, do some preparations and run the command in it.
    /// It returns a `Pid`, a `JoinHandle` of supervising thread, and a `ThreadPool` handle of syscall user functions.
    /// It's recommended to use `Supervisor::wait()` to wait for the child process.
    pub fn spawn(self, mut command: crate::unshare::Command) -> anyhow::Result<Child> {
        let seccomp_filter = self
            .setup_seccomp()
            .context("Failed to set up seccomp filters.")?;
        command
            .seccomp_filter(seccomp_filter)
            .context("Failed to setup socket pair.")?;

        // Spawn child under sandbox.
        let sandbox = command
            .spawn()
            .context("Failed to spawn sandbox process.")?;
        let pid = sandbox.id();
        #[allow(clippy::disallowed_methods)]
        let fd = sandbox.seccomp_fd.unwrap();

        // Ignore some signals to ensure uniterrupted tracing.
        let _ = crate::ignore_signal(SIGTSTP);
        let _ = crate::ignore_signal(SIGTTIN);
        let _ = crate::ignore_signal(SIGTTOU);
        let _ = crate::ignore_signal(SIGHUP);

        #[allow(clippy::cast_possible_wrap)]
        let pid = Pid::from_raw(pid as i32);
        let mut sandbox = self.sandbox.write();
        sandbox.set_child_pid(pid);
        drop(sandbox);

        self.supervise(pid, fd)
    }

    // Set up seccomp for the sandbox process.
    #[allow(clippy::cognitive_complexity)]
    fn setup_seccomp(&self) -> IOResult<ScmpFilterContext> {
        let mut ctx =
            ScmpFilterContext::new_filter(ScmpAction::Errno(nix::libc::EACCES)).map_err(|e| {
                io::Error::new(
                    io::ErrorKind::Other,
                    format!("failed to create seccomp filter: {}", e),
                )
            })?;
        // Enforce the NO_NEW_PRIVS functionality before
        // loading the seccomp filter into the kernel.
        let _ = ctx.set_ctl_nnp(true);
        // We don't want ECANCELED, we want actual errnos.
        let _ = ctx.set_api_sysrawrc(true);
        // We deny with EACCES for bad system call, and kill process for bad arch.
        let _ = ctx.set_act_badarch(ScmpAction::KillProcess);
        // We log all filter actions other than Allow to kernel log if the log level is <=Debug.
        if log_enabled!(Level::Debug) {
            let _ = ctx.set_ctl_log(true);
        }

        seccomp_add_architectures(&mut ctx)?;

        // Acquire the read lock to sandbox configuration.
        let sandbox = self.sandbox.read();
        let fake_root = sandbox.get_root();
        drop(sandbox);

        // Fakeroot
        let id_action = if fake_root {
            ScmpAction::Errno(0)
        } else {
            ScmpAction::Allow
        };
        for sysname in GET_ID_SYSCALLS {
            let syscall = ScmpSyscall::new(sysname);
            ctx.add_rule(id_action, syscall).map_err(|e| {
                io::Error::new(
                    io::ErrorKind::Other,
                    format!(
                        "failed to add rule for system call {} ( {} ): {}",
                        syscall, sysname, e
                    ),
                )
            })?;
        }

        // Add notify rules for system calls with handlers.
        let syscall_notif: Vec<_> = self.handlers.keys().collect();
        for name in &syscall_notif {
            let syscall = ScmpSyscall::new(name);
            let _ = ctx.set_syscall_priority(syscall, crate::syscall_priority(name));
            ctx.add_rule(ScmpAction::Notify, syscall).map_err(|e| {
                io::Error::new(
                    io::ErrorKind::Other,
                    format!(
                        "failed to add rule for system call {} ( {} ): {}",
                        syscall, name, e
                    ),
                )
            })?;
        }

        // Add allow rules for system calls in the default allow list.
        let syscall_allow: Vec<_> = self.sysallow.iter().copied().collect();
        for syscall in &syscall_allow {
            ctx.add_rule(ScmpAction::Allow, *syscall).map_err(|e| {
                io::Error::new(
                    io::ErrorKind::Other,
                    format!(
                        "failed to add rule for system call {} ( {} ): {}",
                        *syscall,
                        ScmpSyscall::get_name(*syscall).unwrap_or("?".to_string()),
                        e
                    ),
                )
            })?;
        }

        // Apply sandbox restrictions as necessary.
        let sandbox = self.sandbox.read();
        let restrict_ioctl = !sandbox.allow_unsafe_ioctl();
        let restrict_prctl = !sandbox.allow_unsafe_prctl();
        let restrict_getrandom = !sandbox.allow_unsafe_getrandom();
        drop(sandbox);

        // Restriction 1: Allowlist known-safe ioctls.
        #[allow(clippy::disallowed_methods)]
        let sys_ioctl = ScmpSyscall::from_name("ioctl").unwrap();
        if restrict_ioctl {
            for (opt, _) in ALLOWLIST_IOCTL {
                let cmp = ScmpArgCompare::new(1, ScmpCompareOp::Equal, *opt);
                ctx.add_rule_conditional(ScmpAction::Allow, sys_ioctl, &[cmp])
                    .map_err(|error| {
                        io::Error::new(io::ErrorKind::Other, format!("SeccompError: {error}"))
                    })?;
            }
        } else {
            ctx.add_rule(ScmpAction::Allow, sys_ioctl)
                .map_err(|error| {
                    io::Error::new(io::ErrorKind::Other, format!("SeccompError: {error}"))
                })?;
        }

        // Restriction 2: Allowlist known-safe prctls.
        // Things like PR_SET_MM, PR_SET_PTRACER, and PR_SET_SPECULATION_CTRL are left out.
        #[allow(clippy::disallowed_methods)]
        let sys_prctl = ScmpSyscall::from_name("prctl").unwrap();
        if restrict_prctl {
            for (opt, _) in ALLOWLIST_PRCTL {
                let cmp = ScmpArgCompare::new(0, ScmpCompareOp::Equal, *opt);
                ctx.add_rule_conditional(ScmpAction::Allow, sys_prctl, &[cmp])
                    .map_err(|error| {
                        io::Error::new(io::ErrorKind::Other, format!("SeccompError: {error}"))
                    })?;
            }
        } else {
            ctx.add_rule(ScmpAction::Allow, sys_prctl)
                .map_err(|error| {
                    io::Error::new(io::ErrorKind::Other, format!("SeccompError: {error}"))
                })?;
        }

        // Restriction 3: Disallow getrandom accessing /dev/random.
        // SAFETY: We must ensure the sandbox process can't
        // drain the system entropy. Accessing /dev/urandom is fine.
        #[allow(clippy::disallowed_methods)]
        let sys_getrandom = ScmpSyscall::from_name("getrandom").unwrap();
        if restrict_getrandom {
            ctx.add_rule_conditional(
                ScmpAction::Allow,
                sys_getrandom,
                &[ScmpArgCompare::new(
                    2,
                    ScmpCompareOp::MaskedEqual(u64::from(nix::libc::GRND_RANDOM)),
                    0,
                )],
            )
            .map_err(|error| {
                io::Error::new(io::ErrorKind::Other, format!("SeccompError: {error}"))
            })?;
        } else {
            ctx.add_rule(ScmpAction::Allow, sys_getrandom)
                .map_err(|error| {
                    io::Error::new(io::ErrorKind::Other, format!("SeccompError: {error}"))
                })?;
        }

        // Export seccomp rules if requested.
        if let Some(mode) = self.export {
            self.seccomp_export(&ctx, mode)?;
        }

        Ok(ctx)
    }

    /// Set up seccomp for the Sydb☮x process.
    /// This is important to restrict potential attack space in case
    /// Sydb☮x process is compromised anyhow.
    fn setup_seccomp_parent() -> IOResult<ScmpFilterContext> {
        let mut ctx = ScmpFilterContext::new_filter(ScmpAction::Allow).map_err(|e| {
            io::Error::new(
                io::ErrorKind::Other,
                format!("failed to create seccomp filter for Sydb☮x: {}", e),
            )
        })?;
        // We don't want ECANCELED, we want actual errnos.
        let _ = ctx.set_api_sysrawrc(true);
        // We kill for bad system cal and bad arch.
        let _ = ctx.set_act_badarch(ScmpAction::KillProcess);
        // We log all filter actions other than Allow to kernel log.
        let _ = ctx.set_ctl_log(true);

        seccomp_add_architectures(&mut ctx)?;

        // SAFETY: Add some more common calls to limit potential attack-space.
        let additional_syscalls = ["ptrace", "chroot", "pivot_root"];
        for sysname in SET_ID_SYSCALLS.iter().chain(&additional_syscalls) {
            let syscall = ScmpSyscall::new(sysname);
            ctx.add_rule(ScmpAction::Errno(nix::libc::EACCES), syscall)
                .map_err(|e| {
                    io::Error::new(
                        io::ErrorKind::Other,
                        format!(
                            "failed to add rule for system call {} ( {} ): {}",
                            syscall, sysname, e
                        ),
                    )
                })?;
        }

        // SAFETY: Make sync calls no-op to prevent potential local DoS.
        let ignore_syscalls = ["sync", "syncfs"];
        for sysname in ignore_syscalls {
            let syscall = ScmpSyscall::new(sysname);
            ctx.add_rule(ScmpAction::Errno(0), syscall).map_err(|e| {
                io::Error::new(
                    io::ErrorKind::Other,
                    format!(
                        "failed to add rule for system call {} ( {} ): {}",
                        syscall, sysname, e
                    ),
                )
            })?;
        }

        // Export seccomp rules if requested.
        if std::env::var("SYD_SECX").is_ok() {
            println!("# SydB☮x parent rules");
            ctx.export_pfc(&mut io::stdout()).map_err(|error| {
                io::Error::new(
                    io::ErrorKind::Other,
                    format!("failed to export rules in pfc format: {error}"),
                )
            })?;
        }

        Ok(ctx)
    }

    /// Export a seccomp context as bpf or pfc.
    fn seccomp_export(&self, ctx: &ScmpFilterContext, mode: ExportMode) -> IOResult<()> {
        match mode {
            ExportMode::BerkeleyPacketFilter => ctx.export_bpf(&mut io::stdout()),
            ExportMode::PseudoFiltercode => {
                #[allow(clippy::disallowed_methods)]
                let fd = match mkstemp("syd-tmp-XXXXXX") {
                    Ok((fd, path)) => {
                        unlink(path.as_path()).expect("Failed to delete temporary file.");
                        fd
                    }
                    Err(error) => {
                        return Err(error.into());
                    }
                };
                // SAFETY: In libc, we trust.
                let mut file = unsafe { File::from_raw_fd(fd) };

                if let Err(error) = ctx.export_pfc(&mut file) {
                    return Err(io::Error::new(io::ErrorKind::Other, format!("{error}")));
                }

                file.seek(SeekFrom::Start(0))?;
                let mut buffer = Vec::new();
                file.read_to_end(&mut buffer)?;

                let output = String::from_utf8_lossy(&buffer);
                let output = output.replace("0x7fc00000", "NOTIFY");

                #[allow(clippy::disallowed_methods)]
                let libver = ScmpVersion::current()
                    .expect("Failed to determine libseccomp library version.");
                println!(
                    "# Sydb☮x v{} seccomp rules generated by libsecc☮mp v{}.{}.{}",
                    env!("CARGO_PKG_VERSION"),
                    libver.major,
                    libver.minor,
                    libver.micro
                );
                println!("# API Version: {API_VERSION}");

                #[allow(clippy::disallowed_methods)]
                let mut syscall_allow: Vec<_> = self
                    .sysallow
                    .iter()
                    .copied()
                    .map(|sys| sys.get_name().unwrap())
                    .collect();
                #[allow(clippy::disallowed_methods)]
                let mut syscall_notif: Vec<_> = self.handlers.keys().collect();
                syscall_allow.sort();
                syscall_notif.sort();

                println!("# System calls with Action=ALLOW: {}", syscall_allow.len());
                println!("# System calls with Action=NOTIF: {}", syscall_notif.len());

                let uidcall = GET_ID_SYSCALLS.to_vec().join(", ");
                let sandbox = self.sandbox.read();
                println!(
                    "# Fake Root: {} ( {uidcall} )",
                    if sandbox.get_root() { "yes" } else { "no" }
                );
                println!(
                    "{}",
                    sandbox
                        .to_string()
                        .lines()
                        .map(|line| format!("# {}", line))
                        .collect::<Vec<_>>()
                        .join("\n")
                );
                drop(sandbox);

                println!("# Action=NOTIF: {}", syscall_notif.len());
                for name in &syscall_notif {
                    println!("#    - {name}");
                }
                println!("# Action=ALLOW: {}", syscall_allow.len());
                for name in &syscall_allow {
                    println!("#    - {name}");
                }
                print!("{output}");
                Ok(())
            }
        }
        .map_err(|error| {
            io::Error::new(
                io::ErrorKind::Other,
                format!("failed to export rules in {mode:?} format: {error}"),
            )
        })
    }

    /// Logic for the supervise child thread.
    fn supervise(self, pid: Pid, fd: RawFd) -> anyhow::Result<Child> {
        let pool_handle = self.thread_pool.clone();

        let thread_handle = Builder::new().name("syd::p☮ll".into()).spawn(move || {
            // SAFETY: Set thread priority to a relatively high value.
            let _ = unsafe { nix::libc::setpriority(nix::libc::PRIO_PROCESS, 0, 1) };

            // 0. Set (process-wide) umask to 0.
            let _ = umask(Mode::empty());

            // 1. Wrap fd in an OwnedFd to ensure it gets closed properly.
            // SAFETY: We're passing a valid fd to OwnedFd.
            let _guard_fd = unsafe { OwnedFd::from_raw_fd(fd) };

            // 2. Create an epoll instance
            let epfd = epoll_create1(EpollCreateFlags::EPOLL_CLOEXEC)?;

            // 3. Wrap epoll fd in an OwnedFd to ensure it gets closed properly.
            // SAFETY: We're passing a valid fd to OwnedFd.
            let _guard_epfd = unsafe { OwnedFd::from_raw_fd(epfd) };

            // 4. Add the file descriptor to the epoll instance
            #[allow(clippy::cast_sign_loss)]
            let mut event = EpollEvent::new(EpollFlags::EPOLLIN, fd as u64);
            epoll_ctl(epfd, EpollOp::EpollCtlAdd, fd, Some(&mut event))?;

            // 5. Wait for events
            loop {
                let mut events = [EpollEvent::empty(); 1];
                let nfds = match epoll_wait(epfd, &mut events, -1) {
                    Ok(nfds) => nfds,
                    Err(Errno::EINTR) => {
                        // An epoll_wait call can - in absence of any
                        // user defined signal handlers - fail with
                        // EINTR on SIGTRAP (eg, attaching strace to a
                        // running process) or SIGCONT (continuing a
                        // stopped process):
                        // https://lwn.net/Articles/851813/
                        continue;
                    },
                    Err(error) => return Err(error),
                };

                if nfds == 0 {
                    error!("ctx": "epoll", "error": "no file descriptors returned from epoll_wait");
                    return Err(Errno::EINVAL);
                } else if events[0].events().contains(EpollFlags::EPOLLHUP) {
                    debug!("ctx": "epoll", "error": "unexpected epoll hung up event", "events": format!("{:?}", events[0].events()));
                    return Err(Errno::EINVAL);
                } else if !events[0].events().contains(EpollFlags::EPOLLIN) {
                    error!("ctx": "epoll", "error": "unexpected epoll event", "events": format!("{:?}", events[0].events()));
                    return Err(Errno::EINVAL);
                }

                #[allow(unused_variables)]
                match ScmpNotifReq::receive(fd) {
                    Ok(req) => {
                        if !self.syscall(fd, req) {
                            // Handling system call failed, break.
                            return Ok(())
                        }
                    }
                    Err(error) => {
                        // SAFETY: We need to continue with epoll after
                        // any error as otherwise the sandbox process will
                        // fail all syscalls with ENOSYS or will just deadlock.
                        match Errno::last() {
                            Errno::ENOENT => {
                                // ENOENT: The task was interrupted by a fatal
                                // signal between the time we get the poll event
                                // and when we attempted to receive the notification.
                            },
                            errno => {
                                debug!("ctx": "seccomp_notif_recv", "error": error.to_string(), "errno": errno.to_string());
                            }
                        }
                    }
                }
            }
        }).context("Failed to spawn the poller thread.")?;

        Ok((pid, thread_handle, pool_handle))
    }

    fn syscall(&self, fd: RawFd, req: ScmpNotifReq) -> bool {
        let event_req = UNotifyEventRequest::new(req, fd, Arc::clone(&self.sandbox));

        let syscall_id = req.data.syscall;
        let syscall_arch = req.data.arch;
        let syscall_handler = match syscall_id.get_name_by_arch(syscall_arch) {
            Ok(name) => self.handlers.get(&name),
            Err(error) => {
                error!("ctx": "syscall_lookup",
                    "pid": req.pid,
                    "sys": syscall_id.to_string(),
                    "arch": format!("{syscall_arch:?}"),
                    "args": req.data.args,
                    "err": error.to_string());
                None
            }
        };
        match syscall_handler {
            Some(handler) => {
                let handler_in_thread = Arc::clone(handler);
                self.thread_pool.execute(move || {
                    PRIORITY_SET.with(|priority_set| {
                        if !priority_set.get() {
                            // SAFETY: Set thread priority to a low value.
                            let _ =
                                unsafe { nix::libc::setpriority(nix::libc::PRIO_PROCESS, 0, 20) };
                            // SAFETY: Set I/O priority to idle.
                            let _ = set_io_priority_idle();

                            priority_set.set(true);
                        }
                    });
                    let response = handler_in_thread(&event_req);
                    if event_req.is_valid() {
                        #[allow(unused_variables)]
                        if let Err(error) = response.respond(fd) {
                            let errno = Errno::last();
                            debug!("ctx": "spawn_response",
                                "pid": req.pid,
                                "sys": syscall_id.to_string(),
                                "arch": format!("{syscall_arch:?}"),
                                "args": req.data.args,
                                "err": error.to_string(),
                                "errno": errno.to_string());
                        }
                    } else {
                        debug!("ctx": "spawn_response",
                            "pid": req.pid,
                            "sys": syscall_id.to_string(),
                            "arch": format!("{syscall_arch:?}"),
                            "args": req.data.args,
                            "err": "invalid request",
                            "errno": Errno::EINVAL.to_string());
                    }
                });
                true
            }
            None => {
                error!("ctx": "syscall_handle",
                    "pid": req.pid,
                    "sys": syscall_id.to_string(),
                    "arch": format!("{syscall_arch:?}"),
                    "args": req.data.args,
                    "err": "unknown syscall",
                    "errno": Errno::EINVAL.to_string());
                #[allow(unused_variables)]
                if let Err(error) = event_req.fail_syscall(libc::ENOSYS).respond(fd) {
                    let errno = Errno::last();
                    debug!("ctx": "spawn_response",
                        "pid": req.pid,
                        "sys": syscall_id.to_string(),
                        "arch": format!("{syscall_arch:?}"),
                        "args": req.data.args,
                        "err": error.to_string(),
                        "errno": errno.to_string());
                }
                true
            }
        }
    }

    /// Wait for the child process to exit and cleanup the supervisor thread and thread pool.
    /// It returns `WaitStatus` of the child process.
    ///
    /// # Examples
    ///
    /// ```ignore
    /// let status = Supervisor::wait(pid, thread_handle, pool).unwrap();
    /// ```
    pub fn wait(
        child: Pid,
        thread_handle: ChildHandle,
        pool_handle: ThreadPool,
        wait_all: bool,
    ) -> Result<i32, Errno> {
        let mut exit_code: i32 = 127;
        loop {
            #[allow(clippy::cast_possible_truncation)]
            #[allow(clippy::cast_sign_loss)]
            match waitpid(None, Some(WaitPidFlag::__WNOTHREAD)) {
                Ok(WaitStatus::Exited(pid, code)) if pid == child => {
                    exit_code = code;
                    if !wait_all {
                        break;
                    }
                }
                Ok(WaitStatus::Signaled(pid, signal, _)) if pid == child => {
                    exit_code = 128_i32.saturating_add(signal as i32);
                    if !wait_all {
                        break;
                    }
                }
                Err(Errno::ECHILD) => break,
                Err(Errno::EINTR) => {}
                Err(errno) => return Err(errno),
                _ => {}
            }
        }
        if wait_all {
            let _ = thread_handle.join().map_err(|_| Errno::EPIPE)?;
            pool_handle.join();
        }
        Ok(exit_code)
    }
}

/// Processes the address family of a `SockaddrStorage` object and performs logging or other
/// required operations specific to the syscall being handled.
///
/// This helper function isolates the logic involved in dealing with different address families
/// and reduces code duplication across different syscall handler functions.
///
/// # Parameters
///
/// - `addr`: Reference to a `SockaddrStorage`, representing the socket address involved in the syscall.
/// - `syscall_name`: A string slice holding the name of the syscall being handled, used for logging purposes.
///
/// # Safety
///
/// The function contains unsafe blocks due to potential TOCTOU (Time-of-Check Time-of-Use)
/// vulnerabilities. Each unsafe block within this function has been annotated with a detailed
/// safety comment to ensure that unsafe operations are used correctly and securely.
///
/// # Errors
///
/// The function returns an `io::Error` in cases where:
/// - The conversion from `SockaddrStorage` to a specific address family representation fails.
/// - Any other unexpected error condition occurs during the processing of the address family.
///
/// # Returns
///
/// Returns an `Result<(), Errno>`:
/// - `Ok(())` if the processing is successful.
/// - `Err(Errno)` containing a description of the error, if any error occurs during processing.
fn sandbox_addr(
    proc: &RemoteProcess,
    request: &UNotifyEventRequest,
    req: &ScmpNotifReq,
    addr: &SockaddrStorage,
    syscall_name: &str,
    caps: Capability,
) -> Result<(), Errno> {
    match addr.family() {
        Some(AddressFamily::Unix) => {
            sandbox_addr_unix(proc, request, req, addr, syscall_name, caps)
        }
        Some(AddressFamily::Inet) => {
            sandbox_addr_inet(proc, request, req, addr, syscall_name, caps)
        }
        Some(AddressFamily::Inet6) => {
            sandbox_addr_inet6(proc, request, req, addr, syscall_name, caps)
        }
        Some(_) | None => sandbox_addr_notsup(proc, request),
    }
}

/// Process a `AddressFamily::Unix` socket address.
fn sandbox_addr_unix(
    proc: &RemoteProcess,
    request: &UNotifyEventRequest,
    req: &ScmpNotifReq,
    addr: &SockaddrStorage,
    syscall_name: &str,
    caps: Capability,
) -> Result<(), Errno> {
    let addr = addr.as_unix_addr().ok_or(Errno::EINVAL)?;
    let (path, abs) = match (addr.path(), addr.as_abstract()) {
        (Some(path), _) => {
            let path = path.as_os_str().as_bytes();
            let null = path.iter().position(|&b| b == 0).unwrap_or(path.len());
            let path = PathBuf::from(OsStr::from_bytes(&path[..null]));
            trace!("ctx": "syscall", "sys": syscall_name, "addr": format!("{}", path.display()), "cap": caps);
            (path, false)
        }
        (_, Some(path)) => {
            let null = path.iter().position(|&b| b == 0).unwrap_or(path.len());
            let path = PathBuf::from(OsStr::from_bytes(&path[..null]));
            trace!("ctx": "syscall", "sys": syscall_name, "addr": format!("{}", path.display()), "cap": caps);
            (path, true)
        }
        _ => {
            // unnamed unix socket
            return Ok(());
        }
    };

    let path = if path.is_relative() {
        let arg = SyscallPathArgument {
            dirfd: None,
            path: None,
            null: false,
            empty: false,
            resolve: true,
            dotlast: None,
        };
        proc.read_directory(request, &arg)?.join(&path)
    } else {
        crate::fs::canonicalize(path, true, crate::fs::MissingHandling::Missing)?
    };

    // Check for access.
    let sandbox = request.get_sandbox(false);
    let action = sandbox.check_unix(caps, &path);
    drop(sandbox);

    if action == Action::Deny {
        // Report access violation.
        warn!("ctx": "access", "cap": caps, "unix": format!("{}", path.display()), "abs": abs, "pid": req.pid, "sys": syscall_name);
    }
    if action == Action::Allow {
        Ok(())
    } else {
        // Deny or Filter.
        Err(Errno::EACCES)
    }
}

/// Process an `AddressFamily::Inet` socket address.
fn sandbox_addr_inet(
    _proc: &RemoteProcess,
    request: &UNotifyEventRequest,
    req: &ScmpNotifReq,
    addr: &SockaddrStorage,
    syscall_name: &str,
    caps: Capability,
) -> Result<(), Errno> {
    let addr = addr.as_sockaddr_in().ok_or(Errno::EINVAL)?;
    let port = addr.port();
    let addr = IpAddr::V4(Ipv4Addr::from(addr.ip()));
    trace!("ctx": "syscall", "sys": syscall_name, "addr": format!("{addr}!{port}"), "cap": caps);

    // Check for access.
    let sandbox = request.get_sandbox(false);
    let action = sandbox.check_ip(caps, addr, port);
    drop(sandbox);

    if action == Action::Deny {
        // Report access violation.
        warn!("ctx": "access", "cap": caps, "addr": format!("{addr}!{port}"), "pid": req.pid, "sys": syscall_name);
    }
    if action == Action::Allow {
        Ok(())
    } else {
        // Deny or Filter.
        Err(Errno::EACCES)
    }
}

/// Process an `AddressFamily::Inet6` socket address.
fn sandbox_addr_inet6(
    _proc: &RemoteProcess,
    request: &UNotifyEventRequest,
    req: &ScmpNotifReq,
    addr: &SockaddrStorage,
    syscall_name: &str,
    caps: Capability,
) -> Result<(), Errno> {
    let addr = addr.as_sockaddr_in6().ok_or(Errno::EINVAL)?;
    let port = addr.port();
    let addr = IpAddr::V6(addr.ip());
    trace!("ctx": "syscall", "sys": syscall_name, "addr": format!("{addr}!{port}"), "cap": caps);

    // Check for access.
    let sandbox = request.get_sandbox(false);
    let action = sandbox.check_ip(caps, addr, port);
    drop(sandbox);

    if action == Action::Deny {
        // Report access violation.
        warn!("ctx": "access", "cap": caps, "addr": format!("{addr}!{port}"), "pid": req.pid, "sys": syscall_name);
    }
    if action == Action::Allow {
        Ok(())
    } else {
        // Deny or Filter.
        Err(Errno::EACCES)
    }
}

/// Process a socket address of an unsupported socket family.
fn sandbox_addr_notsup(_proc: &RemoteProcess, request: &UNotifyEventRequest) -> Result<(), Errno> {
    let sandbox = request.get_sandbox(false);
    let ok = sandbox.allow_unsupported_socket_families();
    drop(sandbox);

    if ok {
        Ok(())
    } else {
        Err(Errno::EAFNOSUPPORT)
    }
}

/// Process the first path argument.
fn sandbox_path_1(
    paths: &NonEmpty<PathBuf>,
    caps: Capability,
    syscall_name: &str,
    request: &UNotifyEventRequest,
    _: &RemoteProcess,
) -> Result<(), Errno> {
    let req = request.get_request();
    let path = paths.first(); // paths are NonEmpty, so first always exists.
    trace!("ctx": "syscall", "sys": syscall_name, "cap": caps, "path": format!("{}", path.display()));

    if caps.is_empty() {
        return Err(Errno::EINVAL);
    }

    // Special case for system calls that must create the path argument.
    // Note: dangling symbolic links are considered existing.
    if (syscall_name.starts_with("mkdir") || syscall_name.starts_with("mknod"))
        && lstat(Path::new(path)).is_ok()
    {
        debug!("ctx": "sandbox_path_1", "err": "EEXIST", "path": format!("{}", path.display()));
        return Err(Errno::EEXIST);
    }

    let sandbox = request.get_sandbox(false);
    let mut action = Action::Allow;

    // Sandboxing
    if caps.contains(Capability::CAP_READ) {
        action = sandbox.check_path(Capability::CAP_READ, path);
    }
    if action == Action::Allow && caps.contains(Capability::CAP_STAT) {
        action = sandbox.check_path(Capability::CAP_STAT, path);
    }
    if action == Action::Allow && caps.contains(Capability::CAP_WRITE) {
        action = sandbox.check_path(Capability::CAP_WRITE, path);
    }
    if action == Action::Allow && caps.contains(Capability::CAP_EXEC) {
        action = sandbox.check_path(Capability::CAP_EXEC, path);
    }

    // exec/kill
    if caps.contains(Capability::CAP_EXEC) && sandbox.check_exec(path) == Action::Kill {
        warn!("ctx": "exec/kill", "cap": caps, "path": format!("{}", path.display()), "pid": req.pid, "sys": syscall_name);
        #[allow(clippy::cast_possible_wrap)]
        let _ = kill(Pid::from_raw(req.pid as i32), Some(SIGKILL));
        action = Action::Kill;
    }

    match action {
        Action::Allow => Ok(()),
        _ if caps.contains(Capability::CAP_STAT) => {
            // We do not report violations for Stat capability because we are essentially hiding paths.
            Err(Errno::ENOENT)
        }
        Action::Deny | Action::Filter if sandbox.is_hidden(path) => {
            // SAFETY: We do not report violations when the path is
            // hidden and return ENOENT, so as to make read/write/exec
            // sandboxing consistent with stat sandboxing. With this
            // restriction in place, it is not possible to enumerate
            // existing hidden paths by attempting to read, write or
            // execute them.

            // Debug log access violation to ease troubleshooting
            // using the log feature (compiled out by default).
            debug!("ctx": "access", "act": action, "cap": caps, "path": format!("{}", path.display()), "pid": req.pid, "sys": syscall_name);

            Err(Errno::ENOENT)
        }
        Action::Filter | Action::Kill => Err(Errno::EACCES),
        Action::Deny => {
            // Report access violation.
            warn!("ctx": "access", "cap": caps, "path": format!("{}", path.display()), "pid": req.pid, "sys": syscall_name);
            Err(Errno::EACCES)
        }
    }
}

/// Process both the first and the second path argument.
fn sandbox_path_2(
    paths: &NonEmpty<PathBuf>,
    caps: Capability,
    syscall_name: &str,
    request: &UNotifyEventRequest,
    proc: &RemoteProcess,
) -> Result<(), Errno> {
    let (source, target) = (paths.first(), paths.last()); // paths are NonEmpty, so first/last always exists.
    trace!("ctx": "syscall", "sys": syscall_name, "cap": caps, "source": format!("{}", source.display()), "target": format!("{}", target.display()));

    sandbox_path_1(
        &nonempty![source.to_path_buf()],
        caps,
        syscall_name,
        request,
        proc,
    )?;
    sandbox_path_1(
        &nonempty![target.to_path_buf()],
        caps,
        syscall_name,
        request,
        proc,
    )?;
    Ok(())
}

/*
 * System call handlers
 */
// System page size
static PAGE_SIZE: Lazy<u64> = Lazy::new(|| {
    sysconf(SysconfVar::PAGE_SIZE)
        .unwrap_or(Some(4096))
        .unwrap_or(4096) as u64
});

fn sys_brk(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_mem_handler(request, "brk")
}

fn sys_mmap(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_mem_handler(request, "mmap")
}

fn sys_mmap2(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_mem_handler(request, "mmap2")
}

fn sys_mremap(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();
    #[allow(clippy::cast_possible_truncation)]
    let old_size = req.data.args[1] as usize;
    #[allow(clippy::cast_possible_truncation)]
    let new_size = req.data.args[2] as usize;
    if new_size < old_size {
        // SAFETY: System call wants to shrink memory.
        return unsafe { request.continue_syscall() };
    }
    syscall_mem_handler(request, "mremap")
}

#[allow(clippy::cognitive_complexity)]
fn syscall_mem_handler(request: &UNotifyEventRequest, name: &str) -> ScmpNotifResp {
    // Get mem & vm max.
    let sandbox = request.get_sandbox(false);
    let enabled = sandbox.enabled(Capability::CAP_MEM);
    let mem_max = sandbox.mem_max;
    let mem_vm_max = sandbox.mem_vm_max;
    let mem_filter = sandbox.mem_filter;
    let kill_mem   = sandbox.kill_mem();
    drop(sandbox); // release the read lock.

    if !enabled || (mem_max == 0 && mem_vm_max == 0) {
        // SAFETY: No pointer dereference in security check.
        // This is safe to continue.
        return unsafe { request.continue_syscall() };
    }

    // Get process entry.
    let req = request.get_request();
    #[allow(clippy::cast_possible_wrap)]
    let proc = match Process::new(req.pid as nix::libc::pid_t) {
        Ok(proc) => proc,
        Err(error) => {
            if !mem_filter {
                warn!("ctx" : "access",
                    "cap" : "m",
                    "sys" : name,
                    "pid" : req.pid,
                    "max" : mem_max,
                    "max_vm" : mem_vm_max,
                    "kill" : kill_mem,
                    "err" : error.to_string());
            }
            if kill_mem {
                #[allow(clippy::cast_possible_wrap)]
                let _ = kill(Pid::from_raw(req.pid as i32), Some(SIGKILL));
            }
            return request.fail_syscall(nix::libc::ENOMEM);
        }
    };

    // Check VmSize
    if mem_vm_max > 0 {
        match proc.statm() {
            Ok(statm) if statm.size.saturating_mul(*PAGE_SIZE) >= mem_vm_max => {
                if !mem_filter {
                    warn!("ctx" : "access",
                        "cap" : "m",
                        "pid" : req.pid,
                        "sys" : name,
                        "kill" : kill_mem,
                        "max" : mem_max,
                        "max_vm": mem_vm_max);
                }
                if kill_mem {
                    #[allow(clippy::cast_possible_wrap)]
                    let _ = kill(Pid::from_raw(req.pid as i32), Some(SIGKILL));
                }
                return request.fail_syscall(nix::libc::ENOMEM);
            }
            Err(error) => {
                if !mem_filter {
                    warn!("ctx" : "access",
                        "cap" : "m",
                        "sys" : name,
                        "pid" : req.pid,
                        "kill" : kill_mem,
                        "max" : mem_max,
                        "max_vm" : mem_vm_max,
                        "err" : error.to_string());
                }
                if kill_mem {
                    #[allow(clippy::cast_possible_wrap)]
                    let _ = kill(Pid::from_raw(req.pid as i32), Some(SIGKILL));
                }
                return request.fail_syscall(nix::libc::ENOMEM);
            }
            _ => { /* fall through */ }
        }
    }

    // Check PSS
    match proc_mem_limit(&proc, mem_max) {
        Ok(false) => {
            // SAFETY: No pointer dereference in security check.
            // This is safe to continue.
            unsafe { request.continue_syscall() }
        }
        Ok(true) => {
            if !mem_filter {
                warn!("ctx" : "access",
                    "cap" : "m",
                    "pid" : req.pid,
                    "sys" : name,
                    "kill" : kill_mem,
                    "max" : mem_max,
                    "max_vm": mem_vm_max);
            }
            if kill_mem {
                #[allow(clippy::cast_possible_wrap)]
                let _ = kill(Pid::from_raw(req.pid as i32), Some(SIGKILL));
            }
            request.fail_syscall(nix::libc::ENOMEM)
        }
        Err(error) => {
            if !mem_filter {
                warn!("ctx" : "access",
                    "cap" : "m",
                    "sys" : name,
                    "pid" : req.pid,
                    "kill" : kill_mem,
                    "max" : mem_max,
                    "max_vm" : mem_vm_max,
                    "err" : error.to_string());
            }
            if kill_mem {
                #[allow(clippy::cast_possible_wrap)]
                let _ = kill(Pid::from_raw(req.pid as i32), Some(SIGKILL));
            }
            request.fail_syscall(nix::libc::ENOMEM)
        }
    }
}

fn sys_fork(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_pid_handler(request, "fork")
}

fn sys_vfork(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_pid_handler(request, "vfork")
}

fn sys_clone(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_pid_handler(request, "clone")
}

fn sys_clone3(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_pid_handler(request, "clone3")
}

#[allow(clippy::cognitive_complexity)]
fn syscall_pid_handler(request: &UNotifyEventRequest, name: &str) -> ScmpNotifResp {
    // Get pid max.
    let sandbox = request.get_sandbox(false);
    let enabled = sandbox.enabled(Capability::CAP_PID);
    let pid_max = sandbox.pid_max;
    let pid_filter = sandbox.pid_filter;
    let kill_pid = sandbox.kill_pid();
    drop(sandbox); // release the read lock.

    if !enabled || pid_max == 0 {
        // SAFETY: No pointer dereference in security check.
        // This is safe to continue.
        return unsafe { request.continue_syscall() };
    }

    // Get pid count.
    let req = request.get_request();
    match proc_task_limit(pid_max) {
        Ok(false) => {
            // SAFETY: No pointer dereference in security check.
            // This is safe to continue.
            unsafe { request.continue_syscall() }
        }
        Ok(true) => {
            if !pid_filter {
                warn!("ctx" : "access",
                    "cap" : "p",
                    "pid" : req.pid,
                    "sys" : name,
                    "max" : pid_max,
                    "kill" : kill_pid);
            }
            if kill_pid {
                #[allow(clippy::cast_possible_wrap)]
                let _ = kill(Pid::from_raw(req.pid as i32), Some(SIGKILL));
            }
            request.fail_syscall(nix::libc::EACCES)
        }
        Err(error) => {
            if !pid_filter {
                warn!("ctx" : "access",
                    "cap" : "p",
                    "sys" : name,
                    "pid" : req.pid,
                    "max" : pid_max,
                    "kill" : kill_pid,
                    "err" : error.to_string());
            }
            if kill_pid {
                #[allow(clippy::cast_possible_wrap)]
                let _ = kill(Pid::from_raw(req.pid as i32), Some(SIGKILL));
            }
            request.fail_syscall(nix::libc::EACCES)
        }
    }
}

fn sys_kill(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_signal_handler(request, false)
}

fn sys_tkill(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_signal_handler(request, false)
}

fn sys_tgkill(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_signal_handler(request, true)
}

fn sys_pidfd_open(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_signal_handler(request, false)
}

fn sys_socketcall(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let (req, proc) = request.prepare();

    // Determine system call name
    let name = match req.data.args[0] {
        0x2 => "bind",
        0x3 => "connect",
        0xb => "sendto",
        0xc => "recvfrom",
        _ => {
            // SAFETY: safe network call, continue.
            return unsafe { request.continue_syscall() };
        }
    };

    // Determine system call arguments
    let args = {
        let mut args = [0u32; 6];
        let size = std::mem::size_of_val(&args);
        let mut buf = vec![0u8; size];
        #[allow(clippy::cast_possible_truncation)]
        match proc.read_mem(&mut buf, req.data.args[1] as usize, request) {
            Ok(bytes_read) if bytes_read >= size => {
                // unsigned long is typically 32 bits (4 bytes) in size.
                for (i, chunk) in buf.chunks_exact(4).enumerate() {
                    match chunk.try_into() {
                        Ok(bytes) => args[i] = u32::from_ne_bytes(bytes),
                        Err(error) => {
                            error!("ctx": "socketcall_decode",
                                "pid": req.pid,
                                "addr": req.data.args[1],
                                "read": bytes_read,
                                "size": size,
                                "err": error.to_string());
                            return request.fail_syscall(nix::libc::EFAULT);
                        }
                    }
                }
                args.map(u64::from)
            }
            Ok(n) => {
                error!("ctx": "socketcall_decode",
                    "pid": req.pid,
                    "addr": req.data.args[1],
                    "read": n,
                    "size": size,
                    "err": "short read");
                return request.fail_syscall(nix::libc::EFAULT);
            }
            Err(error) => {
                error!("ctx": "socketcall_decode",
                    "pid": req.pid,
                    "addr": req.data.args[1],
                    "read": 0,
                    "size": size,
                    "err": error.to_string());
                return request.fail_syscall(nix::libc::EFAULT);
            }
        }
    };

    // SAFETY: If addr or receiver/send buffer is NULL,
    // we must return EFAULT here without further processing.
    if args[1] == 0 {
        return request.fail_syscall(nix::libc::EFAULT);
    }

    match req.data.args[0] {
        0x2 => {
            // bind
            syscall_network_handler(request, &args, 1, 2, false, name, sandbox_addr)
        }
        0x3 => {
            // connect
            syscall_network_handler(request, &args, 1, 2, false, name, sandbox_addr)
        }
        0xb => {
            // sendto
            syscall_network_handler(request, &args, 4, 5, false, name, sandbox_addr)
        }
        0xc => {
            // recvfrom
            syscall_network_handler(request, &args, 4, 5, true, name, sandbox_addr)
        }
        _ => unreachable!(),
    }
}

fn sys_bind(request: &UNotifyEventRequest) -> ScmpNotifResp {
    // SAFETY: If addr is NULL we must return EFAULT
    // here without further processing.
    let req = request.get_request();
    if req.data.args[1] == 0 {
        return request.fail_syscall(nix::libc::EFAULT);
    }
    syscall_network_handler(request, &req.data.args, 1, 2, false, "bind", sandbox_addr)
}

fn sys_connect(request: &UNotifyEventRequest) -> ScmpNotifResp {
    // SAFETY: If addr is NULL we must return EFAULT
    // here without further processing.
    let req = request.get_request();
    if req.data.args[1] == 0 {
        return request.fail_syscall(nix::libc::EFAULT);
    }
    syscall_network_handler(
        request,
        &req.data.args,
        1,
        2,
        false,
        "connect",
        sandbox_addr,
    )
}

fn sys_recvfrom(request: &UNotifyEventRequest) -> ScmpNotifResp {
    // SAFETY: If recieve buffer is NULL we must
    // return EFAULT here without further
    // processing.
    let req = request.get_request();
    if req.data.args[1] == 0 {
        return request.fail_syscall(nix::libc::EFAULT);
    }
    syscall_network_handler(
        request,
        &req.data.args,
        4,
        5,
        true,
        "recvfrom",
        sandbox_addr,
    )
}

fn sys_sendto(request: &UNotifyEventRequest) -> ScmpNotifResp {
    // SAFETY: If send buffer is NULL we must
    // return EFAULT here without further
    // processing.
    let req = request.get_request();
    if req.data.args[1] == 0 {
        return request.fail_syscall(nix::libc::EFAULT);
    }
    syscall_network_handler(request, &req.data.args, 4, 5, false, "sendto", sandbox_addr)
}

fn sys_execve(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(request, "execve", argv, sandbox_path_1, |_, _, _, _| {
        // SAFETY: This is vulnerable to TOCTOU,
        // unfortunately we cannot emulate exec,
        // so we have to live with it...
        Ok(unsafe { request.continue_syscall() })
    })
}

fn sys_execveat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();
    let empty = req.data.args[4] & nix::libc::AT_EMPTY_PATH as u64 != 0;
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: false,
        empty,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(request, "execveat", argv, sandbox_path_1, |_, _, _, _| {
        // SAFETY: This is vulnerable to TOCTOU,
        // unfortunately we cannot emulate exec,
        // so we have to live with it...
        Ok(unsafe { request.continue_syscall() })
    })
}

fn sys_chdir(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(request, "chdir", argv, sandbox_path_1, |_, _, _, _| {
        // SAFETY: This is vulnerable to TOCTOU,
        // unfortunately we cannot emulate chdir,
        // so we have to live with it...
        Ok(unsafe { request.continue_syscall() })
    })
}

fn sys_fchdir(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: None,
        null: false,
        empty: true,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(request, "chdir", argv, sandbox_path_1, |_, _, _, _| {
        // SAFETY: This is vulnerable to TOCTOU,
        // unfortunately we cannot emulate fchdir,
        // so we have to live with it...
        Ok(unsafe { request.continue_syscall() })
    })
}

fn sys_getdents(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_handler!(request, |req: &ScmpNotifReq, proc: &RemoteProcess| {
        if req.data.args[1] == 0 {
            // SAFETY: If the second argument which must hold a pointer to a
            // linux_dirent structure is NULL, we must return EFAULT
            // without further processing here.
            return Err(Errno::EFAULT);
        } else if req.data.args[2] == 0 {
            // SAFETY:Result buffer is too small
            return Err(Errno::EINVAL);
        }

        // If sandboxing for List capability is off, return immediately.
        let sandbox = request.get_sandbox(false);
        #[allow(clippy::cast_possible_wrap)]
        let is_lock = sandbox.locked_for_pid(req.pid as nix::libc::pid_t);
        let is_stat = sandbox.enabled(Capability::CAP_STAT);
        drop(sandbox); // release the read lock.
        if is_lock && !is_stat {
            // SAFETY: This is unsafe due to vulnerability to TOCTOU.
            // Since stat sandboxing is disabled here, this is safe.
            return unsafe { Ok(request.continue_syscall()) };
        }

        let fd = proc.get_fd(req.data.args[0] as RawFd)?;
        let dir = match crate::fs::canonicalize(
            format!("/proc/self/fd/{}", fd.as_raw_fd()),
            true,
            crate::fs::MissingHandling::Missing,
        ) {
            Ok(dir) => dir,
            Err(error) => {
                error!("ctx": "canonicalize", "fd": req.data.args[0], "err": format!("{error}"));
                return Err(Errno::ENOENT);
            }
        };

        // SAFETY: The count argument to the getdents call
        // must not be fully trusted, it can be overly large,
        // and allocating a Vector of that capacity may overflow.
        // This bug was discovered by trinity in this build:
        // https://builds.sr.ht/~alip/job/1077263
        #[allow(clippy::cast_possible_truncation)]
        let count = req.data.args[2] as usize;
        let count = count.min(1000000); // Cap count at 1mio
        let mut entries = Vec::with_capacity(count.saturating_add(1) /* /dev/syd */);
        loop {
            match getdents(fd.as_raw_fd(), count) {
                Err(error) => {
                    return Err(error);
                }
                Ok(None) => {
                    // getdents returned None, there are no more entries.
                    return Ok(request.return_syscall(0));
                }
                Ok(Some(e)) => {
                    for entry in e.iter() {
                        let name = entry.to_path_buf();
                        let path = dir.join(&name);
                        if is_stat {
                            match sandbox_path_1(
                                &nonempty![path.clone()],
                                Capability::CAP_STAT,
                                "getdents",
                                request,
                                proc,
                            ) {
                                Ok(_) /*allow*/ => {
                                    trace!("ctx": "stat", "path": format!("{}", path.display()), "act": "allow");
                                    entries.push(entry.clone());
                                },
                                _ => {
                                    /* this entry is denied, skip it. */
                                    debug!("ctx": "stat", "path": format!("{}", path.display()), "act": "hide");
                                },
                            };
                        } else {
                            // Stat sandboxing is off.
                            // We're here because the lock is off.
                            trace!("ctx": "stat", "path": format!("{}", path.display()), "act": "allow");
                            entries.push(entry.clone());
                        }
                    }
                    if !entries.is_empty() {
                        break; // exit the loop once we have allowed entries
                    }
                }
            };
        }

        let list_syd = !is_lock
            && crate::DEV_STAT.map_or(false, |(dev, ino)| {
                fstat(fd.as_raw_fd()).map_or(false, |metadata| {
                    metadata.st_dev == dev && metadata.st_ino == ino
                })
            });

        if list_syd {
            crate::compat::DirEntry::append_dev_syd(&mut entries);
        }
        crate::compat::DirEntry::fix_offsets(&mut entries);

        let mut buffer = Vec::new();
        for entry in &entries {
            let bytes = &entry.dirent;

            // Ensure we don't append more bytes than the buffer can hold.
            if buffer.len().saturating_add(bytes.len()) > count {
                break;
            }

            buffer.extend_from_slice(bytes);
        }

        #[allow(clippy::cast_possible_truncation)]
        proc.write_mem(&buffer, req.data.args[1] as usize, request)?;

        #[allow(clippy::cast_possible_wrap)]
        Ok(request.return_syscall(buffer.len() as i64))
    })
}

fn sys_access(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "access",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            #[allow(clippy::cast_possible_truncation)]
            let mode = AccessFlags::from_bits_truncate(req.data.args[1] as nix::libc::c_int);
            faccessat(Some(fd), &name, mode, AtFlags::empty()).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_faccessat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "faccessat",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            #[allow(clippy::cast_possible_truncation)]
            let mode = AccessFlags::from_bits_truncate(req.data.args[2] as nix::libc::c_int);
            faccessat(Some(fd), &name, mode, AtFlags::empty()).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_faccessat2(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();
    let resolve = req.data.args[3] & nix::libc::AT_SYMLINK_NOFOLLOW as u64 == 0;
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: false,
        empty: false,
        dotlast: None,
        resolve,
    });
    syscall_path_handler(
        request,
        "faccessat2",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            #[allow(clippy::cast_possible_truncation)]
            let mode = AccessFlags::from_bits_truncate(req.data.args[2] as nix::libc::c_int);
            #[allow(clippy::cast_possible_truncation)]
            let flags = AtFlags::from_bits_truncate(req.data.args[3] as nix::libc::c_int);
            faccessat(Some(fd), &name, mode, flags).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_chmod(request: &UNotifyEventRequest) -> ScmpNotifResp {
    // SAFETY: Disallow setting setuid/setgid bits as necessary.
    let sandbox = request.get_sandbox(false);
    let restrict_chmod = !sandbox.allow_unsafe_chmod();
    drop(sandbox);
    if restrict_chmod {
        let req = request.get_request();
        if req.data.args[1] & u64::from(nix::libc::S_ISUID | nix::libc::S_ISGID) != 0 {
            return request.fail_syscall(nix::libc::EACCES);
        }
    }
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: false,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "chmod",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            #[allow(clippy::cast_possible_truncation)]
            let mode = Mode::from_bits_truncate(req.data.args[1] as u32);
            fchmodat(Some(fd), &name, mode, FchmodatFlags::FollowSymlink)
                .map(|_| request.return_syscall(0))
        },
    )
}

fn sys_fchmod(request: &UNotifyEventRequest) -> ScmpNotifResp {
    // SAFETY: Disallow setting setuid/setgid bits as necessary.
    let sandbox = request.get_sandbox(false);
    let restrict_chmod = !sandbox.allow_unsafe_chmod();
    drop(sandbox);
    if restrict_chmod {
        let req = request.get_request();
        if req.data.args[1] & u64::from(nix::libc::S_ISUID | nix::libc::S_ISGID) != 0 {
            return request.fail_syscall(nix::libc::EACCES);
        }
    }
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: None,
        null: false,
        empty: true,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "fchmod",
        argv,
        sandbox_path_1,
        |_, _, request, proc| {
            let req = request.get_request();
            #[allow(clippy::cast_possible_truncation)]
            let mode = Mode::from_bits_truncate(req.data.args[1] as u32);
            let fd = proc.get_fd(req.data.args[0] as RawFd)?;
            fchmod(fd.as_raw_fd(), mode).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_fchmodat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();

    // SAFETY: Disallow setting setuid/setgid bits as necessary.
    let sandbox = request.get_sandbox(false);
    let restrict_chmod = !sandbox.allow_unsafe_chmod();
    drop(sandbox);
    if restrict_chmod && req.data.args[2] & u64::from(nix::libc::S_ISUID | nix::libc::S_ISGID) != 0
    {
        return request.fail_syscall(nix::libc::EACCES);
    }
    let empty = req.data.args[3] & nix::libc::AT_EMPTY_PATH as u64 != 0;
    let resolve = req.data.args[3] & nix::libc::AT_SYMLINK_NOFOLLOW as u64 == 0;
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: false,
        dotlast: None,
        empty,
        resolve,
    });
    syscall_path_handler(
        request,
        "fchmodat",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            #[allow(clippy::cast_possible_truncation)]
            let mode = Mode::from_bits_truncate(req.data.args[2] as u32);
            let fd = fds.first().as_raw_fd();
            let path = paths.first();
            if !path.is_empty() {
                let name = crate::path_filename(path);
                let flags = if resolve {
                    FchmodatFlags::FollowSymlink
                } else {
                    FchmodatFlags::NoFollowSymlink
                };
                fchmodat(Some(fd), &name, mode, flags)
            } else {
                fchmod(fd, mode)
            }
            .map(|_| request.return_syscall(0))
        },
    )
}

fn sys_chown(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "chown",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            #[allow(clippy::cast_possible_wrap)]
            let owner = if req.data.args[1] as i64 == -1 {
                None
            } else {
                Some(Uid::from_raw(req.data.args[1] as nix::libc::uid_t))
            };
            #[allow(clippy::cast_possible_wrap)]
            let group = if req.data.args[2] as i64 == -1 {
                None
            } else {
                Some(Gid::from_raw(req.data.args[2] as nix::libc::gid_t))
            };
            fchownat(Some(fd), &name, owner, group, FchownatFlags::FollowSymlink)
                .map(|_| request.return_syscall(0))
        },
    )
}

fn sys_lchown(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: false,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "chown",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            #[allow(clippy::cast_possible_wrap)]
            let owner = if req.data.args[1] as i64 == -1 {
                None
            } else {
                Some(Uid::from_raw(req.data.args[1] as nix::libc::uid_t))
            };
            #[allow(clippy::cast_possible_wrap)]
            let group = if req.data.args[2] as i64 == -1 {
                None
            } else {
                Some(Gid::from_raw(req.data.args[2] as nix::libc::gid_t))
            };
            fchownat(
                Some(fd),
                &name,
                owner,
                group,
                FchownatFlags::NoFollowSymlink,
            )
            .map(|_| request.return_syscall(0))
        },
    )
}

fn sys_fchown(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: None,
        null: false,
        empty: true,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "fchown",
        argv,
        sandbox_path_1,
        |_, _, request, proc| {
            let req = request.get_request();
            #[allow(clippy::cast_possible_wrap)]
            let owner = if req.data.args[1] as i64 == -1 {
                None
            } else {
                Some(Uid::from_raw(req.data.args[1] as nix::libc::uid_t))
            };
            #[allow(clippy::cast_possible_wrap)]
            let group = if req.data.args[2] as i64 == -1 {
                None
            } else {
                Some(Gid::from_raw(req.data.args[2] as nix::libc::gid_t))
            };
            let fd = proc.get_fd(req.data.args[0] as RawFd)?;
            fchown(fd.as_raw_fd(), owner, group).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_fchownat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();
    let empty = req.data.args[4] & nix::libc::AT_EMPTY_PATH as u64 != 0;
    let resolve = req.data.args[4] & nix::libc::AT_SYMLINK_NOFOLLOW as u64 == 0;
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: false,
        dotlast: None,
        empty,
        resolve,
    });
    syscall_path_handler(
        request,
        "fchownat",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            #[allow(clippy::cast_possible_wrap)]
            let owner = if req.data.args[2] as i64 == -1 {
                None
            } else {
                Some(Uid::from_raw(req.data.args[2] as nix::libc::uid_t))
            };
            #[allow(clippy::cast_possible_wrap)]
            let group = if req.data.args[3] as i64 == -1 {
                None
            } else {
                Some(Gid::from_raw(req.data.args[3] as nix::libc::gid_t))
            };
            let path = paths.first();
            if !path.is_empty() {
                let name = crate::path_filename(path);
                let flags = if resolve {
                    FchownatFlags::FollowSymlink
                } else {
                    FchownatFlags::NoFollowSymlink
                };
                fchownat(Some(fd), &name, owner, group, flags)
            } else {
                fchown(fd, owner, group)
            }
            .map(|_| request.return_syscall(0))
        },
    )
}

fn sys_creat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_open_handler(request, "creat")
}

fn sys_link(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = nonempty![
        SyscallPathArgument {
            dirfd: None,
            path: Some(0),
            null: false,
            empty: false,
            resolve: false,
            dotlast: None,
        },
        SyscallPathArgument {
            dirfd: None,
            path: Some(1),
            null: false,
            empty: false,
            resolve: false,
            dotlast: Some(Errno::ENOENT),
        }
    ];
    syscall_path_handler(
        request,
        "link",
        argv,
        sandbox_path_2,
        |fds, paths, request, _proc| {
            let fd_old = fds.first().as_raw_fd();
            let fd_new = fds.last().as_raw_fd();
            let name_old = crate::path_filename(paths.first());
            let name_new = crate::path_filename(paths.last());

            linkat(
                Some(fd_old),
                &name_old,
                Some(fd_new),
                &name_new,
                LinkatFlags::NoSymlinkFollow,
            )
            .map(|_| request.return_syscall(0))
        },
    )
}

fn sys_symlink(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_handler!(request, |req: &ScmpNotifReq, proc: &RemoteProcess| {
        // SAFETY: No checking of the target is done.
        // This is consistent with the system call.
        let path_arg = SyscallPathArgument {
            dirfd: None,
            path: Some(1),
            null: false,
            empty: false,
            resolve: false,
            dotlast: None,
        };

        // If sandboxing for all the selected capabilities is off, return immediately.
        let mut check = false;
        let sandbox = request.get_sandbox(false);
        if sandbox.enabled(Capability::CAP_WRITE) {
            check = true;
        }
        drop(sandbox); // release the read lock.
        if !check {
            // SAFETY: This is unsafe due to vulnerability to TOCTOU,
            // however since sandboxing for the respective Capability
            // is off, this is safe here.
            return unsafe { Ok(request.continue_syscall()) };
        }

        // SAFETY: symlink() returns ENOENT if target is an empty string.
        let target = remote_path_n!(proc, req, 0, &request)?;
        if target.is_empty() {
            return Err(Errno::ENOENT);
        }
        let target = OsStr::from_bytes(target.to_bytes());

        let path = proc.read_path(request, &path_arg)?;
        sandbox_path_1(
            &nonempty![path.clone()],
            Capability::CAP_WRITE,
            "symlink",
            request,
            proc,
        )?;

        // SAFETY: We have already resolved symbolic
        // links as necessary, from this point on we
        // do not resolve any longer for safety as
        // the filesystem may change which may
        // result in a TOCTOU.
        let mut how = openat2::OpenHow::new(nix::libc::O_PATH, 0);
        how.resolve = openat2::ResolveFlags::NO_SYMLINKS;
        let parent = path.parent().unwrap_or(&path);
        let newdirfd = openat2::openat2(None, parent, &how)
            .map_err(|e| Errno::from_i32(e.raw_os_error().unwrap_or(nix::libc::EINVAL)))?;
        // SAFETY: Valid FD
        let newdirfd = unsafe { OwnedFd::from_raw_fd(newdirfd) };
        let linkpath = crate::path_filename(path);

        symlinkat(target, Some(newdirfd.as_raw_fd()), &linkpath).map(|_| request.return_syscall(0))
    })
}

fn sys_unlink(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: false,
        dotlast: Some(Errno::EINVAL),
    });
    syscall_path_handler(
        request,
        "unlink",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            unlinkat(Some(fd), &name, UnlinkatFlags::NoRemoveDir).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_linkat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();
    #[allow(clippy::cast_possible_truncation)]
    let flags = req.data.args[4] as nix::libc::c_int;
    let empty = flags & nix::libc::AT_EMPTY_PATH != 0;
    let resolve = flags & nix::libc::AT_SYMLINK_FOLLOW != 0;
    let argv = nonempty![
        SyscallPathArgument {
            dirfd: Some(0),
            path: Some(1),
            null: false,
            dotlast: None,
            empty,
            resolve,
        },
        SyscallPathArgument {
            dirfd: Some(2),
            path: Some(3),
            null: false,
            empty: false,
            resolve: false,
            dotlast: Some(Errno::ENOENT),
        }
    ];
    syscall_path_handler(
        request,
        "linkat",
        argv,
        sandbox_path_2,
        |fds, paths, request, _proc| {
            let fd_old = fds.first().as_raw_fd();
            let fd_new = fds.last().as_raw_fd();
            let name_old = crate::path_filename(paths.first());
            let name_new = crate::path_filename(paths.last());
            let flags = if resolve {
                LinkatFlags::SymlinkFollow
            } else {
                LinkatFlags::NoSymlinkFollow
            };
            linkat(Some(fd_old), &name_old, Some(fd_new), &name_new, flags)
                .map(|_| request.return_syscall(0))
        },
    )
}

fn sys_symlinkat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_handler!(request, |req: &ScmpNotifReq, proc: &RemoteProcess| {
        // SAFETY: No checking of the target is done.
        // This is consistent with the system call.
        let path_arg = SyscallPathArgument {
            dirfd: Some(1),
            path: Some(2),
            null: false,
            empty: false,
            resolve: false,
            dotlast: None,
        };

        // If sandboxing for all the selected capabilities is off, return immediately.
        let mut check = false;
        let sandbox = request.get_sandbox(false);
        if sandbox.enabled(Capability::CAP_WRITE) {
            check = true;
        }
        drop(sandbox); // release the read lock.
        if !check {
            // SAFETY: This is unsafe due to vulnerability to TOCTOU,
            // however since sandboxing for the respective Capability
            // is off, this is safe here.
            return unsafe { Ok(request.continue_syscall()) };
        }

        // SAFETY: symlinkat() returns ENOENT if target is an empty string.
        let target = remote_path_n!(proc, req, 0, &request)?;
        if target.is_empty() {
            return Err(Errno::ENOENT);
        }
        let target = OsStr::from_bytes(target.to_bytes());

        let path = proc.read_path(request, &path_arg)?;
        sandbox_path_1(
            &nonempty![path.clone()],
            Capability::CAP_WRITE,
            "symlinkat",
            request,
            proc,
        )?;

        // SAFETY: We have already resolved symbolic
        // links as necessary, from this point on we
        // do not resolve any longer for safety as
        // the filesystem may change which may
        // result in a TOCTOU.
        let mut how = openat2::OpenHow::new(nix::libc::O_PATH, 0);
        how.resolve = openat2::ResolveFlags::NO_SYMLINKS;
        let parent = path.parent().unwrap_or(&path);
        let newdirfd = openat2::openat2(None, parent, &how)
            .map_err(|e| Errno::from_i32(e.raw_os_error().unwrap_or(nix::libc::EINVAL)))?;
        // SAFETY: Valid FD
        let newdirfd = unsafe { OwnedFd::from_raw_fd(newdirfd) };
        let linkpath = crate::path_filename(path);

        symlinkat(target, Some(newdirfd.as_raw_fd()), &linkpath).map(|_| request.return_syscall(0))
    })
}

fn sys_unlinkat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: false,
        empty: false,
        resolve: false,
        dotlast: Some(Errno::EINVAL),
    });
    syscall_path_handler(
        request,
        "unlinkat",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            #[allow(clippy::cast_possible_truncation)]
            let flags = if req.data.args[2] as nix::libc::c_int & nix::libc::AT_REMOVEDIR != 0 {
                UnlinkatFlags::RemoveDir
            } else {
                UnlinkatFlags::NoRemoveDir
            };
            unlinkat(Some(fd), &name, flags).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_mkdir(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: false,
        dotlast: Some(Errno::ENOENT),
    });
    syscall_path_handler(
        request,
        "mkdir",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            // SAFETY: SydB☮x' umask is 0 here so we respect process' umask.
            #[allow(clippy::cast_possible_wrap)]
            let pid = Pid::from_raw(req.pid as i32);
            let mask = proc_umask(Some(pid))?.bits();
            #[allow(clippy::cast_possible_truncation)]
            #[allow(clippy::cast_sign_loss)]
            let mode =
                Mode::from_bits_truncate((req.data.args[1] as nix::libc::c_int as u32) & !mask);
            mkdirat(fd, &name, mode).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_rmdir(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: Some(Errno::EINVAL),
    });
    syscall_path_handler(
        request,
        "rmdir",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            unlinkat(Some(fd), &name, UnlinkatFlags::RemoveDir).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_mkdirat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: false,
        empty: false,
        resolve: false,
        dotlast: Some(Errno::ENOENT),
    });
    syscall_path_handler(
        request,
        "mkdirat",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            // SAFETY: SydB☮x' umask is 0 here so we respect process' umask.
            #[allow(clippy::cast_possible_wrap)]
            let pid = Pid::from_raw(req.pid as i32);
            let mask = proc_umask(Some(pid))?.bits();
            #[allow(clippy::cast_possible_truncation)]
            #[allow(clippy::cast_sign_loss)]
            let mode =
                Mode::from_bits_truncate((req.data.args[2] as nix::libc::c_int as u32) & !mask);
            mkdirat(fd, &name, mode).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_mknod(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();
    #[allow(clippy::cast_possible_truncation)]
    let kind =
        SFlag::from_bits_truncate(req.data.args[1] as nix::libc::mode_t & SFlag::S_IFMT.bits());
    if !matches!(kind, SFlag::S_IFIFO | SFlag::S_IFREG | SFlag::S_IFSOCK) {
        // SAFETY: We do not allow:
        // 1. Device special files because this may
        //    circumvent path based access control.
        return request.fail_syscall(nix::libc::EACCES);
    }
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "mknod",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            // SAFETY: SydB☮x' umask is 0 here so we respect process' umask.
            #[allow(clippy::cast_possible_wrap)]
            let pid = Pid::from_raw(req.pid as i32);
            let mask = proc_umask(Some(pid))?.bits();
            #[allow(clippy::cast_possible_truncation)]
            let perm = Mode::from_bits_truncate(req.data.args[1] as u32 & !SFlag::S_IFMT.bits());
            let perm = Mode::from_bits_truncate(perm.bits() & !mask);
            #[allow(clippy::cast_possible_truncation)]
            let dev = req.data.args[2] as nix::libc::dev_t;
            mknodat(fd, &name, kind, perm, dev).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_mknodat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();
    #[allow(clippy::cast_possible_truncation)]
    let kind =
        SFlag::from_bits_truncate(req.data.args[2] as nix::libc::mode_t & SFlag::S_IFMT.bits());
    if !matches!(kind, SFlag::S_IFIFO | SFlag::S_IFREG | SFlag::S_IFSOCK) {
        // SAFETY: We do not allow:
        // 1. Device special files because this may
        //    circumvent path based access control.
        return request.fail_syscall(nix::libc::EACCES);
    }
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "mknodat",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            // SAFETY: SydB☮x' umask is 0 here so we respect process' umask.
            #[allow(clippy::cast_possible_wrap)]
            let pid = Pid::from_raw(req.pid as i32);
            let mask = proc_umask(Some(pid))?.bits();
            #[allow(clippy::cast_possible_truncation)]
            let perm = Mode::from_bits_truncate(req.data.args[2] as u32 & !SFlag::S_IFMT.bits());
            let perm = Mode::from_bits_truncate(perm.bits() & !mask);
            #[allow(clippy::cast_possible_truncation)]
            let dev = req.data.args[3] as nix::libc::dev_t;
            mknodat(fd, &name, kind, perm, dev).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_mount(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = nonempty![
        SyscallPathArgument {
            dirfd: None,
            path: Some(0),
            null: true,
            empty: false,
            resolve: true,
            dotlast: None,
        },
        SyscallPathArgument {
            dirfd: None,
            path: Some(1),
            null: false,
            empty: false,
            resolve: true,
            dotlast: None,
        }
    ];
    syscall_path_handler(
        request,
        "mount",
        argv,
        sandbox_path_2,
        |_fds, paths, request, proc| {
            let req = request.get_request();
            let source: Option<&PathBuf> = if req.data.args[0] == 0 {
                None
            } else {
                Some(paths.first())
            };
            let target = if req.data.args[1] == 0 {
                return Err(Errno::EFAULT);
            } else {
                paths.last()
            };
            let fstype = if req.data.args[2] == 0 {
                None
            } else {
                let mut fstype = vec![0u8; nix::libc::PATH_MAX as usize];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut fstype, req.data.args[2] as usize, request)?;
                Some(PathBuf::from(OsString::from_vec(fstype)))
            };
            let flags = MsFlags::from_bits_truncate(req.data.args[3] as nix::libc::c_ulong);
            let data = if req.data.args[4] == 0 {
                None
            } else {
                let mut data = vec![0u8; nix::libc::PATH_MAX as usize];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut data, req.data.args[4] as usize, request)?;
                Some(PathBuf::from(OsString::from_vec(data)))
            };
            mount(source, target, fstype.as_ref(), flags, data.as_ref())
                .map(|_| request.return_syscall(0))
        },
    )
}

fn sys_umount(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "umount",
        argv,
        sandbox_path_1,
        |_fds, paths, request, _proc| {
            let path = paths.first();
            umount(path).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_umount2(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();
    #[allow(clippy::cast_possible_truncation)]
    let flags = MntFlags::from_bits_truncate(req.data.args[1] as nix::libc::c_int);
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: !flags.contains(MntFlags::UMOUNT_NOFOLLOW),
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "umount2",
        argv,
        sandbox_path_1,
        |_fds, paths, request, _proc| {
            let path = paths.first();
            umount2(path, flags).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_open(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_open_handler(request, "open")
}

fn sys_openat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_open_handler(request, "openat")
}

fn sys_openat2(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_open_handler(request, "openat2")
}

fn sys_rename(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = nonempty![
        SyscallPathArgument {
            dirfd: None,
            path: Some(0),
            null: false,
            empty: false,
            resolve: false,
            dotlast: Some(Errno::EINVAL),
        },
        SyscallPathArgument {
            dirfd: None,
            path: Some(1),
            null: false,
            empty: false,
            resolve: false,
            dotlast: Some(Errno::EINVAL),
        }
    ];
    syscall_path_handler(
        request,
        "rename",
        argv,
        sandbox_path_2,
        |fds, paths, request, _proc| {
            let fd_old = fds.first().as_raw_fd();
            let fd_new = fds.last().as_raw_fd();
            let name_old = crate::path_filename(paths.first());
            let name_new = crate::path_filename(paths.last());
            renameat(Some(fd_old), &name_old, Some(fd_new), &name_new)
                .map(|_| request.return_syscall(0))
        },
    )
}

fn sys_renameat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = nonempty![
        SyscallPathArgument {
            dirfd: Some(0),
            path: Some(1),
            null: false,
            empty: false,
            resolve: false,
            dotlast: Some(Errno::EINVAL),
        },
        SyscallPathArgument {
            dirfd: Some(2),
            path: Some(3),
            null: false,
            empty: false,
            resolve: false,
            dotlast: Some(Errno::EINVAL),
        }
    ];
    syscall_path_handler(
        request,
        "renameat",
        argv,
        sandbox_path_2,
        |fds, paths, request, _proc| {
            let fd_old = fds.first().as_raw_fd();
            let fd_new = fds.last().as_raw_fd();
            let name_old = crate::path_filename(paths.first());
            let name_new = crate::path_filename(paths.last());
            renameat(Some(fd_old), &name_old, Some(fd_new), &name_new)
                .map(|_| request.return_syscall(0))
        },
    )
}

fn sys_renameat2(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = nonempty![
        SyscallPathArgument {
            dirfd: Some(0),
            path: Some(1),
            null: false,
            empty: false,
            resolve: false,
            dotlast: Some(Errno::EINVAL),
        },
        SyscallPathArgument {
            dirfd: Some(2),
            path: Some(3),
            null: false,
            empty: false,
            resolve: false,
            dotlast: Some(Errno::EINVAL),
        }
    ];
    syscall_path_handler(
        request,
        "renameat2",
        argv,
        sandbox_path_2,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd_old = fds.first().as_raw_fd();
            let fd_new = fds.last().as_raw_fd();
            let name_old = crate::path_filename(paths.first());
            let name_new = crate::path_filename(paths.last());
            let name_old =
                CString::new(name_old.as_os_str().as_bytes()).map_err(|_| Errno::EINVAL)?;
            let name_new =
                CString::new(name_new.as_os_str().as_bytes()).map_err(|_| Errno::EINVAL)?;
            #[allow(clippy::cast_possible_truncation)]
            let flags = req.data.args[4] as u32;
            // Note: musl does not have renameat2 yet.
            // SAFETY: The call to `nix::libc::syscall` for
            // `SYS_renameat2` is safe provided the arguments are
            // correct. `fd_old` and `fd_new` are assumed to be valid
            // file descriptors, and `name_old.as_ptr()` and
            // `name_new.as_ptr()` provide valid pointers to
            // null-terminated strings. `flags` is a valid flag for the
            // syscall. Assuming these conditions, the syscall does not
            // lead to undefined behavior.
            if unsafe {
                nix::libc::syscall(
                    nix::libc::SYS_renameat2,
                    fd_old,
                    name_old.as_ptr(),
                    fd_new,
                    name_new.as_ptr(),
                    flags,
                )
            } == 0
            {
                Ok(request.return_syscall(0))
            } else {
                Err(Errno::last())
            }
        },
    )
}

fn sys_stat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_stat_handler(request, "stat", argv)
}

fn sys_fstat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: None,
        null: false,
        empty: true,
        resolve: true,
        dotlast: None,
    });
    syscall_stat_handler(request, "fstat", argv)
}

fn sys_lstat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: false,
        dotlast: None,
    });
    syscall_stat_handler(request, "lstat", argv)
}

fn sys_statx(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();
    let empty = req.data.args[2] & nix::libc::AT_EMPTY_PATH as u64 != 0;
    let resolve = req.data.args[2] & nix::libc::AT_SYMLINK_NOFOLLOW as u64 == 0;
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: false,
        dotlast: None,
        empty,
        resolve,
    });
    syscall_stat_handler(request, "statx", argv)
}

fn sys_newfstatat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();
    let empty = req.data.args[3] & nix::libc::AT_EMPTY_PATH as u64 != 0;
    let resolve = req.data.args[3] & nix::libc::AT_SYMLINK_NOFOLLOW as u64 == 0;
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: false,
        dotlast: None,
        empty,
        resolve,
    });
    syscall_stat_handler(request, "newfstatat", argv)
}

fn sys_utime(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "utime",
        argv,
        sandbox_path_1,
        |fds, paths, request, proc| {
            let req = request.get_request();
            let name = if req.data.args[0] != 0 {
                Some(crate::path_filename(paths.first()))
            } else {
                None
            };
            let name = if let Some(name) = name {
                Some(CString::new(name.as_os_str().as_bytes()).map_err(|_| Errno::EINVAL)?)
            } else {
                None
            };
            #[allow(clippy::cast_possible_truncation)]
            let time = match proc.remote_utimbuf(req.data.args[1] as usize, request)? {
                Some(time) => time.as_ptr(),
                None => std::ptr::null(),
            };
            let r = if let Some(name) = name {
                let (fd, name) = (fds.first(), name.as_ptr());
                // SAFETY: To be able to pass the arguments
                // path and time as NULL to the system call
                // we need to directly call it here.
                unsafe {
                    nix::libc::syscall(nix::libc::SYS_utimensat, fd.as_raw_fd(), name, time, 0)
                }
            } else {
                let (fd, name) = (proc.get_fd(req.data.args[0] as RawFd)?, 0);
                // SAFETY: The call to `nix::libc::syscall` for
                // `SYS_utimensat` is safe given that the arguments are
                // correct. 'fd.as_raw_fd()' is a valid file descriptor
                // encapsulated in an OwnedFd. 'name' is a valid
                // pointer (in this case, 0, representing a null
                // pointer). 'time' is assumed to be a valid pointer to
                // a timespec array or null. The last argument '0' is a
                // valid flag for the syscall. Assuming these
                // conditions, the syscall does not lead to undefined
                // behavior.
                unsafe {
                    nix::libc::syscall(nix::libc::SYS_utimensat, fd.as_raw_fd(), name, time, 0)
                }
            };
            if r == 0 {
                Ok(request.return_syscall(0))
            } else {
                Err(Errno::last())
            }
        },
    )
}

fn sys_utimes(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "utimes",
        argv,
        sandbox_path_1,
        |fds, paths, request, proc| {
            let req = request.get_request();
            let name = if req.data.args[0] != 0 {
                Some(crate::path_filename(paths.first()))
            } else {
                None
            };
            let name = if let Some(name) = name {
                Some(CString::new(name.as_os_str().as_bytes()).map_err(|_| Errno::EINVAL)?)
            } else {
                None
            };
            #[allow(clippy::cast_possible_truncation)]
            let time = match proc.remote_timeval(req.data.args[1] as usize, request)? {
                Some(time) => time.as_ptr(),
                None => std::ptr::null(),
            };
            let r = if let Some(name) = name {
                let (fd, name) = (fds.first(), name.as_ptr());
                // SAFETY: To be able to pass the arguments
                // path and time as NULL to the system call
                // we need to directly call it here.
                unsafe {
                    nix::libc::syscall(nix::libc::SYS_utimensat, fd.as_raw_fd(), name, time, 0)
                }
            } else {
                let (fd, name) = (proc.get_fd(req.data.args[0] as RawFd)?, 0);
                // SAFETY: The call to `nix::libc::syscall` for
                // `SYS_utimensat` is safe given that the arguments are
                // correct. 'fd.as_raw_fd()' is a valid file descriptor
                // encapsulated in an OwnedFd. 'name' is a valid
                // pointer (in this case, 0, representing a null
                // pointer). 'time' is assumed to be a valid pointer to
                // a timespec array or null. The last argument '0' is a
                // valid flag for the syscall. Assuming these
                // conditions, the syscall does not lead to undefined
                // behavior.
                unsafe {
                    nix::libc::syscall(nix::libc::SYS_utimensat, fd.as_raw_fd(), name, time, 0)
                }
            };
            if r == 0 {
                Ok(request.return_syscall(0))
            } else {
                Err(Errno::last())
            }
        },
    )
}

fn sys_futimesat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    // SAFETY: pathname may be NULL here:
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: true,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "futimesat",
        argv,
        sandbox_path_1,
        |fds, paths, request, proc| {
            let req = request.get_request();
            let name = if req.data.args[1] != 0 {
                Some(crate::path_filename(paths.first()))
            } else {
                None
            };
            let name = if let Some(name) = name {
                Some(CString::new(name.as_os_str().as_bytes()).map_err(|_| Errno::EINVAL)?)
            } else {
                None
            };
            #[allow(clippy::cast_possible_truncation)]
            let time = match proc.remote_timeval(req.data.args[2] as usize, request)? {
                Some(time) => time.as_ptr(),
                None => std::ptr::null(),
            };
            let r = if let Some(name) = name {
                let (fd, name) = (fds.first(), name.as_ptr());
                // SAFETY: To be able to pass the arguments
                // path and time as NULL to the system call
                // we need to directly call it here.
                unsafe {
                    nix::libc::syscall(nix::libc::SYS_utimensat, fd.as_raw_fd(), name, time, 0)
                }
            } else {
                let (fd, name) = (proc.get_fd(req.data.args[0] as RawFd)?, 0);
                // SAFETY: The call to `nix::libc::syscall` for
                // `SYS_utimensat` is safe given that the arguments are
                // correct. 'fd.as_raw_fd()' is a valid file descriptor
                // encapsulated in an OwnedFd. 'name' is a valid
                // pointer (in this case, 0, representing a null
                // pointer). 'time' is assumed to be a valid pointer to
                // a timespec array or null. The last argument '0' is a
                // valid flag for the syscall. Assuming these
                // conditions, the syscall does not lead to undefined
                // behavior.
                unsafe {
                    nix::libc::syscall(nix::libc::SYS_utimensat, fd.as_raw_fd(), name, time, 0)
                }
            };
            if r == 0 {
                Ok(request.return_syscall(0))
            } else {
                Err(Errno::last())
            }
        },
    )
}

fn sys_utimensat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();
    #[allow(clippy::cast_possible_truncation)]
    let resolve = req.data.args[3] as nix::libc::c_int & nix::libc::AT_SYMLINK_NOFOLLOW == 0;
    // SAFETY: pathname may be NULL here:
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: true,
        empty: false,
        dotlast: None,
        resolve,
    });
    syscall_path_handler(
        request,
        "utimensat",
        argv,
        sandbox_path_1,
        |fds, paths, request, proc| {
            let req = request.get_request();
            let name = if req.data.args[1] != 0 {
                Some(crate::path_filename(paths.first()))
            } else {
                None
            };
            let name = if let Some(name) = name {
                Some(CString::new(name.as_os_str().as_bytes()).map_err(|_| Errno::EINVAL)?)
            } else {
                None
            };
            #[allow(clippy::cast_possible_truncation)]
            let time = match proc.remote_timespec(req.data.args[2] as usize, request)? {
                Some(time) => time.as_ptr(),
                None => std::ptr::null(),
            };
            let r = if let Some(name) = name {
                let (fd, name) = (fds.first(), name.as_ptr());
                // SAFETY: To be able to pass the arguments
                // path and time as NULL to the system call
                // we need to directly call it here.
                unsafe {
                    nix::libc::syscall(
                        nix::libc::SYS_utimensat,
                        fd.as_raw_fd(),
                        name,
                        time,
                        req.data.args[3],
                    )
                }
            } else {
                let (fd, name) = (proc.get_fd(req.data.args[0] as RawFd)?, 0);
                // SAFETY: The call to `nix::libc::syscall` with
                // `SYS_utimensat` is safe provided that the arguments
                // are correct. `fd.as_raw_fd()` provides a valid file
                // descriptor. `name`, if not null, points to a valid
                // CString, and if null, it correctly represents a null
                // pointer. `time` is assumed to be a valid pointer to
                // a timespec array or null. The last argument,
                // `req.data.args[3]`, is expected to be a valid flag.
                // As long as these conditions are met, the syscall
                // should not lead to undefined behavior.
                unsafe {
                    nix::libc::syscall(
                        nix::libc::SYS_utimensat,
                        fd.as_raw_fd(),
                        name,
                        time,
                        req.data.args[3],
                    )
                }
            };
            if r == 0 {
                Ok(request.return_syscall(0))
            } else {
                Err(Errno::last())
            }
        },
    )
}

fn sys_truncate(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "truncate",
        argv,
        sandbox_path_1,
        |_fds, paths, request, _proc| {
            let req = request.get_request();
            #[allow(clippy::cast_possible_wrap)]
            let len = req.data.args[1] as nix::libc::off_t;
            let path = paths.first();
            truncate(path, len).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_ftruncate(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: None,
        null: false,
        empty: true,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "ftruncate",
        argv,
        sandbox_path_1,
        |_, _, request, proc| {
            let req = request.get_request();
            #[allow(clippy::cast_possible_wrap)]
            let len = req.data.args[1] as nix::libc::off_t;
            let fd = proc.get_fd(req.data.args[0] as RawFd)?;
            ftruncate(fd.as_raw_fd(), len).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_getxattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "getxattr",
        argv,
        sandbox_path_1,
        |_fds, paths, request, proc| {
            let req = request.get_request();
            let path = paths.first();
            let path = CString::new(path.as_os_str().as_bytes()).map_err(|_| Errno::ENOENT)?;
            let name = if req.data.args[1] != 0 {
                let mut buf = vec![0u8; nix::libc::PATH_MAX as usize];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[1] as usize, request)?;
                Some(buf)
            } else {
                None
            };
            let name = if let Some(ref name) = name {
                CStr::from_bytes_until_nul(name)
                    .map_err(|_| Errno::E2BIG)?
                    .as_ptr()
            } else {
                std::ptr::null()
            };
            // SAFETY: The size argument to the getxattr call
            // must not be fully trusted, it can be overly large,
            // and allocating a Vector of that capacity may overflow.
            #[allow(clippy::cast_possible_truncation)]
            let len = req.data.args[3] as usize;
            let len = len.min(nix::libc::PATH_MAX as usize); // Cap count at PATH_MAX
            let mut buf = if len == 0 { None } else { Some(vec![0u8; len]) };
            let ptr = match buf.as_mut() {
                Some(b) => b.as_mut_ptr(),
                None => std::ptr::null_mut(),
            };
            // SAFETY: In libc we trust.
            let n = unsafe {
                nix::libc::getxattr(
                    path.as_ptr(),
                    name,
                    ptr as *mut _ as *mut nix::libc::c_void,
                    len,
                )
            };
            let n = if n == -1 {
                return Err(Errno::last());
            } else {
                n as usize
            };
            if let Some(buf) = buf {
                #[allow(clippy::cast_possible_truncation)]
                proc.write_mem(&buf[..n], req.data.args[2] as usize, request)?;
            }
            #[allow(clippy::cast_possible_wrap)]
            Ok(request.return_syscall(n as i64))
        },
    )
}

fn sys_lgetxattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: false,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "lgetxattr",
        argv,
        sandbox_path_1,
        |_fds, paths, request, proc| {
            let req = request.get_request();
            let path = paths.first();
            let path = CString::new(path.as_os_str().as_bytes()).map_err(|_| Errno::ENOENT)?;
            let name = if req.data.args[1] != 0 {
                let mut buf = vec![0u8; nix::libc::PATH_MAX as usize];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[1] as usize, request)?;
                Some(buf)
            } else {
                None
            };
            let name = if let Some(ref name) = name {
                CStr::from_bytes_until_nul(name)
                    .map_err(|_| Errno::E2BIG)?
                    .as_ptr()
            } else {
                std::ptr::null()
            };
            // SAFETY: The size argument to the lgetxattr call
            // must not be fully trusted, it can be overly large,
            // and allocating a Vector of that capacity may overflow.
            #[allow(clippy::cast_possible_truncation)]
            let len = req.data.args[3] as usize;
            let len = len.min(nix::libc::PATH_MAX as usize); // Cap count at PATH_MAX
            let mut buf = if len == 0 { None } else { Some(vec![0u8; len]) };
            let ptr = match buf.as_mut() {
                Some(b) => b.as_mut_ptr(),
                None => std::ptr::null_mut(),
            };
            // SAFETY: In libc we trust.
            let n = unsafe {
                nix::libc::lgetxattr(
                    path.as_ptr(),
                    name,
                    ptr as *mut _ as *mut nix::libc::c_void,
                    len,
                )
            };
            let n = if n == -1 {
                return Err(Errno::last());
            } else {
                n as usize
            };
            if let Some(buf) = buf {
                #[allow(clippy::cast_possible_truncation)]
                proc.write_mem(&buf[..n], req.data.args[2] as usize, request)?;
            }
            #[allow(clippy::cast_possible_wrap)]
            Ok(request.return_syscall(n as i64))
        },
    )
}

fn sys_fgetxattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: None,
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "fgetxattr",
        argv,
        sandbox_path_1,
        |_fds, _paths, request, proc| {
            let req = request.get_request();
            let name = if req.data.args[1] != 0 {
                let mut buf = vec![0u8; nix::libc::PATH_MAX as usize];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[1] as usize, request)?;
                Some(buf)
            } else {
                None
            };
            let name = if let Some(ref name) = name {
                CStr::from_bytes_until_nul(name)
                    .map_err(|_| Errno::E2BIG)?
                    .as_ptr()
            } else {
                std::ptr::null()
            };
            // SAFETY: The size argument to the fgetxattr call
            // must not be fully trusted, it can be overly large,
            // and allocating a Vector of that capacity may overflow.
            #[allow(clippy::cast_possible_truncation)]
            let len = req.data.args[3] as usize;
            let len = len.min(nix::libc::PATH_MAX as usize); // Cap count at PATH_MAX
            let mut buf = if len == 0 { None } else { Some(vec![0u8; len]) };
            let ptr = match buf.as_mut() {
                Some(b) => b.as_mut_ptr(),
                None => std::ptr::null_mut(),
            };
            let fd = proc.get_fd(req.data.args[0] as RawFd)?;
            // SAFETY: In libc we trust.
            let n = unsafe {
                nix::libc::fgetxattr(
                    fd.as_raw_fd(),
                    name,
                    ptr as *mut _ as *mut nix::libc::c_void,
                    len,
                )
            };
            let e = Errno::last();
            let n = if n == -1 {
                return Err(e);
            } else {
                n as usize
            };
            if let Some(buf) = buf {
                #[allow(clippy::cast_possible_truncation)]
                proc.write_mem(&buf[..n], req.data.args[2] as usize, request)?;
            }
            #[allow(clippy::cast_possible_wrap)]
            Ok(request.return_syscall(n as i64))
        },
    )
}

fn sys_setxattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "setxattr",
        argv,
        sandbox_path_1,
        |_fds, paths, request, proc| {
            let req = request.get_request();
            let path = paths.first();
            let path = CString::new(path.as_os_str().as_bytes()).map_err(|_| Errno::EINVAL)?;
            let name = if req.data.args[1] != 0 {
                let mut buf = vec![0u8; nix::libc::PATH_MAX as usize];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[1] as usize, request)?;
                Some(buf)
            } else {
                None
            };
            let name = if let Some(ref name) = name {
                CStr::from_bytes_until_nul(name)
                    .map_err(|_| Errno::E2BIG)?
                    .as_ptr()
            } else {
                std::ptr::null()
            };
            // SAFETY: The size argument to the setxattr call
            // must not be fully trusted, it can be overly large,
            // and allocating a Vector of that capacity may overflow.
            let (buf, len) = if req.data.args[3] == 0 {
                (None, 0)
            } else {
                #[allow(clippy::cast_possible_truncation)]
                let len = req.data.args[3] as usize;
                let len = len.min(nix::libc::PATH_MAX as usize); // Cap count at PATH_MAX
                let mut buf = vec![0u8; len];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[2] as usize, request)?;
                (Some(buf), len)
            };
            let buf = if let Some(mut buf) = buf {
                buf.as_mut_ptr()
            } else {
                std::ptr::null_mut()
            };
            #[allow(clippy::cast_possible_truncation)]
            let flags = req.data.args[4] as nix::libc::c_int;

            // SAFETY: In libc we trust.
            if unsafe {
                nix::libc::setxattr(
                    path.as_ptr(),
                    name,
                    buf as *mut _ as *mut nix::libc::c_void,
                    len,
                    flags,
                )
            } == 0
            {
                Ok(request.return_syscall(0))
            } else {
                Err(Errno::last())
            }
        },
    )
}

fn sys_lsetxattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: false,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "lsetxattr",
        argv,
        sandbox_path_1,
        |_fds, paths, request, proc| {
            let req = request.get_request();
            let path = paths.first();
            let path = CString::new(path.as_os_str().as_bytes()).map_err(|_| Errno::EINVAL)?;
            let name = if req.data.args[1] != 0 {
                let mut buf = vec![0u8; nix::libc::PATH_MAX as usize];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[1] as usize, request)?;
                Some(buf)
            } else {
                None
            };
            let name = if let Some(ref name) = name {
                CStr::from_bytes_until_nul(name)
                    .map_err(|_| Errno::E2BIG)?
                    .as_ptr()
            } else {
                std::ptr::null()
            };
            // SAFETY: The size argument to the lsetxattr call
            // must not be fully trusted, it can be overly large,
            // and allocating a Vector of that capacity may overflow.
            let (buf, len) = if req.data.args[3] == 0 {
                (None, 0)
            } else {
                #[allow(clippy::cast_possible_truncation)]
                let len = req.data.args[3] as usize;
                let len = len.min(nix::libc::PATH_MAX as usize); // Cap count at PATH_MAX
                let mut buf = vec![0u8; len];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[2] as usize, request)?;
                (Some(buf), len)
            };
            let buf = if let Some(mut buf) = buf {
                buf.as_mut_ptr()
            } else {
                std::ptr::null_mut()
            };
            #[allow(clippy::cast_possible_truncation)]
            let flags = req.data.args[4] as nix::libc::c_int;
            // SAFETY: In libc we trust.
            if unsafe {
                nix::libc::lsetxattr(
                    path.as_ptr(),
                    name,
                    buf as *mut _ as *mut nix::libc::c_void,
                    len,
                    flags,
                )
            } == 0
            {
                Ok(request.return_syscall(0))
            } else {
                Err(Errno::last())
            }
        },
    )
}

fn sys_fsetxattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: None,
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "fsetxattr",
        argv,
        sandbox_path_1,
        |_fds, _paths, request, proc| {
            let req = request.get_request();
            let name = if req.data.args[1] != 0 {
                let mut buf = vec![0u8; nix::libc::PATH_MAX as usize];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[1] as usize, request)?;
                Some(buf)
            } else {
                None
            };
            let name = if let Some(ref name) = name {
                CStr::from_bytes_until_nul(name)
                    .map_err(|_| Errno::E2BIG)?
                    .as_ptr()
            } else {
                std::ptr::null()
            };
            // SAFETY: The size argument to the lsetxattr call
            // must not be fully trusted, it can be overly large,
            // and allocating a Vector of that capacity may overflow.
            let (buf, len) = if req.data.args[3] == 0 {
                (None, 0)
            } else {
                #[allow(clippy::cast_possible_truncation)]
                let len = req.data.args[3] as usize;
                let len = len.min(nix::libc::PATH_MAX as usize); // Cap count at PATH_MAX
                let mut buf = vec![0u8; len];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[2] as usize, request)?;
                (Some(buf), len)
            };
            let buf = if let Some(mut buf) = buf {
                buf.as_mut_ptr()
            } else {
                std::ptr::null_mut()
            };
            #[allow(clippy::cast_possible_truncation)]
            let flags = req.data.args[4] as nix::libc::c_int;
            let fd = proc.get_fd(req.data.args[0] as RawFd)?;
            // SAFETY: In libc we trust.
            if unsafe {
                nix::libc::fsetxattr(
                    fd.as_raw_fd(),
                    name,
                    buf as *mut _ as *mut nix::libc::c_void,
                    len,
                    flags,
                )
            } == 0
            {
                Ok(request.return_syscall(0))
            } else {
                Err(Errno::last())
            }
        },
    )
}

fn sys_listxattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "listxattr",
        argv,
        sandbox_path_1,
        |_fds, paths, request, proc| {
            let req = request.get_request();
            let path = paths.first();
            let path = CString::new(path.as_os_str().as_bytes()).map_err(|_| Errno::ENOENT)?;
            // SAFETY: The size argument to the listxattr call
            // must not be fully trusted, it can be overly large,
            // and allocating a Vector of that capacity may overflow.
            #[allow(clippy::cast_possible_truncation)]
            let len = req.data.args[2] as usize;
            let len = len.min(nix::libc::PATH_MAX as usize); // Cap count at PATH_MAX
            let mut buf = if len == 0 { None } else { Some(vec![0u8; len]) };
            let ptr = match buf.as_mut() {
                Some(b) => b.as_mut_ptr(),
                None => std::ptr::null_mut(),
            };
            // SAFETY: In libc we trust.
            let n = unsafe {
                nix::libc::listxattr(path.as_ptr(), ptr as *mut _ as *mut nix::libc::c_char, len)
            };
            let n = if n == -1 {
                return Err(Errno::last());
            } else {
                n as usize
            };
            if let Some(buf) = buf {
                #[allow(clippy::cast_possible_truncation)]
                proc.write_mem(&buf[..n], req.data.args[1] as usize, request)?;
            }
            #[allow(clippy::cast_possible_wrap)]
            Ok(request.return_syscall(n as i64))
        },
    )
}

fn sys_flistxattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: None,
        null: false,
        empty: true,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "flistxattr",
        argv,
        sandbox_path_1,
        |_fds, _paths, request, proc| {
            let req = request.get_request();
            // SAFETY: The size argument to the flistxattr call
            // must not be fully trusted, it can be overly large,
            // and allocating a Vector of that capacity may overflow.
            #[allow(clippy::cast_possible_truncation)]
            let len = req.data.args[2] as usize;
            let len = len.min(nix::libc::PATH_MAX as usize); // Cap count at PATH_MAX
            let mut buf = if len == 0 { None } else { Some(vec![0u8; len]) };
            let ptr = match buf.as_mut() {
                Some(b) => b.as_mut_ptr(),
                None => std::ptr::null_mut(),
            };
            let fd = proc.get_fd(req.data.args[0] as RawFd)?;
            // SAFETY: In libc we trust.
            let n = unsafe {
                nix::libc::flistxattr(fd.as_raw_fd(), ptr as *mut _ as *mut nix::libc::c_char, len)
            };
            let n = if n == -1 {
                return Err(Errno::last());
            } else {
                n as usize
            };
            if let Some(buf) = buf {
                #[allow(clippy::cast_possible_truncation)]
                proc.write_mem(&buf[..n], req.data.args[1] as usize, request)?;
            }
            #[allow(clippy::cast_possible_wrap)]
            Ok(request.return_syscall(n as i64))
        },
    )
}

fn sys_llistxattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: false,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "llistxattr",
        argv,
        sandbox_path_1,
        |_fds, paths, request, proc| {
            let req = request.get_request();
            let path = paths.first();
            let path = CString::new(path.as_os_str().as_bytes()).map_err(|_| Errno::ENOENT)?;
            // SAFETY: The size argument to the llistxattr call
            // must not be fully trusted, it can be overly large,
            // and allocating a Vector of that capacity may overflow.
            #[allow(clippy::cast_possible_truncation)]
            let len = req.data.args[2] as usize;
            let len = len.min(nix::libc::PATH_MAX as usize); // Cap count at PATH_MAX
            let mut buf = if len == 0 { None } else { Some(vec![0u8; len]) };
            let ptr = match buf.as_mut() {
                Some(b) => b.as_mut_ptr(),
                None => std::ptr::null_mut(),
            };
            // SAFETY: In libc we trust.
            let n = unsafe {
                nix::libc::llistxattr(path.as_ptr(), ptr as *mut _ as *mut nix::libc::c_char, len)
            };
            let n = if n == -1 {
                return Err(Errno::last());
            } else {
                n as usize
            };
            if let Some(buf) = buf {
                #[allow(clippy::cast_possible_truncation)]
                proc.write_mem(&buf[..n], req.data.args[1] as usize, request)?;
            }
            #[allow(clippy::cast_possible_wrap)]
            Ok(request.return_syscall(n as i64))
        },
    )
}

fn sys_removexattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "removexattr",
        argv,
        sandbox_path_1,
        |_fds, paths, request, proc| {
            let req = request.get_request();
            let path = paths.first();
            let path = CString::new(path.as_os_str().as_bytes()).map_err(|_| Errno::ENOENT)?;
            let name = if req.data.args[1] != 0 {
                let mut buf = vec![0u8; nix::libc::PATH_MAX as usize];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[1] as usize, request)?;
                Some(buf)
            } else {
                None
            };
            let name = if let Some(ref name) = name {
                CStr::from_bytes_until_nul(name)
                    .map_err(|_| Errno::E2BIG)?
                    .as_ptr()
            } else {
                std::ptr::null()
            };
            // SAFETY: In libc we trust.
            if unsafe { nix::libc::removexattr(path.as_ptr(), name) } == 0 {
                Ok(request.return_syscall(0))
            } else {
                Err(Errno::last())
            }
        },
    )
}

fn sys_fremovexattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: None,
        null: false,
        empty: true,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "fremovexattr",
        argv,
        sandbox_path_1,
        |_paths, _cwd, request, proc| {
            let req = request.get_request();
            let mut buf = vec![0u8; nix::libc::PATH_MAX as usize];
            let name = if req.data.args[1] != 0 {
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[1] as usize, request)?;
                Some(CStr::from_bytes_until_nul(&buf).map_err(|_| Errno::E2BIG)?)
            } else {
                None
            };
            let name = if let Some(name) = name {
                name.as_ptr()
            } else {
                std::ptr::null()
            };
            let fd = proc.get_fd(req.data.args[0] as RawFd)?;
            // SAFETY: In libc we trust.
            if unsafe { nix::libc::fremovexattr(fd.as_raw_fd(), name) } == 0 {
                Ok(request.return_syscall(0))
            } else {
                Err(Errno::last())
            }
        },
    )
}

fn sys_lremovexattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: false,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "lremovexattr",
        argv,
        sandbox_path_1,
        |_fds, paths, request, proc| {
            let req = request.get_request();
            let path = paths.first();
            let path = CString::new(path.as_os_str().as_bytes()).map_err(|_| Errno::EINVAL)?;
            let name = if req.data.args[1] != 0 {
                let mut buf = vec![0u8; nix::libc::PATH_MAX as usize];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[1] as usize, request)?;
                Some(buf)
            } else {
                None
            };
            let name = if let Some(ref name) = name {
                CStr::from_bytes_until_nul(name)
                    .map_err(|_| Errno::E2BIG)?
                    .as_ptr()
            } else {
                std::ptr::null()
            };
            // SAFETY: In libc we trust.
            if unsafe { nix::libc::lremovexattr(path.as_ptr(), name) } == 0 {
                Ok(request.return_syscall(0))
            } else {
                Err(Errno::last())
            }
        },
    )
}

/// Handles syscalls related to signal handling, protecting the SydB☮x
/// process and their threads from signals.
///
/// # Parameters
///
/// - `request`: User notification request from seccomp.
/// - `group`: true if the system call has both progress group id and process id (tgkill), false otherwise.
/// - `syscall_name`: The name of the syscall being handled, used for logging and error reporting.
/// # Returns
///
/// - `ScmpNotifResp`: Response indicating the result of the syscall handling.
fn syscall_signal_handler(request: &UNotifyEventRequest, group: bool) -> ScmpNotifResp {
    let req = request.get_request();

    #[allow(clippy::cast_possible_truncation)]
    let pid = [
        req.data.args[0] as nix::libc::pid_t,
        req.data.args[1] as nix::libc::pid_t,
    ];

    let mut bad = false;
    for (idx, tid) in pid.iter().enumerate() {
        if faccessat(
            None,
            format!("/proc/self/task/{tid}").as_str(),
            AccessFlags::F_OK,
            AtFlags::AT_SYMLINK_NOFOLLOW,
        )
        .is_ok()
        {
            bad = true;
            break;
        } else if idx >= 1 || !group {
            // all except tgkill get a single pid argument.
            break;
        }
    }

    if bad {
        return request.fail_syscall(Errno::EACCES as i32);
    }
    // SAFETY: This is safe because we haven't dereferenced
    // any pointers during access check.
    unsafe { request.continue_syscall() }
}

///
/// Handles syscalls related to paths, reducing code redundancy and ensuring a uniform way of dealing with paths.
///
/// # Parameters
///
/// - `request`: User notification request from seccomp.
/// - `syscall_name`: The name of the syscall being handled, used for logging and error reporting.
/// - `arg_mappings`: Non-empty list of argument mappings containing dirfd and path indexes, if applicable.
/// - `handler`: Closure that processes the constructed canonical paths and performs additional syscall-specific operations.
///
/// # Returns
///
/// - `ScmpNotifResp`: Response indicating the result of the syscall handling.
#[allow(clippy::cognitive_complexity)]
fn syscall_path_handler<C, H>(
    request: &UNotifyEventRequest,
    syscall_name: &str,
    path_argv: NonEmpty<SyscallPathArgument>,
    checker: C,
    handler: H,
) -> ScmpNotifResp
where
    C: Fn(
        &NonEmpty<PathBuf>,
        Capability,
        &str,
        &UNotifyEventRequest,
        &RemoteProcess,
    ) -> Result<(), Errno>,
    H: Fn(
        &NonEmpty<OwnedFd>,
        &NonEmpty<PathBuf>,
        &UNotifyEventRequest,
        &RemoteProcess,
    ) -> Result<ScmpNotifResp, Errno>,
{
    syscall_handler!(request, |req: &ScmpNotifReq, proc: &RemoteProcess| {
        // Determine system call capabilities.
        let caps = Capability::try_from((proc, request, req, syscall_name))?;

        // If sandboxing for all the selected capabilities is off, return immediately.
        // Exception: CAP_EXEC is available and we have patterns in exec/kill list.
        let mut check = false;
        let sandbox = request.get_sandbox(false);
        if caps.contains(Capability::CAP_EXEC) && sandbox.has_exec_kill() {
            check = true;
        } else {
            for cap in [
                Capability::CAP_READ,
                Capability::CAP_STAT,
                Capability::CAP_WRITE,
                Capability::CAP_EXEC,
                Capability::CAP_CONNECT,
                Capability::CAP_BIND,
            ] {
                if sandbox.enabled(cap) {
                    check = true;
                    break;
                }
            }
        }
        drop(sandbox); // release the read lock.
        if !check {
            // SAFETY: This is unsafe due to vulnerability to TOCTOU,
            // however since sandboxing for the respective Capability
            // is off, this is safe here.
            return unsafe { Ok(request.continue_syscall()) };
        }

        let mut paths: Vec<PathBuf> = Vec::new();
        for arg in path_argv.iter() {
            paths.push(proc.read_path(request, arg)?);
        }

        // At this point, we're absolutely sure the
        // path vectors have at least a single
        // element. Hence, using `unwrap` is fine
        // here.
        #[allow(clippy::disallowed_methods)]
        let mut non_empty_paths = NonEmpty::from_vec(paths).unwrap();
        checker(&non_empty_paths, caps, syscall_name, request, proc)?;

        // SAFETY: We have already resolved symbolic
        // links as necessary, from this point on we
        // do not resolve any longer for safety as
        // the filesystem may change which may
        // result in a TOCTOU.
        let mut how = openat2::OpenHow::new(nix::libc::O_PATH, 0);
        how.resolve = openat2::ResolveFlags::NO_SYMLINKS;

        let mut fds = Vec::new();
        for idx in 0..non_empty_paths.len() {
            let path = &non_empty_paths[idx];
            // System calls that have a pathless version.
            // fstat is already handled in the stat handler.
            if matches!(syscall_name, "fchmodat" | "fchownat") {
                // Resolve /proc file descriptor links.
                // Use get_fd if fd belongs to the current thread.
                let fd = if let Some(captures) = RE_PROC_FD.captures(path.as_os_str().as_bytes()) {
                    // SAFETY: Regex ensures the pid capture is a valid pid_t.
                    #[allow(clippy::arithmetic_side_effects)]
                    let proc_pid = captures["pid"]
                        .iter()
                        .fold(0 as nix::libc::pid_t, |acc, &b| {
                            acc * 10 + nix::libc::pid_t::from(b - b'0')
                        });
                    #[allow(clippy::cast_possible_wrap)]
                    if proc_pid == req.pid as nix::libc::pid_t {
                        let proc_tid = if let Some(proc_tid) = captures.name("tid") {
                            // SAFETY: Regex ensures the tid capture is a valid pid_t.
                            #[allow(clippy::arithmetic_side_effects)]
                            proc_tid
                                .as_bytes()
                                .iter()
                                .fold(0 as nix::libc::pid_t, |acc, &b| {
                                    acc * 10 + nix::libc::pid_t::from(b - b'0')
                                })
                        } else {
                            0
                        };

                        #[allow(clippy::cast_possible_wrap)]
                        if proc_tid == 0 || proc_tid == req.pid as nix::libc::pid_t {
                            // SAFETY: Regex ensures the fd capture is a valid RawFd.
                            #[allow(clippy::arithmetic_side_effects)]
                            Some(
                                captures["fd"]
                                    .iter()
                                    .fold(0 as RawFd, |acc, &b| acc * 10 + RawFd::from(b - b'0')),
                            )
                        } else {
                            None
                        }
                    } else {
                        None
                    }
                } else {
                    None
                };

                if let Some(fd) = fd {
                    non_empty_paths[idx] = PathBuf::from("");
                    fds.push(proc.get_fd(fd)?);
                    continue;
                }
            }
            let path = match path.parent() {
                Some(path) => {
                    how.flags |= nix::libc::O_DIRECTORY as u64;
                    path
                }
                None => path,
            };
            let fd = openat2::openat2(None, path, &how)
                .map_err(|e| Errno::from_i32(e.raw_os_error().unwrap_or(nix::libc::EINVAL)))?;
            fds.push(unsafe { OwnedFd::from_raw_fd(fd) });
        }
        #[allow(clippy::disallowed_methods)]
        let non_empty_fds = NonEmpty::from_vec(fds).unwrap();

        handler(&non_empty_fds, &non_empty_paths, request, proc)
    })
}

#[allow(clippy::cognitive_complexity)]
fn syscall_open_handler(request: &UNotifyEventRequest, syscall_name: &str) -> ScmpNotifResp {
    syscall_handler!(request, |req: &ScmpNotifReq, proc: &RemoteProcess| {
        #[allow(clippy::cast_possible_truncation)]
        let (mut flags, mode) = match syscall_name {
            "creat" => {
                let flags = OFlag::O_CREAT | OFlag::O_WRONLY | OFlag::O_TRUNC;
                let mode = Mode::from_bits_truncate(req.data.args[1] as nix::libc::mode_t);
                (flags, mode)
            }
            "open" => {
                let flags = OFlag::from_bits_truncate(req.data.args[1] as nix::libc::c_int);
                let mode = Mode::from_bits_truncate(req.data.args[2] as nix::libc::mode_t);
                (flags, mode)
            }
            "openat" => {
                let flags = OFlag::from_bits_truncate(req.data.args[2] as nix::libc::c_int);
                let mode = Mode::from_bits_truncate(req.data.args[3] as nix::libc::mode_t);
                (flags, mode)
            }
            "openat2" => {
                let open_how = proc.remote_ohow(
                    req.data.args[2] as usize,
                    req.data.args[3] as usize,
                    request,
                )?;
                let flags = OFlag::from_bits_truncate(open_how.flags as nix::libc::c_int);
                let mode = Mode::from_bits_truncate(open_how.mode as nix::libc::mode_t);
                (flags, mode)
            }
            _ => unreachable!(),
        };

        if flags.contains(OFlag::O_PATH) {
            // SAFETY: O_PATH obtains a file descriptor for
            // low-level access. The file is not opened and
            // other operations will fail.  Hence we consider
            // this secure, and continue the system call here.
            return unsafe { Ok(request.continue_syscall()) };
        }

        let cap = if flags.contains(OFlag::O_CREAT) {
            Capability::CAP_WRITE
        } else {
            match flags & OFlag::O_ACCMODE {
                OFlag::O_RDONLY => Capability::CAP_READ,
                _ => Capability::CAP_WRITE,
            }
        };

        let sandbox = request.get_sandbox(false);
        #[allow(clippy::cast_possible_wrap)]
        let is_lock = sandbox.locked_for_pid(req.pid as nix::libc::pid_t);
        let enabled = sandbox.enabled(cap);
        if is_lock && !enabled {
            // SAFETY: This is unsafe due to vulnerability to TOCTOU.
            // Since the sandbox is off, this is fine here.
            return unsafe { Ok(request.continue_syscall()) };
        }
        drop(sandbox);

        #[allow(clippy::cast_possible_wrap)]
        let pid = Pid::from_raw(req.pid as i32);
        let resolve =
            !(flags.contains(OFlag::O_NOFOLLOW) || flags.contains(OFlag::O_CREAT | OFlag::O_EXCL));

        // Step 1: Read the remote path.
        let (path, arg) = match syscall_name {
            "open" | "creat" => {
                let arg = SyscallPathArgument {
                    dirfd: None,
                    path: Some(0),
                    null: false,
                    empty: false,
                    dotlast: None,
                    resolve,
                };
                (remote_path_n!(proc, req, 0, &request)?, arg)
            }
            _ => {
                // "openat" | "openat2"
                let arg = SyscallPathArgument {
                    dirfd: Some(0),
                    path: Some(1),
                    null: false,
                    empty: false,
                    dotlast: None,
                    resolve,
                };
                (remote_path_n!(proc, req, 1, &request)?, arg)
            }
        };
        let path = PathBuf::from(OsStr::from_bytes(path.to_bytes()));

        // Step 2: Handle the special /dev/syd paths.
        if path.as_os_str().as_bytes() == MAGIC_PREFIX.as_bytes() {
            if is_lock {
                return Err(Errno::ENOENT);
            }
            let sandbox = request.get_sandbox(false);
            let content = if cap.contains(Capability::CAP_READ) {
                Some(serde_json::to_string(&*sandbox).map_err(|_| Errno::EINVAL)?)
            } else {
                None
            };
            drop(sandbox);

            let fd = if let Some(content) = content {
                let fd = open(
                    ".",
                    OFlag::O_RDWR | OFlag::O_TMPFILE,
                    Mode::S_IRUSR | Mode::S_IWUSR,
                )?;
                write(fd, content.as_bytes())?;
                lseek(fd, 0, Whence::SeekSet)?;
                fd
            } else {
                open("/dev/null", flags, mode)?
            };

            // Send the file descriptor to the process and return the fd no.
            let result = request.add_fd(fd, true);
            let _ = close(fd);
            return result.map(|fd| request.return_syscall(i64::from(fd)));
        } else if !enabled {
            // SAFETY: This is unsafe due to vulnerability to TOCTOU.
            // Since the sandbox is off, this is fine here.
            return unsafe { Ok(request.continue_syscall()) };
        }

        // Step 3: Canonicalize the path
        //
        // Step 3.1: Ensure path is not empty, normalize.
        if path.is_empty() && !flags.contains(OFlag::O_TMPFILE) {
            return Err(Errno::ENOENT);
        }
        let path = normalize_path(path);
        // Step 3.2: Ensure the path is absolute.
        let abspath = if path.is_absolute() {
            path.clone()
        } else if path == PathBuf::from(".") {
            // SAFETY: Add a trailing slash to the directory
            // to assert it must be a directory.
            proc.read_directory(request, &arg)?.join("")
        } else {
            // SAFETY: We join path even if it's empty.
            // This way e.g. a final dot adds a trailing
            // slash to the path, ensuring the system call
            // to fail with ENOTDIR as necessary.
            proc.read_directory(request, &arg)?.join(&path)
        };

        // Step 3.3: Check for magic paths and replace.
        let abspath = replace_self_with_pid(abspath, pid);

        // Step 3.4: Canonicalize the resulting path.
        // FIXME: if we remove the &abspath borrow as clippy suggests
        // we get a use after move error during compilation.
        #[allow(clippy::needless_borrows_for_generic_args)]
        let abspath =
            match crate::fs::canonicalize(&abspath, resolve, crate::fs::MissingHandling::Missing) {
                Ok(path) => {
                    trace!("ctx": "path",
                        "act": "resolve",
                        "arg": arg,
                        "path_fini": format!("{}", abspath.display()),
                        "path_orig": format!("{}", path.display()),
                        "pid": pid.as_raw(),
                        "sys": syscall_name);
                    path
                }
                Err(error) => {
                    debug!("ctx": "path",
                        "act": "resolve",
                        "arg": arg,
                        "path_fini": format!("{}", abspath.display()),
                        "path_orig": format!("{}", path.display()),
                        "err": error.to_string(),
                        "pid": pid.as_raw(),
                        "sys": syscall_name);
                    return Err(error);
                }
            };

        // Step 3.5: Check for magic paths, and replace.
        // SAFETY: We have to do this one more time to ensure
        // relative paths are properly handled.
        let mut abspath = replace_self_with_pid(abspath, pid);

        // SAFETY: We must provide safe access to
        // sandbox process' controlling terminal.
        if GLOB_TTY.is_match(&abspath) {
            let dev_tty = proc_tty(pid)?;
            if !request.is_valid() {
                return Err(Errno::EACCES);
            }
            abspath = dev_tty;
        }

        // Step 3.6: Add O_NOFOLLOW to flags.
        // SAFETY: We have already resolved the symbolic
        // links in the path as necessary, to prevent a
        // time-of-check to time-of-use vector we add
        // O_NOFOLLOW to flags here.
        flags.insert(OFlag::O_NOFOLLOW);

        // Step 4: Check for access.
        let sandbox = request.get_sandbox(false);
        let action = sandbox.check_path(cap, &abspath);
        let hidden = if action != Action::Allow {
            sandbox.is_hidden(&abspath)
        } else {
            // No need for hidden check if we're allowing.
            false
        };
        drop(sandbox);

        // Step 5: Perform action: allow->emulate, deny->log.
        match action {
            Action::Allow => {
                // The system call is allowlisted.
                // To prevent TOCTOU, we open the file ourselves,
                // and put the file descriptor to the process'
                // address space with SECCOMP_IOCTL_NOTIF_ADDFD.

                // Resolve /proc file descriptor links.
                // Use get_fd if fd belongs to the current thread.
                let fd = if let Some(captures) = RE_PROC_FD.captures(abspath.as_os_str().as_bytes())
                {
                    // SAFETY: Regex ensures the pid capture is a valid pid_t.
                    #[allow(clippy::arithmetic_side_effects)]
                    let proc_pid = captures["pid"]
                        .iter()
                        .fold(0 as nix::libc::pid_t, |acc, &b| {
                            acc * 10 + nix::libc::pid_t::from(b - b'0')
                        });
                    if proc_pid == pid.as_raw() {
                        let proc_tid = if let Some(proc_tid) = captures.name("tid") {
                            // SAFETY: Regex ensures the tid capture is a valid pid_t.
                            #[allow(clippy::arithmetic_side_effects)]
                            proc_tid
                                .as_bytes()
                                .iter()
                                .fold(0 as nix::libc::pid_t, |acc, &b| {
                                    acc * 10 + nix::libc::pid_t::from(b - b'0')
                                })
                        } else {
                            0
                        };

                        if proc_tid == 0 || proc_tid == pid.as_raw() {
                            // SAFETY: Regex ensures the fd capture is a valid RawFd.
                            #[allow(clippy::arithmetic_side_effects)]
                            Some(
                                captures["fd"]
                                    .iter()
                                    .fold(0 as RawFd, |acc, &b| acc * 10 + RawFd::from(b - b'0')),
                            )
                        } else {
                            None
                        }
                    } else {
                        None
                    }
                } else {
                    None
                };

                let (fd, name) = if let Some(fd) = fd {
                    (
                        proc.get_fd(fd).map_err(|e| {
                            if e == Errno::EBADF {
                                Errno::ENOENT
                            } else {
                                e
                            }
                        })?,
                        None,
                    )
                } else {
                    // SAFETY: We have already resolved symbolic
                    // links as necessary, from this point on we
                    // do not resolve any longer for safety as
                    // the filesystem may change which may
                    // result in a TOCTOU.
                    let mut how = openat2::OpenHow::new(nix::libc::O_PATH, 0);
                    how.resolve = openat2::ResolveFlags::NO_SYMLINKS;
                    let parent = abspath.parent().unwrap_or(&abspath);
                    let name = crate::path_filename(&abspath);
                    let fd = openat2::openat2(None, parent, &how).map_err(|e| {
                        Errno::from_i32(e.raw_os_error().unwrap_or(nix::libc::EINVAL))
                    })?;

                    // SAFETY: Refuse to open the file is not one of:
                    // 1. Regular file
                    // 2. Directory
                    // 3. Socket
                    // 4. Character device
                    // 5. Named pipe
                    // This leaves out block devices above all.
                    if let Ok(stat) = fstatat(fd, &name, AtFlags::empty()) {
                        match SFlag::from_bits_truncate(stat.st_mode & SFlag::S_IFMT.bits()) {
                            SFlag::S_IFIFO | SFlag::S_IFREG | SFlag::S_IFSOCK | SFlag::S_IFCHR => {}
                            SFlag::S_IFDIR => {
                                // SAFETY: Opening a directory for write must open with EISDIR.
                                if flags.contains(
                                    OFlag::O_WRONLY
                                        | OFlag::O_RDWR
                                        | OFlag::O_CREAT
                                        | OFlag::O_TRUNC,
                                ) {
                                    let _ = close(fd);
                                    return Err(Errno::EISDIR);
                                }
                            }
                            _ => {
                                let _ = close(fd);
                                return Err(Errno::EACCES);
                            }
                        }
                    }
                    // SAFETY: Valid FD.
                    (unsafe { OwnedFd::from_raw_fd(fd) }, Some(name))
                };

                let result = if let Some(name) = name {
                    // SAFETY: SydB☮x' umask is 0 here so we respect process' umask.
                    let mode = mode & !proc_umask(Some(pid))?;
                    openat(fd.as_raw_fd(), &name, flags, mode)
                } else {
                    Ok(fd.as_raw_fd())
                };

                let fd_this = match result {
                    Ok(fd) => {
                        trace!("ctx": "open",
                            "pid": pid.as_raw(),
                            "path_orig": format!("{}", path.display()),
                            "path_fini": format!("{}", abspath.display()),
                            "sys": syscall_name,
                            "flags": crate::oflag_fmt(flags),
                            "mode": mode.bits());
                        fd
                    }
                    Err(error)
                        if error == Errno::EEXIST
                            || error == Errno::ENOENT
                            || error == Errno::ENOTDIR =>
                    {
                        debug!("ctx": "open",
                            "pid": pid.as_raw(),
                            "err": error.to_string(),
                            "path_orig": format!("{}", path.display()),
                            "path_fini": format!("{}", abspath.display()),
                            "sys": syscall_name,
                            "flags": crate::oflag_fmt(flags),
                            "mode": mode.bits());
                        return Err(error);
                    }
                    Err(error) => {
                        info!("ctx": "open",
                            "pid": pid.as_raw(),
                            "err": error.to_string(),
                            "path_orig": path.to_string_lossy(),
                            "path_fini": abspath,
                            "sys": syscall_name,
                            "flags": crate::oflag_fmt(flags),
                            "mode": mode.bits());
                        return Err(error);
                    }
                };
                let result = request.add_fd(fd_this, flags.contains(OFlag::O_CLOEXEC));
                let _ = close(fd_this);
                result.map(|fd| request.return_syscall(i64::from(fd)))
            }
            Action::Deny | Action::Filter if hidden => {
                // SAFETY: We do not report violations when the path is
                // hidden and return ENOENT, so as to make read/write/exec
                // sandboxing consistent with stat sandboxing. With this
                // restriction in place, it is not possible to enumerate
                // existing hidden paths by attempting to read, write or
                // execute them.

                // Debug log access violation to ease troubleshooting
                // using the log feature (compiled out by default).
                debug!("ctx": "access",
                    "act": action,
                    "cap": cap,
                    "orig": format!("{}", path.display()),
                    "path": format!("{}", abspath.display()),
                    "flags": oflag_fmt(flags),
                    "mode": mode.bits(),
                    "pid": pid.as_raw(),
                    "sys": syscall_name);

                Err(Errno::ENOENT)
            }
            Action::Filter | Action::Kill => Err(Errno::EACCES),
            Action::Deny => {
                // Report access violation.
                warn!("ctx": "access",
                    "cap": cap,
                    "orig": format!("{}", path.display()),
                    "path": format!("{}", abspath.display()),
                    "flags": oflag_fmt(flags),
                    "mode": mode.bits(),
                    "pid": pid.as_raw(),
                    "sys": syscall_name);
                Err(Errno::EACCES)
            }
        }
    })
}

#[allow(clippy::cognitive_complexity)]
fn syscall_stat_handler(
    request: &UNotifyEventRequest,
    syscall_name: &str,
    path_argv: NonEmpty<SyscallPathArgument>,
) -> ScmpNotifResp {
    syscall_handler!(request, |req: &ScmpNotifReq, proc: &RemoteProcess| {
        // If sandboxing for CAP_STAT is off, and magic lock is set, return immediately.
        let sandbox = request.get_sandbox(false);
        #[allow(clippy::cast_possible_wrap)]
        let is_lock = sandbox.locked_for_pid(req.pid as nix::libc::pid_t);
        let is_stat = sandbox.enabled(Capability::CAP_STAT);
        if is_lock && !is_stat {
            // SAFETY: This is unsafe due to vulnerability to TOCTOU.
            // However, since stat sandboxing is disabled here this is
            // safe to call.
            return unsafe { Ok(request.continue_syscall()) };
        }
        drop(sandbox); // release the read lock.

        let arg = path_argv.first();
        let (mut path, path_raw) = proc.read_path_opt(request, arg, false)?;
        trace!("ctx": "syscall",
            "sys": syscall_name,
            "cap": Capability::CAP_STAT,
            "arg": arg,
            "path": format!("{}", path.display()));

        // SAFETY: Carefully handle magic symlinks.
        if !arg.resolve {
            if let Some(path_raw) = path_raw {
                let bytes = path_raw.as_os_str().as_bytes();
                if matches!(
                    bytes,
                    b"/proc/self"
                        | b"/proc/thread-self"
                        | b"/dev/fd"
                        | b"/dev/stdin"
                        | b"/dev/stdout"
                        | b"/dev/stderr"
                ) {
                    // We're not resolving symbolic links,
                    // and the path is an exact match.
                    // Use the raw path as otherwise the
                    // sandbox process will see e.g. /proc/self
                    // as a directory rather than a symlink which
                    // can be most confusing.
                    path = path_raw;
                }
            }
        }

        let (path, fd, flags) = if let Ok(path) = path.strip_prefix(MAGIC_PREFIX) {
            // Handle magic prefix (ie /dev/syd)
            let sandbox = request.get_sandbox(false);
            #[allow(clippy::cast_possible_wrap)]
            if sandbox.locked_for_pid(req.pid as nix::libc::pid_t) {
                // Magic commands locked, return ENOENT.
                return Ok(request.fail_syscall(nix::libc::ENOENT));
            }
            drop(sandbox);

            // Execute magic command.
            let mut sandbox = request.get_sandbox(true);
            if path.is_empty() {
                sandbox.config("")?;
            } else if path.as_os_str().as_bytes() == b"panic" {
                #[allow(clippy::disallowed_methods)]
                exit(127);
            } else if let Ok(path) = path.strip_prefix("load") {
                // We handle load specially here as it involves process access.
                match parse_fd(path) {
                    Ok(remote_fd) => {
                        let fd = proc.get_fd(remote_fd)?;
                        let file = BufReader::new(File::from(fd));
                        if sandbox.parse_config(file).is_err() {
                            return Ok(request.fail_syscall(nix::libc::EINVAL));
                        }
                        if let Err(error) = sandbox.build_globsets() {
                            error!("ctx": "config",
                                "pid": req.pid,
                                "sys": syscall_name,
                                "cfg": "load",
                                "fd": remote_fd,
                                "error": error.to_string());
                        } else {
                            info!("ctx": "config",
                                "pid": req.pid,
                                "sys": syscall_name,
                                "cfg": "load",
                                "fd": remote_fd);
                        }
                        // Fall through to emulate as /dev/null.
                    }
                    Err(errno) => {
                        return Ok(request.fail_syscall(errno as i32));
                    }
                }
            } else {
                // SAFETY: Conversion from PathBuf to String is OK here,
                // since sandbox config function does not work on the
                // filesystem, rather treats the Strings as sandbox
                // commands.
                match sandbox.config(path.to_string_lossy().as_ref()) {
                    Ok(_) => {
                        if let Err(error) = sandbox.build_globsets() {
                            error!("ctx": "config",
                                "pid": req.pid,
                                "sys": syscall_name,
                                "cfg": format!("{}", path.display()),
                                "err": error.to_string());
                        } else {
                            info!("ctx": "config",
                                "pid": req.pid,
                                "sys": syscall_name,
                                "cfg": format!("{}", path.display()));
                        }
                    }
                    Err(Errno::ENOENT) => {
                        info!("ctx": "config",
                            "pid": req.pid,
                            "sys": syscall_name,
                            "cfg": format!("{}", path.display()),
                            "err": Errno::ENOENT.to_string());
                        return Err(Errno::ENOENT);
                    }
                    Err(error) => {
                        error!("ctx": "config",
                            "pid": req.pid,
                            "sys": syscall_name,
                            "cfg": format!("{}", path.display()),
                            "err": error.to_string());
                        return Err(error);
                    }
                };
            }
            drop(sandbox);

            // If the stat buffer is NULL, return immediately.
            let idx = match syscall_name {
                "stat" | "lstat" => 1,
                "newfstatat" => 2,
                "statx" => 4,
                _ => 0,
            };
            if idx > 0 && req.data.args[idx] == 0 {
                return Ok(request.return_syscall(0));
            }

            // SAFETY: We trust the contents of the MAGIC_FILE.
            let path = unsafe { CString::from_vec_unchecked(MAGIC_FILE.into()) };

            (path, None, 0)
        } else if is_stat {
            let non_empty_paths = nonempty![path.clone()];
            sandbox_path_1(
                &non_empty_paths,
                Capability::CAP_STAT,
                syscall_name,
                request,
                proc,
            )?;

            // Resolve /proc file descriptor links.
            // Use get_fd if fd belongs to the current thread.
            #[allow(clippy::arithmetic_side_effects)]
            #[allow(clippy::cast_possible_truncation)]
            let fd = if syscall_name == "fstat"
                || arg.empty
                || (req.data.args[0] as nix::libc::c_int != nix::libc::AT_FDCWD
                    && GLOB_PROC.is_match(&path))
            {
                Some(req.data.args[0] as RawFd)
            } else if let Some(captures) = RE_PROC_FD.captures(path.as_os_str().as_bytes()) {
                #[allow(clippy::cast_possible_truncation)]
                #[allow(clippy::cast_possible_wrap)]
                let pid = Pid::from_raw(req.pid as i32);

                // SAFETY: Regex ensures the pid capture is a valid pid_t.
                #[allow(clippy::arithmetic_side_effects)]
                let proc_pid = captures["pid"]
                    .iter()
                    .fold(0 as nix::libc::pid_t, |acc, &b| {
                        acc * 10 + nix::libc::pid_t::from(b - b'0')
                    });
                if proc_pid == pid.as_raw() {
                    let proc_tid = if let Some(proc_tid) = captures.name("tid") {
                        // SAFETY: Regex ensures the tid capture is a valid pid_t.
                        #[allow(clippy::arithmetic_side_effects)]
                        proc_tid
                            .as_bytes()
                            .iter()
                            .fold(0 as nix::libc::pid_t, |acc, &b| {
                                acc * 10 + nix::libc::pid_t::from(b - b'0')
                            })
                    } else {
                        0
                    };

                    if proc_tid == 0 || proc_tid == pid.as_raw() {
                        // SAFETY: Regex ensures the fd capture is a valid RawFd.
                        #[allow(clippy::arithmetic_side_effects)]
                        Some(
                            captures["fd"]
                                .iter()
                                .fold(0 as RawFd, |acc, &b| acc * 10 + RawFd::from(b - b'0')),
                        )
                    } else {
                        None
                    }
                } else {
                    None
                }
            } else {
                None
            };

            let fd = if let Some(fd) = fd {
                proc.get_fd(fd)
                    .map_err(|e| if e == Errno::EBADF { Errno::ENOENT } else { e })?
            } else {
                // SAFETY: We have already resolved symbolic
                // links as necessary, from this point on we
                // do not resolve any longer for safety as
                // the filesystem may change which may
                // result in a TOCTOU.
                let mut how = openat2::OpenHow::new(nix::libc::O_PATH, 0);
                how.resolve = openat2::ResolveFlags::NO_SYMLINKS;
                if !arg.resolve {
                    how.flags |= nix::libc::O_NOFOLLOW as u64;
                    // SAFETY: Remove trailing slash if it exists,
                    // or else we may get ELOOP on symbolic links.
                    let path_bytes = path.as_os_str().as_bytes();
                    #[allow(clippy::arithmetic_side_effects)]
                    if path_bytes.ends_with(&[b'/']) && path_bytes.len() > 1 {
                        // SAFETY: Since we're operating on valid path bytes, getting a slice is safe.
                        // This excludes the root path "/" to avoid turning it into an empty path.
                        path = PathBuf::from(OsStr::from_bytes(&path_bytes[..path_bytes.len() - 1]))
                    }
                }
                let fd = openat2::openat2(None, path, &how)
                    .map_err(|e| Errno::from_i32(e.raw_os_error().unwrap_or(nix::libc::EINVAL)))?;
                // SAFETY: Valid FD.
                unsafe { OwnedFd::from_raw_fd(fd) }
            };

            match syscall_name {
                "stat" => (CString::default(), Some(fd), nix::libc::AT_EMPTY_PATH),
                "lstat" => (
                    CString::default(),
                    Some(fd),
                    nix::libc::AT_EMPTY_PATH | nix::libc::AT_SYMLINK_NOFOLLOW,
                ),
                "fstat" => (CString::default(), Some(fd), nix::libc::AT_EMPTY_PATH),
                "statx" | "newfstatat" => {
                    let flags = if arg.resolve {
                        nix::libc::AT_EMPTY_PATH
                    } else {
                        nix::libc::AT_EMPTY_PATH | nix::libc::AT_SYMLINK_NOFOLLOW
                    };
                    (CString::default(), Some(fd), flags)
                }
                _ => unreachable!(),
            }
        } else {
            // Continue system call normally.
            // SAFETY: This is unsafe due to vulnerability to TOCTOU,
            // however if we're here stat sandboxing is disabled,
            // so this is safe to call.
            return unsafe { Ok(request.continue_syscall()) };
        };

        match syscall_name {
            "statx" => {
                let mut statx = MaybeUninit::<crate::compat::statx>::uninit();
                #[allow(clippy::cast_possible_truncation)]
                let flags = flags
                    | (req.data.args[2] as nix::libc::c_int
                        & !(nix::libc::AT_EMPTY_PATH | nix::libc::AT_SYMLINK_NOFOLLOW));
                #[allow(clippy::cast_possible_truncation)]
                let mask = req.data.args[3] as nix::libc::c_int;
                // SAFETY: In libc we trust.
                let ret = if let Some(fd) = fd {
                    unsafe {
                        nix::libc::syscall(
                            nix::libc::SYS_statx,
                            fd.as_raw_fd(),
                            path.as_ptr(),
                            flags,
                            mask,
                            statx.as_mut_ptr(),
                        )
                    }
                } else {
                    unsafe {
                        nix::libc::syscall(
                            nix::libc::SYS_statx,
                            nix::libc::AT_FDCWD,
                            path.as_ptr(),
                            flags,
                            mask,
                            statx.as_mut_ptr(),
                        )
                    }
                };
                let errno = Errno::last();
                if ret == 0 {
                    trace!("ctx": "stat",
                        "path": &path,
                        "pipe": flags & nix::libc::AT_EMPTY_PATH != 0,
                        "sys": syscall_name);
                } else if matches!(errno, Errno::EPERM | Errno::ENOENT | Errno::ENOTDIR) {
                    debug!("ctx": "stat",
                        "path": format!("{}", path.to_string_lossy()),
                        "pipe": flags & nix::libc::AT_EMPTY_PATH != 0,
                        "sys": syscall_name,
                        "err": errno.to_string());
                    return Err(errno);
                } else {
                    info!("ctx": "stat",
                        "path": format!("{}", path.to_string_lossy()),
                        "pipe": flags & nix::libc::AT_EMPTY_PATH != 0,
                        "sys": syscall_name,
                        "err": errno.to_string());
                    return Err(errno);
                }

                // SAFETY: The following block creates an immutable byte slice representing the memory of `statx`.
                // We ensure that the slice covers the entire memory of `statx` using `std::mem::size_of_val`.
                // Since `statx` is a stack variable and we're only borrowing its memory for the duration of the slice,
                // there's no risk of `statx` being deallocated while the slice exists.
                // Additionally, we ensure that the slice is not used outside of its valid lifetime.
                let statx = unsafe {
                    std::slice::from_raw_parts(
                        statx.as_ptr() as *const u8,
                        std::mem::size_of_val(&statx),
                    )
                };
                #[allow(clippy::cast_possible_truncation)]
                proc.write_mem(statx, req.data.args[4] as usize, request)?;
            }
            _ => {
                // "stat" | "fstat" | "lstat" | "newfstatat"
                let mut stat = MaybeUninit::<nix::libc::stat>::uninit();
                Errno::clear();
                // SAFETY: In libc we trust.
                let ret = if let Some(fd) = fd {
                    unsafe {
                        nix::libc::fstatat(fd.as_raw_fd(), path.as_ptr(), stat.as_mut_ptr(), flags)
                    }
                } else {
                    unsafe {
                        nix::libc::fstatat(
                            nix::libc::AT_FDCWD,
                            path.as_ptr(),
                            stat.as_mut_ptr(),
                            flags,
                        )
                    }
                };
                let errno = Errno::last();
                if ret == 0 {
                    trace!("ctx": "stat",
                        "path": format!("{}", path.to_string_lossy()),
                        "pipe": flags & nix::libc::AT_EMPTY_PATH != 0,
                        "sys": syscall_name);
                } else if matches!(errno, Errno::EPERM | Errno::ENOENT | Errno::ENOTDIR) {
                    debug!("ctx": "stat",
                        "path": format!("{}", path.to_string_lossy()),
                        "pipe": flags & nix::libc::AT_EMPTY_PATH != 0,
                        "sys": syscall_name,
                        "err": errno.to_string());
                    return Err(errno);
                } else {
                    info!("ctx": "stat",
                        "path": format!("{}", path.to_string_lossy()),
                        "pipe": flags & nix::libc::AT_EMPTY_PATH != 0,
                        "sys": syscall_name,
                        "err": errno.to_string());
                    return Err(errno);
                }

                // SAFETY: stat returned success, stat struct is properly populated.
                unsafe { stat.assume_init() };
                // SAFETY: The following block creates an immutable byte slice representing the memory of `stat`.
                // We ensure that the slice covers the entire memory of `stat` using `std::mem::size_of_val`.
                // Since `stat` is a stack variable and we're only borrowing its memory for the duration of the slice,
                // there's no risk of `stat` being deallocated while the slice exists.
                // Additionally, we ensure that the slice is not used outside of its valid lifetime.
                let stat = unsafe {
                    std::slice::from_raw_parts(
                        std::ptr::addr_of!(stat) as *const u8,
                        std::mem::size_of_val(&stat),
                    )
                };
                #[allow(clippy::cast_possible_truncation)]
                let addr = req.data.args[if syscall_name == "newfstatat" { 2 } else { 1 }] as usize;
                if addr != 0 {
                    proc.write_mem(stat, addr, request)?;
                }
            }
        }

        // stat system call successfully emulated.
        Ok(request.return_syscall(0))
    })
}

/// A helper function to handle network-related syscalls.
///
/// This function abstracts the common logic involved in handling network syscalls such as `bind`,
/// `connect`, `recvfrom`, and `sendto` in a seccomp-based sandboxing environment. It reduces code
/// duplication across different syscall handler functions.
///
/// # Parameters
///
/// - `request`: Reference to `UNotifyEventRequest`, the user notification request from seccomp.
/// - `arg_addr`: Index of the argument representing the address in the syscall arguments.
/// - `arg_len`: Index of the argument representing the length of the address in the syscall arguments.
/// - `syscall_name`: A string slice holding the name of the syscall being handled, used for error reporting.
/// - `handler`: A closure that handles the processing of the `SockaddrStorage` object and is responsible
///   for logging or any other required operations, specific to the syscall being handled.
///
/// # Safety
///
/// This function contains unsafe blocks due to direct memory access and potential TOCTOU (Time-of-Check
/// Time-of-Use) vulnerabilities, which are acceptable in this context as per requirements. The invoked
/// unsafe blocks are commented with detailed explanations of safety considerations.
///
/// # Errors
///
/// This function returns an error if there is an issue reading memory, converting the address types,
/// or any error returned by the passed `handler` closure. All errors are wrapped in `ScmpNotifResp`
/// and are returned as syscall failures.
///
/// # Returns
///
/// Returns `ScmpNotifResp` indicating the result of the syscall handling:
/// - If successful, it contains a continued syscall.
/// - If an error occurs, it contains a failed syscall with an `EACCES` error code.
#[allow(clippy::cognitive_complexity)]
fn syscall_network_handler<F>(
    request: &UNotifyEventRequest,
    args: &[u64; 6],
    arg_addr: usize,
    arg_len: usize,
    len_ptr: bool,
    syscall_name: &str,
    handler: F,
) -> ScmpNotifResp
where
    F: Fn(
        &RemoteProcess,
        &UNotifyEventRequest,
        &ScmpNotifReq,
        &SockaddrStorage,
        &str,
        Capability,
    ) -> Result<(), Errno>,
{
    syscall_handler!(request, |req: &ScmpNotifReq, proc: &RemoteProcess| {
        let cap = match syscall_name {
            "bind" => Capability::CAP_BIND,
            _ => Capability::CAP_CONNECT,
        };

        let sandbox = request.get_sandbox(false);
        let (check, allow_successful_bind, allow_unsupported_socket_families) = (
            sandbox.enabled(cap),
            sandbox.allow_successful_bind(),
            sandbox.allow_unsupported_socket_families(),
        );
        drop(sandbox);

        // Return immediately if sandboxing is not enabled for current capability,
        if !check {
            // SAFETY: This is unsafe due to vulnerability to TOCTOU,
            // however since the sandboxing for the requested capability
            // is disabled this is safe here.
            return unsafe { Ok(request.continue_syscall()) };
        }

        #[allow(clippy::cast_possible_truncation)]
        let (addr_remote, addr_len) = (
            args[arg_addr] as usize,
            get_addr_len(args, arg_len, len_ptr, proc, request)?,
        );

        if addr_remote == 0 || addr_len == 0 {
            if matches!(syscall_name, "recvfrom" | "sendto") {
                // SAFETY: Connection mode socket, safe to continue.
                return Ok(unsafe { request.continue_syscall() });
            } else {
                return Err(Errno::EFAULT);
            }
        }

        let mut addr = get_addr(proc, request, addr_remote, addr_len)?;

        // SAFETY: Make sure relative UNIX socket paths match process CWD.
        #[allow(clippy::cast_possible_truncation)]
        if let Some(path) = addr.as_unix_addr().and_then(|a| a.path()) {
            if path.is_relative() {
                // Concatenate with /proc/pid/cwd.
                let mut buf = itoa::Buffer::new();
                let mut proc = PathBuf::from("/proc");
                let pid = request.get_request().pid;
                proc.push(buf.format(pid));
                proc.push("cwd");
                proc.push(path);
                let path = proc;
                // Convert the path to a CString
                let osstr: &OsStr = path.as_ref();
                let bytes = osstr.as_bytes();
                let null = bytes.iter().position(|&b| b == 0).unwrap_or(bytes.len());
                let cstr = CString::new(&bytes[..null]).map_err(|_| Errno::EINVAL)?;

                // Create sockaddr_un struct.
                const UNIX_PATH_MAX: usize = 108;
                let mut sockaddr = nix::libc::sockaddr_un {
                    sun_family: nix::libc::AF_UNIX as nix::libc::sa_family_t,
                    sun_path: [0; UNIX_PATH_MAX],
                };

                let bytes_with_nul = cstr.as_bytes_with_nul();
                if bytes_with_nul.len() > UNIX_PATH_MAX {
                    return Err(Errno::ENAMETOOLONG);
                }

                // Manually copy the bytes.
                // TODO: Is there a better way?
                #[allow(clippy::cast_possible_wrap)]
                for (dst, &src) in sockaddr.sun_path.iter_mut().zip(bytes_with_nul.iter()) {
                    *dst = src as nix::libc::c_char;
                }

                // SAFETY: We are converting a sockaddr_un to a
                // SockaddrStorage using a raw pointer. The sockaddr_un
                // is valid for the duration of this operation, ensuring
                // the safety of the pointer. However, this operation is
                // inherently unsafe due to direct pointer manipulation.
                addr = unsafe {
                    SockaddrStorage::from_raw(
                        std::ptr::addr_of!(sockaddr) as *const _,
                        Some(
                            bytes_with_nul
                                .len()
                                .saturating_add(std::mem::size_of::<nix::libc::sa_family_t>())
                                as nix::libc::socklen_t,
                        ),
                    )
                }
                .ok_or(Errno::EINVAL)?;
            }
        }

        let family = match addr.family() {
            Some(AddressFamily::Unix) => AddressFamily::Unix,
            Some(AddressFamily::Inet) => AddressFamily::Inet,
            Some(AddressFamily::Inet6) => AddressFamily::Inet6,
            _ if allow_unsupported_socket_families => {
                return unsafe { Ok(request.continue_syscall()) }
            }
            _ => return Err(Errno::EAFNOSUPPORT),
        };

        handler(proc, request, req, &addr, syscall_name, cap)?;

        let fd = proc.get_fd(args[0] as RawFd)?;
        match syscall_name {
            "bind" => handle_bind(&fd, &addr, allow_successful_bind, family, request),
            "connect" => handle_connect(&fd, &addr, request),
            "recvfrom" => handle_recvfrom(&fd, args, proc, request, addr_len, family),
            "sendto" => handle_sendto(&fd, args, proc, request, &addr),
            _ => unreachable!(),
        }
    })
}

#[allow(clippy::cognitive_complexity)]
fn handle_bind(
    fd: &OwnedFd,
    addr: &SockaddrStorage,
    allow_successful_bind: bool,
    family: AddressFamily,
    request: &UNotifyEventRequest,
) -> Result<ScmpNotifResp, Errno> {
    let unix_sock = addr.as_unix_addr().and_then(|a| a.path());

    let result = if unix_sock.is_some() {
        let req = request.get_request();
        #[allow(clippy::cast_possible_wrap)]
        let mask = match proc_umask(Some(Pid::from_raw(req.pid as i32))) {
            Ok(mask) => mask.bits(),
            Err(_) => {
                return Err(Errno::EACCES);
            }
        };
        let fd = fd.as_raw_fd();
        let addr = addr.as_ptr();
        #[allow(clippy::cast_possible_truncation)]
        const ADDR_LEN: nix::libc::socklen_t =
            std::mem::size_of::<nix::libc::sockaddr_un>() as nix::libc::socklen_t;
        let flags = CloneFlags::CLONE_VM | CloneFlags::CLONE_VFORK | CloneFlags::CLONE_SIGHAND;
        let mut nstack = [0u8; 32 * 1024];
        let tid = match clone(
            Box::new(|| -> isize {
                // SAFETY: Honour process' umask.
                // Note, the umask is per-thread here.
                let _ = unsafe { nix::libc::umask(mask) };
                if !request.is_valid() {
                    return Errno::EACCES as isize;
                }
                if unsafe { nix::libc::syscall(nix::libc::SYS_bind, fd, addr, ADDR_LEN) } == 0 {
                    0
                } else {
                    Errno::last() as isize
                }
            }),
            &mut nstack[..],
            flags,
            Some(SIGCHLD as i32),
        ) {
            Ok(tid) => tid,
            Err(_) => {
                return Err(Errno::EACCES);
            }
        };

        // Note: since we vfork, waitpid interruption can only happen
        // due to syd::m☮☮n sending SIGUSR1 which means the request has
        // been invalidated in which case we return EACCES.
        match waitpid(tid, None) {
            Ok(WaitStatus::Exited(_, 0)) => Ok(()),
            Ok(WaitStatus::Exited(_, n)) => Err(Errno::from_i32(n)),
            _ => Err(Errno::EACCES),
        }
    } else {
        bind(fd.as_raw_fd(), addr)
    };
    if let Err(error) = result {
        return Err(error);
    } else if !allow_successful_bind {
        return Ok(request.return_syscall(0));
    }

    // Handle allow_successful_bind
    let cmd = match family {
        AddressFamily::Unix => {
            let addr = addr.as_unix_addr().ok_or(Errno::EINVAL)?;
            match (addr.path(), addr.as_abstract()) {
                (Some(path), _) => {
                    let path = path.as_os_str().as_bytes();
                    let null = path.iter().position(|&b| b == 0).unwrap_or(path.len());
                    let path = PathBuf::from(OsStr::from_bytes(&path[..null]));
                    if let Ok(path) =
                        crate::fs::canonicalize(path, true, crate::fs::MissingHandling::Missing)
                    {
                        Some(format!("allowlist/net/connect+{}", path.display()))
                    } else {
                        None
                    }
                }
                (_, Some(path)) => {
                    let null = path.iter().position(|&b| b == 0).unwrap_or(path.len());
                    let path = PathBuf::from(OsStr::from_bytes(&path[..null]));
                    Some(format!("allowlist/net/connect+{}", path.display()))
                }
                _ => {
                    // Unnamed UNIX socket
                    None
                }
            }
        }
        AddressFamily::Inet => {
            let addr = addr.as_sockaddr_in().ok_or(Errno::EINVAL)?;
            let mut port = addr.port();
            let addr = IpAddr::V4(Ipv4Addr::from(addr.ip()));
            if port == 0 {
                port = getsockname::<SockaddrStorage>(fd.as_raw_fd())?
                    .as_sockaddr_in()
                    .ok_or(Errno::EINVAL)?
                    .port();
            }
            Some(format!("allowlist/net/connect+{addr}!{port}"))
        }
        AddressFamily::Inet6 => {
            let addr = addr.as_sockaddr_in6().ok_or(Errno::EINVAL)?;
            let mut port = addr.port();
            let addr = IpAddr::V6(addr.ip());
            if port == 0 {
                port = getsockname::<SockaddrStorage>(fd.as_raw_fd())?
                    .as_sockaddr_in6()
                    .ok_or(Errno::EINVAL)?
                    .port();
            }
            Some(format!("allowlist/net/connect+{addr}!{port}"))
        }
        _ => None,
    };
    if let Some(cmd) = cmd {
        let mut sandbox = request.get_sandbox(true);
        let result = sandbox.config(&cmd);
        let pid = request.get_request().pid;
        match result {
            Ok(_) => {
                if let Err(error) = sandbox.build_globsets() {
                    error!("ctx": "config",
                        "pid": pid,
                        "sys": "bind",
                        "cfg": cmd,
                        "err": error.to_string());
                } else {
                    info!("ctx": "config",
                        "pid": pid,
                        "sys": "bind",
                        "cfg": cmd);
                }
            }
            Err(error) => {
                error!("ctx": "config",
                    "pid": pid,
                    "sys": "bind",
                    "cfg": cmd,
                    "err": error.to_string());
            }
        };
        drop(sandbox);
    }
    Ok(request.return_syscall(0))
}

fn handle_connect(
    fd: &OwnedFd,
    addr: &SockaddrStorage,
    request: &UNotifyEventRequest,
) -> Result<ScmpNotifResp, Errno> {
    connect(fd.as_raw_fd(), addr).map(|_| request.return_syscall(0))
}

fn handle_recvfrom(
    fd: &OwnedFd,
    args: &[u64; 6],
    proc: &RemoteProcess,
    request: &UNotifyEventRequest,
    addr_len: usize,
    addr_family: AddressFamily,
) -> Result<ScmpNotifResp, Errno> {
    // SAFETY: The length argument to the recvfrom call
    // must not be fully trusted, it can be overly large,
    // and allocating a Vector of that capacity may overflow.
    #[allow(clippy::cast_possible_truncation)]
    let len = args[2] as usize;
    let len = len.min(1000000); // Cap count at 1mio
    let mut buf = vec![0u8; len];
    let (n, addr) = recvfrom::<SockaddrStorage>(fd.as_raw_fd(), &mut buf)?;
    #[allow(clippy::cast_possible_truncation)]
    proc.write_mem(&buf[..n], args[1] as usize, request)?;
    if let Some(addr) = addr {
        #[allow(clippy::cast_possible_truncation)]
        let addr_remote = args[4] as usize;
        match addr_family {
            AddressFamily::Inet => {
                // SAFETY: TODO
                let addr = unsafe {
                    any_as_u8_slice(addr.as_sockaddr_in().ok_or(Errno::EINVAL)?.as_ref())
                };
                proc.write_mem(addr, addr_remote, request)?;
            }
            AddressFamily::Inet6 => {
                // SAFETY: TODO
                let addr = unsafe {
                    any_as_u8_slice(addr.as_sockaddr_in6().ok_or(Errno::EINVAL)?.as_ref())
                };
                proc.write_mem(addr, addr_remote, request)?;
            }
            AddressFamily::Unix => {
                // SAFETY: TODO
                let addr =
                    unsafe { any_as_u8_slice(addr.as_unix_addr().ok_or(Errno::EINVAL)?.as_ref()) };
                proc.write_mem(addr, addr_remote, request)?;
            }
            _ => unreachable!(),
        }
        let addr_len = unsafe { any_as_u8_slice(&addr_len) };
        #[allow(clippy::cast_possible_truncation)]
        proc.write_mem(addr_len, args[5] as usize, request)?;
    }
    #[allow(clippy::cast_possible_wrap)]
    Ok(request.return_syscall(n as i64))
}

fn handle_sendto(
    fd: &OwnedFd,
    args: &[u64; 6],
    proc: &RemoteProcess,
    request: &UNotifyEventRequest,
    addr: &SockaddrStorage,
) -> Result<ScmpNotifResp, Errno> {
    // SAFETY: The length argument to the sendto call
    // must not be fully trusted, it can be overly large,
    // and allocating a Vector of that capacity may overflow.
    #[allow(clippy::cast_possible_truncation)]
    let len = args[2] as usize;
    let len = len.min(1000000); // Cap count at 1mio
    #[allow(clippy::cast_possible_truncation)]
    let flags = MsgFlags::from_bits_truncate(args[3] as nix::libc::c_int);
    let mut buf = vec![0u8; len];
    #[allow(clippy::cast_possible_truncation)]
    proc.read_mem(&mut buf, args[1] as usize, request)?;
    let n = sendto(fd.as_raw_fd(), &buf, addr, flags)?;
    #[allow(clippy::cast_possible_wrap)]
    Ok(request.return_syscall(n as i64))
}

fn get_addr(
    proc: &RemoteProcess,
    request: &UNotifyEventRequest,
    addr_remote: usize,
    addr_len: usize,
) -> Result<SockaddrStorage, Errno> {
    let mut addr = vec![0u8; addr_len];
    proc.read_mem(&mut addr, addr_remote, request)?;

    // SAFETY: Invoking `SockaddrStorage::from_raw` is safe because:
    // 1. The memory location of `sockaddr_ptr` is valid, correctly aligned,
    //    and readable for `addr_len` bytes, preventing segmentation faults.
    // 2. The memory is allocated based on a valid `sockaddr` structure.
    // 3. The `addr_len` is correctly representing the size of the actual `sockaddr` structure.
    // 4. There are no concurrent writes to the memory location while reading.
    match unsafe {
        #[allow(clippy::cast_ptr_alignment)]
        #[allow(clippy::cast_possible_truncation)]
        SockaddrStorage::from_raw(
            addr.as_ptr() as *const nix::libc::sockaddr,
            Some(addr_len as nix::libc::socklen_t),
        )
    } {
        Some(addr) => Ok(addr),
        None => {
            // Invalid socket address of length `addr_len`.
            Err(Errno::EINVAL)
        }
    }
}

fn get_addr_len(
    args: &[u64; 6],
    arg_len: usize,
    len_ptr: bool,
    proc: &RemoteProcess,
    request: &UNotifyEventRequest,
) -> Result<usize, Errno> {
    let addr_len: usize = if !len_ptr {
        args[arg_len].try_into().map_err(|_| Errno::EINVAL)?
    } else if args[arg_len] == 0 {
        // Careful here:
        // recvfrom with NULL address length is required for connection-mode sockets.
        // assert!(syscall_name == "recvfrom");
        0
    } else {
        let mut l: nix::libc::socklen_t = 0;
        // SAFETY:
        // We are using the `any_as_u8_mut_slice` function to obtain a mutable byte slice from
        // `l`. We must ensure that the lifetime of this slice does not exceed `l`, and that
        // the memory representation of `l` is correctly interpreted when accessed as bytes.
        // Also, as we're getting a mutable slice, we need to guarantee that no other parts of
        // the program will simultaneously read or write to `l` or the memory area we're
        // interfacing with to prevent data races and undefined behavior.
        let l_slice = unsafe { any_as_u8_mut_slice(&mut l) };
        #[allow(clippy::cast_possible_truncation)]
        proc.read_mem(l_slice, args[arg_len] as usize, request)?;
        l as usize
    };

    Ok(addr_len)
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn glob_proc() {
        assert!(GLOB_PROC.is_match("/proc/256677/fd/pipe:[2379377]"));
        assert!(GLOB_PROC.is_match("/proc/256677/task/123/fd/pipe:[2379377]"));
        assert!(GLOB_PROC.is_match("/proc/256677/fd/foo (deleted)"));
    }
}

/*
#[cfg(test)]
mod tests {
    use std::{ffi::CStr, time::Duration};

    use nix::sys::signal::{kill, Signal::SIGKILL};

    use super::*;

    #[test]
    fn smoke_test_sleep() {
        fn openat_handler(req: &UNotifyEventRequest) -> libseccomp::ScmpNotifResp {
            let path = req.get_request().data.args[1];
            let remote = RemoteProcess::new(Pid::from_raw(req.request.pid as i32)).unwrap();
            let mut buf = [0u8; 256];
            remote.read_mem(&mut buf, path as usize).unwrap();
            eprintln!("open (read from remote): {:?}", buf);
            let path = CStr::from_bytes_until_nul(&buf).unwrap();
            if !req.is_valid() {
                return req.fail_syscall(libc::EACCES);
            }
            eprintln!("open (path CStr): {:?}", path);
            unsafe { req.continue_syscall() }
        }

        let mut supervisor = Supervisor::new(2).unwrap();
        supervisor.insert_handler(ScmpSyscall::new("openat"), openat_handler);
        let mut cmd = Command::new("/bin/sleep");
        let cmd = cmd.arg("1");
        let (pid, thread_handle, pool) = supervisor.spawn(cmd).unwrap();
        let status = Supervisor::wait(pid, thread_handle, pool).unwrap();
        assert!(status.success());
    }

    #[test]
    fn smoke_test_whoami() {
        fn geteuid_handler(req: &UNotifyEventRequest) -> libseccomp::ScmpNotifResp {
            req.return_syscall(0)
        }

        let mut supervisor = Supervisor::new(2).unwrap();
        supervisor.insert_handler(ScmpSyscall::new("geteuid"), geteuid_handler);
        let mut cmd = Command::new("/usr/bin/whoami");
        let cmd = cmd.stdout(Stdio::piped());
        let (pid, thread_handle, pool) = supervisor.spawn(cmd).unwrap();
        let status = Supervisor::wait(pid, thread_handle, pool).unwrap();
        assert!(status.success());
        let whoami_stdout = child.stdout.as_mut().unwrap();
        let mut buf = String::new();
        whoami_stdout.read_to_string(&mut buf).unwrap();
        assert_eq!(buf.trim(), "root");
    }

    #[test]
    fn test_sleep_blocking_syscall() {
        fn clock_nanosleep_handler(req: &UNotifyEventRequest) -> libseccomp::ScmpNotifResp {
            // sleep for extra 60s
            // Please note that it may bring A LOT OF PROBLEMS if you try using pthread_cancel
            // So here we just use the easy way: check valid in the loop
            let (tx, rx) = std::sync::mpsc::channel();
            let handler = std::thread::spawn(move || {
                for _ in 0..60 {
                    if rx.try_recv().is_ok() {
                        break;
                    }
                    std::thread::sleep(Duration::from_secs(1));
                }
            });
            // while handler is running, check valid in the loop
            loop {
                if !req.is_valid() {
                    // cancel the thread
                    eprintln!("canceling thread as req is invalid now");
                    tx.send(()).unwrap();
                    break;
                }
                std::thread::sleep(Duration::from_millis(100));
            }
            handler.join().unwrap();
            unsafe { req.continue_syscall() }
        }

        let mut supervisor = Supervisor::new(2).unwrap();
        supervisor.insert_handler(ScmpSyscall::new("clock_nanosleep"), clock_nanosleep_handler);
        let mut cmd = Command::new("/bin/sleep");
        let cmd = cmd.arg("120");
        let (pid, thread_handle, pool) = supervisor.spawn(cmd).unwrap();
        std::thread::spawn(move || {
            std::thread::sleep(Duration::from_secs(1));
            // kill the child process
            kill(pid, SIGKILL).unwrap();
        });
        let _ = Supervisor::wait(pid, thread_handle, pool).unwrap();
    }

    #[test]
    fn test_new_fd() {
        fn openat_handler(req: &UNotifyEventRequest) -> libseccomp::ScmpNotifResp {
            let path = req.get_request().data.args[1];
            let remote = RemoteProcess::new(Pid::from_raw(req.request.pid as i32)).unwrap();
            let mut buf = [0u8; 256];
            remote.read_mem(&mut buf, path as usize).unwrap();
            eprintln!("open (read from remote): {:?}", buf);
            let path = CStr::from_bytes_until_nul(&buf).unwrap();
            if !req.is_valid() {
                return req.fail_syscall(libc::EACCES);
            }
            eprintln!("open (path CStr): {:?}", path);
            if path.to_str().unwrap() == "/etc/passwd" {
                // open /etc/resolv.conf instead
                let file = File::open("/etc/resolv.conf").unwrap();
                let fd = file.as_raw_fd();
                let remote_fd = req.add_fd(fd).unwrap();
                req.return_syscall(remote_fd as i64)
            } else {
                unsafe { req.continue_syscall() }
            }
        }

        let mut supervisor = Supervisor::new(2).unwrap();
        supervisor.insert_handler(ScmpSyscall::new("openat"), openat_handler);
        let mut cmd = Command::new("/bin/cat");
        let cmd = cmd.arg("/etc/passwd").stdout(Stdio::piped());
        let (pid, thread_handle, pool) = supervisor.spawn(cmd).unwrap();
        let status = Supervisor::wait(pid, thread_handle, pool).unwrap();
        assert!(status.success());
        let cat_stdout = child.stdout.as_mut().unwrap();
        let mut buf = String::new();
        cat_stdout.read_to_string(&mut buf).unwrap();
        assert!(buf.contains("nameserver"));
    }
}
*/
